/* retawq/stuff.c - general stuff
   This file is part of retawq (<http://retawq.sourceforge.net/>), a network
   client created by Arne Thomassen; retawq is basically released under certain
   versions of the GNU General Public License and WITHOUT ANY WARRANTY.
   Read the file COPYING for license details, README for program information.
   Copyright (C) 2001-2004 Arne Thomassen <arne@arne-thomassen.de>
*/

#include "stuff.h"

declare_local_i18n_buffer
__define_filestuff

tConfiguration config;

const_after_init tProgramMode program_mode = pmEnvironed;
const char* initial_messages = NULL;
unsigned char launch_uri_count = 0;
const char* launch_uri[2]; /* documents to load on launch (if environed) */

#if MIGHT_DO_TERMIOS
tBoolean do_restore_termios = falsE;
struct termios saved_termios;
#endif

tBoolean need_tglib_cleanup = falsE;
const_after_init tBoolean lfdmbs = falsE;

#if CONFIG_TG != TG_GTK
tBoolean use_colors = falsE;
#endif

#if TGC_IS_CURSES
tColorBitmask color_bitmask[cpnMax + 1];
#endif

#if CONFIG_DEBUG
const_after_init int debugfd = 2;
static char debugstrbuf[STRBUF_SIZE];
#endif


/* Strings */

const char strCopyright[] = "retawq " RETAWQ_VERSION
  " - Copyright (C) 2001-2004 Arne Thomassen <arne@arne-thomassen.de>\n";
const char strRetawq[] = "retawq", strTG[] = TGSTR, strOptHelp[] = "--help",
  strVersion[] = "retawq " RETAWQ_VERSION;
const char strEmpty[] = "", strNewline[] = "\n";
const char strMinus[] = "-", strSlash[] = "/", strAsterisk[] = "*",
  strLt[] = "<", strGt[] = ">", strHm[] = "#", strQm[] = "?",
  strDoubleQuote[] = "\"", strSpacedDash[] = " - ", strDoubleDot[] = "..",
  str1dot0[] = "1.0", str1dot1[] = "1.1";
static const char strPercld[] = "%ld";
const char strPercsPercs[] = "%s%s", strPercd[] = "%d";
  /* for sprintf()-style formatting */
const char strHexnum[] = "0123456789abcdef";
const char strA[] = "a", strShtml[] = "shtml", strReset[] = "reset",
  strSelect[] = "select", strSubmit[] = "submit", strText[] = "text",
  strClose[] = "close", strQuit[] = "quit";
const char strLocalhost[] = "localhost";
const char strUnknown[] = N_("(unknown)"), strCheckbox[] = N_("checkbox"),
  strBracedDisabled[] = N_("(disabled)"), strButton[] = N_("button"),
  strImage[] = N_("image"), strYes[] = N_("yes"), strNo[] = N_("no"),
  strFileUc[] = N_("File"), strDirectoryUc[] = N_("Directory"),
  strLink[] = N_("link"), strLoginFailed[] = N_("Login failed"),
  strFwdact[] = N_("feature \"%s\" was disabled at compile-time%s");
#if !CONFIG_SESSIONS
const char strFsessions[] = N_("Fsessions");
#endif
#if !CONFIG_JUMPS
const char strFjumps[] = N_("Fjumps");
#endif
#if OPTION_EXECEXT != EXECEXT_ALL
const char strFexecext[] = N_("F(this kind of) execution of external programs");
#endif
#if CONFIG_CONSOLE
const char strConsoleDiscard[] = "cdis";
#endif
#if (CONFIG_TG == TG_X) || (CONFIG_TG == TG_XCURSES)
const char strCantOpenXws[] = N_("can't open X Window System display");
#endif
#if OPTION_LOCAL_CGI || OPTION_TRAP || (OPTION_EXECEXT & 1)
const char strSoftwareId[] = "retawq/" RETAWQ_VERSION;
#endif


/* scheme strings: */
const char strHttp[] = "http", strFtp[] = "ftp", strLocal[] = "local",
  strFile[] = "file", strAbout[] = "about", strFinger[] = "finger",
  strLocalCgi[] = "local-cgi", strNntp[] = "nntp", strNews[] = "news",
  strCvs[] = "cvs", strGopher[] = "gopher", strInfo[] = "info",
  strMailto[] = "mailto", strJavascript[] = "javascript", strHttps[] = "https",
  strFtps[] = "ftps", strPop[] = "pop", strPops[] = "pops",
  strExecextShell[] = "execext-shell";


/* Functions */

int streqcase3(const char* _str1, const char* _str2)
/* This function is a useful mixture of strcmp() and strcasecmp(): it compares
   the lower-case variant of <str1> with <str2> (not "with the lower-case
   variant of <str2>"!). The "3" in the function name reminds of the
   "tristate" nature of the return value (< 0, == 0, > 0), as opposed to
   the "two-state" (== 0, != 0) return value of streqcase(). */
{ const unsigned char *str1 = (const unsigned char*) _str1,
    *str2 = (const unsigned char*) _str2;
  unsigned char ch1, ch2;
  do
  { ch1 = *str1++; ch2 = *str2++;
    if (ch1 == '\0') break;
    ch1 = (unsigned char) my_tolower(ch1);
  } while (ch1 == ch2);
  return((int) (ch1 - ch2));
}

int strneqcase3(const char* _str1, const char* _str2, size_t maxlen)
/* like streqcase3(), but compares at most <maxlen> characters */
{ int retval;
  if (maxlen <= 0) retval = 0;
  else
  { const unsigned char *str1 = (const unsigned char*) _str1,
      *str2 = (const unsigned char*) _str2;
    unsigned char ch1, ch2;
    do
    { ch1 = *str1++; ch2 = *str2++;
      if (ch1 == '\0') break;
      ch1 = (unsigned char) my_tolower(ch1);
    } while ( (ch1 == ch2) && (--maxlen > 0) );
    retval = ((int) (ch1 - ch2));
  }
  return(retval);
}

void my_atoi(const char* src, int* value, const char** next, int maxvalue)
/* similar to atoi(), but allows only non-negative values, allows to define a
   maximum possible value, and additionally returns the position of the first
   non-numeric character */
{ int v = 0;
  while (1)
  { char ch = *src;
    if (my_isdigit(ch))
    { v = 10 * v + ((int)(ch - '0'));
      src++;
      if (v > maxvalue)
      { v = maxvalue;
        while (my_isdigit(*src)) { src++; } /* skip number */
        goto done;
      }
    }
    else
    { done:
      *value = v;
      if (next != NULL) *next = src;
      return;
    }
  }
}

char* my_strnchr(const char* str, int _ch, int count)
/* like strchr(), but checks at most <count> characters */
{ char ch = ((char) _ch);
  while (count-- > 0)
  { char c = *str++;
    if (c == '\0') break;
    else if (c == ch) return(unconstify(str) - 1);
  }
  return(NULL);
}

char* my_strncasestr(const char* haystack, const char* needle,
  const size_t hlen)
/* some mixture of strncasecmp() and strstr() */
{ const size_t nlen = strlen(needle);
  size_t h = 0, n = 0;
  do
  { if (my_tolower(haystack[h]) == my_tolower(needle[n])) n++;
    else { h -= n; n = 0; }
    h++;
  } while ( (n < nlen) && (h < hlen) );
  if (n >= nlen) return(unconstify(haystack) + h - nlen - 1); /* found */
  else return(NULL);
}

my_inline tBoolean is_suffix(const char* str, const char* suff)
/* returns whether <str> ends with (or equals!) <suff> */
{ size_t str_len = strlen(str), suff_len = strlen(suff);
  if ( (str_len >= suff_len) && (!strcmp(str + (str_len - suff_len), suff)) )
    return(truE);
  else return(falsE);
}

tBoolean my_pattern_matcher(const char* pattern, const char* str)
/* a _very_ simple pattern matching algorithm - but it's enough for us */
{ if (pattern[0] == '*') return(is_suffix(str, pattern + 1));
  else
  { if (!strcmp(pattern, str)) return(truE);
    else return(falsE);
  }
}

static void does_not_return do_quit_msg(int exitcode, const char* msg)
{ if (need_tglib_cleanup)
  {
#if CONFIG_TG == TG_X
    XCloseDisplay(xws_display);
#elif CONFIG_TG == TG_GTK
    gtk_main_quit();
#else
    (void) endwin();
#endif
  }
#if MIGHT_DO_TERMIOS
  if (do_restore_termios)
  { int err;
    unsigned char loopcount = 0;
    do
    { err = tcsetattr(0, TCSAFLUSH, &saved_termios);
    } while ( (err == -1) && (errno == EINTR) && (++loopcount < 100) );
  }
#endif
  if (msg != NULL)
  { /* if (lfdmbs) */
    { int fd = ( (exitcode != 0) ? 2 : 1 );
      my_write_str(fd, msg);
    }
    debugmsg(msg);
  }
  resource_quit();
  exit(exitcode);
}

void do_quit(void)
/* quit without errors */
{ do_quit_msg(0, NULL);
}

void do_quit_sig(void)
/* terminated by a signal */
{ do_quit_msg(2, _("retawq: terminated by a signal\n"));
}

void fatal_error(int err, const char* str)
/* quits the program due to a fatal error; <err> must be errno or an E...
   constant or 0. */
{ static const char str2[] = N_("; error #%d, %s"),
    str1[] = N_("retawq: fatal error: %s%s\n"); /* ugly colon-ialism... */
  static tBoolean is_recursive = falsE;
  tBoolean is_rec = is_recursive;
  char buf1[1000], buf2[1000];
  const char* ptr;
  is_recursive = truE;
  if (err <= 0) ptr = strEmpty;
  else
  { const char* errstr = (is_rec ? __my_strerror(err) : my_strerror(err));
    tBoolean too_long = cond2boolean(strlen(errstr) > 200);
    sprint_safe(buf2, (is_rec ? str2 : _(str2)), err,
      (too_long ? strUnknown : errstr));
    ptr = buf2;
  }
  sprint_safe(buf1, (is_rec ? str1 : _(str1)), str, ptr);
  do_quit_msg(1, buf1);
}

void fatal_error_writefailed(int err)
{ fatal_error(((err == -1) ? errno : 0), _("write() failed"));
}

static void does_not_return out_of_memory(void)
{ fatal_error(ENOMEM, _("out of memory"));
}

#if !CONFIG_DEBUG
my_inline /* gcc/gdb and line numbers... */
#endif
void* __memory_allocate(size_t size)
{ void* ptr = malloc(size);
  if (ptr == NULL) out_of_memory();
  return(ptr);
}

void* memory_allocate(size_t size)
{ void* ptr = __memory_allocate(size);
  my_memclr(ptr, size);
  return(ptr);
}

void* memory_reallocate(void* ptr, size_t size)
{ void* ptr2 = realloc(ptr, size);
  if (ptr2 == NULL) out_of_memory();
  return(ptr2);
}

void memory_deallocate(const void* ptr)
{ /* guess what... */
  free(__unconstify(void*, ptr));
}

#if NEED_SPECIAL_GETTEXT
#define is_special_char(ch) ( (ch == '') || (ch == '') || (ch == '') || \
  (ch == '') || (ch == '') || (ch == '') || (ch == '') )
char* my_do_gettext(char** buffer, const char* str)
{ if (str == NULL) /* deallocate old strings */
  { char* ptr = *buffer;
    while (ptr != NULL)
    { char* next = *((char**) ptr);
      memory_deallocate(ptr);
      ptr = next;
    }
    *buffer = NULL;
    return(NULL);
  }
  else /* translate and make seven-bit-clean */
  { static const char charmap[] = "?????????\t\n??\r?????????????????? !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~??????????????????????????????????!????|????\"?-?-????'??????\"????AAAAAA?CEEEEIIIIDNOOOOO*OUUUUYPsaaaaaa?ceeeeiiiidnooooo/ouuuuypy";
    const char *s = gettext(str), *src;
    char *buf, *dest, *retval, ch;
    size_t len;
    if (s == NULL) return(NULL); /* "should not happen" */
    src = s; len = 0;
    while ( (ch = *src++) != '\0' ) { len++; if (is_special_char(ch)) len++; }
    buf = __memory_allocate(sizeof(char*) + len + 1);
    dest = retval = buf + sizeof(char*);
    src = s;
    while ( (ch = *src++) != '\0' )
    { *dest++ = charmap[(unsigned short) ((unsigned char) ch)];
      if (is_special_char(ch))
      { if (ch == '') *dest++ = 's';
        else *dest++ = 'e';
      }
    }
    *dest = '\0';
    *((char**) buf) = *buffer;
    *buffer = buf;
    return(retval);
  }
}
#undef is_special_char
#endif /* #if NEED_SPECIAL_GETTEXT */

char* my_strdup(const char* str)
/* duplicates a string; the standard library function strdup() isn't portable,
   and callers always have to check against a NULL return value - two reasons
   for using something different... */
{ size_t size = strlen(str) + 1;
  char* retval = __memory_allocate(size);
  my_memcpy(retval, str, size);
  return(retval);
}

char* my_strndup(const char* str, size_t len)
{ char* retval = __memory_allocate(len + 1);
  strncpy(retval, str, len); retval[len] = '\0';
  return(retval);
}

char* my_strdup_tolower(const char* str)
/* duplicates a string, changing uppercase to lowercase */
{ char* retval = __memory_allocate(strlen(str) + 1);
  my_strcpy_tolower(retval, str);
  return(retval);
}

char* my_strndup_tolower(const char* str, size_t len)
{ char *retval = __memory_allocate(len + 1), *ptr = retval;
  while (len-- > 0)
  { const char ch = *str++;
    *ptr++ = my_tolower(ch);
  }
  *ptr = '\0';
  return(retval);
}

const char* get_homepath(void)
/* tries to find out the user's home directory */
{ static const char* homepath = NULL;
  if (homepath == NULL) /* not yet calculated */
  { const char* const env = getenv("HOME"); /* user's home directory */
    if ( (env == NULL) || (env[0] == '\0') ) homepath = strEmpty;
    else
    { tBoolean has_trailslash = cond2boolean(env[strlen(env) - 1] == '/');
      char* spfbuf;
      my_spf(NULL, 0, &spfbuf, strPercsPercs, env,
        (has_trailslash ? strEmpty : strSlash));
      homepath = my_spf_use(spfbuf);
    }
  }
  return(homepath);
}

ssize_t my_write(int fd, const void* buf, size_t _count)
/* will write <_count> bytes (or fail on error) */
{ size_t countleft = _count;
  while (countleft > 0)
  { ssize_t err = __my_write(fd, buf, countleft);
    if (err == 0) { err = -1; errno = EIO; } /* CHECKME! */
    if (err < 0) return(err);
    else if (err > (ssize_t) countleft) break; /* "can't happen" */
    countleft -= err; buf = ((char*) buf) + err;
  }
  return((ssize_t) _count);
}

unsigned char my_mmap_file_readonly(const char* filename, void** _b,
  size_t* _s)
/* tries to open a regular file in read-only mode and to map it into memory;
   return value: 0=error, 1=empty, 2=fine */
{ unsigned char retval = 0; /* assume failure */
  int fd = my_open(filename, O_RDONLY);
  struct stat statbuf;
  size_t size;
  void* filebuf;
  if (fd < 0) goto out;
  if (my_fstat(fd, &statbuf) != 0) goto cleanup;
  if (!S_ISREG(statbuf.st_mode))
  { if (S_ISDIR(statbuf.st_mode)) errno = EISDIR;
    goto cleanup;
  }
  size = statbuf.st_size;
  if (size <= 0) { retval = 1; goto cleanup; }
  filebuf = my_mmap(size, fd);
  if (filebuf == MAP_FAILED) goto cleanup;
  my_madvise_sequential(filebuf, size);
  *_b = filebuf; *_s = size; retval = 2;
  cleanup: { int e = errno; my_close(fd); errno = e; }
  out: return(retval);
}

#if CONFIG_LOCALDIR > 1
void check_localdirsort(char* str)
/* checks and possibly fixes the "sorting" string for local directory sorting,
   in both run-time configuration entries and URI parts */
{ const char* src;
  char ch, *dest;
  unsigned char bits = 0;
  if (*str == '_')
  { /* disables all sorting, no further check necessary */
    *(str + 1) = '\0'; /* ignore any further characters */
    return;
  }
  src = dest = str;
  while ( (ch = *src++) != '\0' )
  { static const char allowed[] = "gimnstu";
      /* (group ID, case-insensitive name, modification time, name, size,
          type (file/directory/...), user ID) */
    const char lch = my_tolower(ch), *pos = my_strchr(allowed, lch);
    if (pos != NULL) /* it's an _allowed_ sorting option */
    { char val = (char) (pos - allowed);
      if (!my_bit_test(&bits, val)) /* only if option not yet used */
      { my_bit_set(&bits, val);
        *dest++ = ch;
      }
    }
  }
  *dest = '\0';
}
#endif


/* Objects */

my_enum1 enum
{ ofNone = 0, ofVanished = 0x01
} my_enum2(unsigned char) tObjectFlags;
/* (ofVanished is the main point of the whole objects system) */

typedef struct
{ tObjectIdentifier id;
  size_t num_watchers;
  tObjectFlags flags;
} tObject;

typedef int tObjectIndex;
#define OBJECT_INDEX_INVALID (-1)

static tObject* objects_watched = NULL;
static tObjectIndex objects_watched_num = 0, objects_watched_maxnum = 0;

tObjectIdentifier object_id_create(void)
{ static tObjectIdentifier id = 1;
  tObjectIdentifier retval = id++;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "object_id_create(): %d\n", retval);
  debugmsg(debugstrbuf);
#endif
  if (retval == 0)
  { /* virtually "can't happen" - which user might create 2^32 objects in one
       session? Even with an automated test/attack tool, the creation of so
       many objects would take... erm... weeks? */
    fatal_error(0, _("too many objects"));
  }
  return(retval);
}

static tObjectIndex object_lookup(tObjectIdentifier id)
/* Usually there might be 1 or 2 watched objects, in rare cases maybe a little
   more. No reason for hashing, sorted lists or any other such stuff. KISS! */
{ tObjectIndex retval = OBJECT_INDEX_INVALID, i;
  for (i = 0; i < objects_watched_num; i++)
  { if (objects_watched[i].id == id) { retval = i; break; /* found */ } }
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "object_lookup(%d): %d\n", id, retval);
  debugmsg(debugstrbuf);
#endif
  return(retval);
}

void object_watch(tObjectIdentifier id)
{ tObjectIndex i = object_lookup(id);
  if (i == OBJECT_INDEX_INVALID) /* not yet being watched */
  { tObject* object;
    if (objects_watched_maxnum <= objects_watched_num)
    { objects_watched_maxnum += 8; /* usually more than enough */
      objects_watched = memory_reallocate(objects_watched,
        objects_watched_maxnum * sizeof(tObject));
    }
    object = &(objects_watched[objects_watched_num]);
    my_memclr(object, sizeof(tObject)); object->id = id;
    i = objects_watched_num++;
  }
  objects_watched[i].num_watchers++;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "object_watch(%d): %d,%d\n", id, i,
    objects_watched[i].num_watchers);
  debugmsg(debugstrbuf);
#endif
}

void object_unwatch(tObjectIdentifier id)
{ tObjectIndex i = object_lookup(id);
  tObject* object;
  if (i == OBJECT_INDEX_INVALID) return; /* "should not happen" */
  object = &(objects_watched[i]);
  if (object->num_watchers > 0) object->num_watchers--; /* "should" be true */
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "object_unwatch(%d): %d,%d\n", id, i,
    object->num_watchers);
  debugmsg(debugstrbuf);
#endif
  if (object->num_watchers <= 0) /* no longer watched by anyone */
  { if (i < objects_watched_num - 1)
      objects_watched[i] = objects_watched[objects_watched_num - 1];
    objects_watched_num--;
  }
}

void object_mark_vanished(tObjectIdentifier id)
{ tObjectIndex i = object_lookup(id);
  if (i != OBJECT_INDEX_INVALID) objects_watched[i].flags |= ofVanished;
  /* "else": nothing is watching the object, so the information is irrelevant;
     this is by far the most likely case */
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "object_mark_vanished(%d): %d\n",id,i);
  debugmsg(debugstrbuf);
#endif
}

tBoolean object_did_vanish(tObjectIdentifier id)
{ tBoolean retval;
  tObjectIndex i = object_lookup(id);
  if (i == OBJECT_INDEX_INVALID) retval = truE; /* "should not happen" */
  else retval = cond2boolean(objects_watched[i].flags & ofVanished);
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "object_did_vanish(%d): %d,%d\n", id, i, retval);
  debugmsg(debugstrbuf);
#endif
  return(retval);
}


/* sprintf() replacement */

#if HAVE_STDARG_H
#include <stdarg.h>
#endif

void my_spf(char* staticbuf, size_t staticbufsize, char** spfbuf,
  const char* format, ...)
/* similar to sprintf(), but without the problem of buffer overflows; if the
   resulting string would not fit in the static buffer, a dynamic buffer is
   allocated. - Call either my_spf_cleanup() or my_spf_use() afterwards. */
/* The design of this function is based on the fact that the static buffer is
   "almost always" large enough, so we do the expensive dynamic allocation
   only in the few cases where it's absolutely necessary for correctness or
   where we'd want a dynamic buffer anyway. */
{ va_list vl;
  char numbuf[50], ch, *dest, *desttmp;
  const char* tmp;
  size_t destlen;
  tBoolean inside, is_long SHUT_UP_COMPILER(falsE);

  /* Find out how long the resulting string can become at most: */
  destlen = strlen(format) + 1;
  tmp = format; inside = falsE;
  va_start(vl, format);
  while ( (ch = *tmp++) != '\0' )
  { if (inside)
    { inside = falsE;
      if (ch == 's') /* found a "%s" specifier */
      { const char* const s = va_arg(vl, char*);
        if (s != NULL) destlen += strlen(s);
      }
      else if (ch == 'd') /* found a "%d" specifier */
      { if (is_long)
        { const long int num = va_arg(vl, long int);
          sprint_safe(numbuf, strPercld, num); /* IMPROVEME! */
        }
        else
        { const int num = va_arg(vl, int);
          sprint_safe(numbuf, strPercd, num); /* IMPROVEME! */
        }
        destlen += strlen(numbuf);
      }
      else if (ch == 'c') /* found a "%c" specifier */
      { (void) va_arg(vl, int); /* (only needed to proceed in the vl list) */
        destlen++;
      }
      else if (ch == 'p') /* found a "%p" specifier */
      { (void) va_arg(vl, void*); /* (only needed to proceed in the vl list) */
        destlen += (2 + 2 * sizeof(void*));
      }
      else if (ch == 'l') { inside = is_long = truE; destlen--; goto dontdec; }
      else goto dontdec;
      destlen -= 2; /* we'll strip the two characters "%s"/"%d"/... */
      dontdec: {}
    }
    else
    { if (ch == '%') { inside = truE; is_long = falsE; }
    }
  }
  va_end(vl);

  /* Build the string: */
  if ( (staticbuf != NULL) && (staticbufsize >= destlen) ) dest = staticbuf;
  else dest = __memory_allocate(destlen);
  desttmp = dest; tmp = format; inside = falsE;
  va_start(vl, format);
  while ( (ch = *tmp++) != '\0' )
  { if (inside)
    { inside = falsE;
      if (ch == 's') /* found a "%s" specifier */
      { const char* s = va_arg(vl, char*);
        char ch0;
        if (s != NULL) { while ( (ch0 = *s++) != '\0' ) *desttmp++ = ch0; }
      }
      else if (ch == 'd') /* found a "%d" specifier */
      { char ch0, *s = numbuf;
        if (is_long)
        { const long int num = va_arg(vl, long int);
          sprint_safe(numbuf, strPercld, num); /* IMPROVEME! */
        }
        else
        { const int num = va_arg(vl, int);
          sprint_safe(numbuf, strPercd, num); /* IMPROVEME! */
        }
        while ( (ch0 = *s++) != '\0' ) *desttmp++ = ch0;
      }
      else if (ch == 'c') /* found a "%c" specifier */
      { const char ch0 = (char) va_arg(vl, int);
        if (ch0 != '\0') *desttmp++ = ch0;
      }
      else if (ch == 'p') /* found a "%p" specifier */
      { void* p = va_arg(vl, void*);
        unsigned long int l = (unsigned long int) p;
        unsigned short count = 2 * sizeof(void*);
        char* t;
        *desttmp++ = '0'; *desttmp++ = 'x';
        t = desttmp = desttmp + count;
        while (count-- > 0)
        { *--t = strHexnum[l & 15];
          l >>= 4;
        }
        /* CHECKME: don't output leading zeroes? */
      }
      else if (ch == 'l') { inside = is_long = truE; }
      else goto append;
    }
    else
    { if (ch == '%') { inside = truE; is_long = falsE; }
      else { append: *desttmp++ = ch; }
    }
  }
  va_end(vl);
  *desttmp = '\0';
  *spfbuf = dest;
}


/* Replacements for broken/missing C library functions; we don't care about
   performance here - the only purpose of these functions is to make the
   program compile and work, nothing more. */

#if !HAVE_MMAP
void* nonbroken_mmap(size_t len, int fd)
{ if (len > 0)
  { void *retval = __memory_allocate(len), *p = retval;
    size_t lenleft = len;
    while (1)
    { ssize_t err = my_read(fd, p, lenleft);
      if (err <= 0) { memory_deallocate(retval); goto failed; }
      else if (err >= (ssize_t) lenleft) return(retval); /* done */
      lenleft -= err; p += err;
    }
  }
  failed:
  return(MAP_FAILED);
}
#endif

#if !HAVE_STRCASECMP
int nonbroken_strcasecmp(const char* _s1, const char* _s2)
{ const unsigned char *s1 = (const unsigned char*) _s1,
    *s2 = (const unsigned char*) _s2;
  while (1)
  { unsigned char _a = *s1++, a = my_tolower(_a),
      _b = *s2++, b = my_tolower(_b);
    if (a != b) return(a - b);
    else if (a == '\0') return(0);
  }
}
#endif

#if 0
#if !HAVE_STRNCASECMP
int nonbroken_strncasecmp(const char* _s1, const char* _s2, size_t n)
/* Nowadays we seem to use strneqcase() instead of my_strncasecmp() everywhere,
   so REMOVEME? */
{ const unsigned char *s1 = (const unsigned char*) _s1,
    *s2 = (const unsigned char*) _s2;
  while (n-- > 0)
  { unsigned char _a = *s1++, a = my_tolower(_a),
      _b = *s2++, b = my_tolower(_b);
    if (a != b) return(a - b);
    else if (a == '\0') return(0);
  }
  return(0);
}
#endif
#endif

#if !HAVE_STRCHR
char* nonbroken_strchr(const char* s, int _wanted_ch)
{ const char wanted_ch = (char) _wanted_ch;
  char ch;
  while ( (ch = *s) != '\0' )
  { if (ch == wanted_ch) return(unconstify(s));
    s++;
  }
  return(NULL);
}
#endif

#if !HAVE_STRRCHR
char* nonbroken_strrchr(const char* s, int _wanted_ch)
{ size_t len = strlen(s);
  if (len > 0)
  { const char wanted_ch = (char) _wanted_ch;
    s += len - 1;
    while (len-- > 0)
    { if (*s == wanted_ch) return(unconstify(s));
      s--;
    }
  }
  return(NULL);
}
#endif

#if !HAVE_STRSTR
char* nonbroken_strstr(const char* haystack, const char* needle)
{ const size_t hlen = strlen(haystack), nlen = strlen(needle);
  size_t h = 0, n = 0;
  do
  { if (haystack[h] == needle[n]) n++;
    else { h -= n; n = 0; }
    h++;
  } while ( (n < nlen) && (h < hlen) );
  if (n >= nlen) return(unconstify(haystack) + h - nlen - 1); /* found */
  else return(NULL);
}
#endif
