/* retawq/stuff.h - general stuff
   This file is part of retawq (<http://retawq.sourceforge.net/>), a network
   client created by Arne Thomassen; retawq is basically released under certain
   versions of the GNU General Public License and WITHOUT ANY WARRANTY.
   Read the file COPYING for license details, README for program information.
   Copyright (C) 2001-2004 Arne Thomassen <arne@arne-thomassen.de>
*/

#ifndef __retawq_stuff_h__
#define __retawq_stuff_h__

/* Compile-time configuration */

#include ".config"

#undef OPTION_TG /* make sure it won't be used anywhere accidentally... */

#if OPTION_BUILTIN_DNS
#undef OPTION_THREADING
#define OPTION_THREADING 0
#endif

#define CONFIG_ASYNC_DNS (OPTION_THREADING || OPTION_BUILTIN_DNS)
#define CONFIG_ANY_MAIL (OPTION_POP || OPTION_IMAP)
#define EXECEXT_ALL (1 /* | 2 | 4 | 8 */)


/* Text/graphics configuration I: numerical values for CONFIG_TG */

#define TG_INVALID 0
#define TG_BACKGROUND 1
#define TG_CONSOLE 2
#define TG_CURSES 3
#define TG_NCURSES 4
#define TG_XCURSES 5
#define TG_FLEXCURSES 6
#define TG_X 7
#define TG_GTK 8
#define TG_SVGA 9
#define TG_FRAMEBUFFER 10

/* text/graphics classes, for convenience */
#define TGC_IS_CURSES (CONFIG_TG == TG_CURSES) || (CONFIG_TG == TG_NCURSES) || (CONFIG_TG == TG_XCURSES) || (CONFIG_TG == TG_FLEXCURSES)
#define TGC_IS_GRAPHICS (CONFIG_TG == TG_X) || (CONFIG_TG == TG_GTK) || (CONFIG_TG == TG_SVGA) || (CONFIG_TG == TG_FRAMEBUFFER)

#if CONFIG_TG == TG_CONSOLE /* default runmode is console, so enable console */
#undef CONFIG_CONSOLE
#define CONFIG_CONSOLE 1
#endif


/* Standard header file includes */

#include <stdio.h>

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <stddef.h>
#else
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#endif

#if HAVE_STRING_H
#if (!STDC_HEADERS) && HAVE_MEMORY_H
#include <memory.h>
#endif
#include <string.h>
#endif

#if HAVE_STRINGS_H
#include <strings.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_ERRNO_H
#include <errno.h>
#endif

#if HAVE_LIMITS_H
#include <limits.h>
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#if HAVE_TERMIOS_H && CONFIG_CONSOLE
#include <termios.h>
#define MIGHT_DO_TERMIOS 1
#else
#define MIGHT_DO_TERMIOS 0
#endif

typedef unsigned char tUint8;
typedef unsigned int tUint32;


/* Strings */

#define STRBUF_SIZE (4096)
extern const char strCopyright[], strRetawq[], strTG[], strOptHelp[],
  strVersion[];
extern /*@observer@*/ const char strEmpty[], strNewline[];
extern const char strMinus[], strSlash[], strAsterisk[], strLt[], strGt[],
  strHm[], strQm[], strDoubleQuote[], strSpacedDash[], strDoubleDot[],
  str1dot0[], str1dot1[];
extern const char strPercsPercs[], strPercd[];
extern const char strHexnum[];
extern const char strA[], strShtml[], strReset[], strSelect[], strSubmit[],
  strText[], strClose[], strQuit[], strLocalhost[], strUnknown[],
  strBracedDisabled[], strImage[], strCheckbox[], strButton[], strYes[],
  strNo[], strFileUc[], strDirectoryUc[], strLink[], strLoginFailed[],
  strFwdact[];
#if !CONFIG_SESSIONS
extern const char strFsessions[];
#endif
#if !CONFIG_JUMPS
extern const char strFjumps[];
#endif
#if OPTION_EXECEXT != EXECEXT_ALL
extern const char strFexecext[];
#endif
#if CONFIG_CONSOLE
extern const char strConsoleDiscard[];
#endif
#if (CONFIG_TG == TG_X) || (CONFIG_TG == TG_XCURSES)
extern const char strCantOpenXws[];
#endif
#if OPTION_LOCAL_CGI || OPTION_TRAP || (OPTION_EXECEXT & 1)
extern const char strSoftwareId[];
#endif

#define strHelp  (strOptHelp + 2)    /* ug */
#define strSpace (strSpacedDash + 2) /* ly */
#define strPercs (strPercsPercs + 2) /* :-) */
#define strHtml  (strShtml + 1)

/* scheme strings: */
extern const char strHttp[], strFtp[], strLocal[], strFile[], strAbout[],
  strFinger[], strLocalCgi[], strNntp[], strNews[], strCvs[], strGopher[],
  strInfo[], strMailto[], strJavascript[], strHttps[], strFtps[], strPop[],
  strPops[], strExecextShell[];


/* Helper macros */

#define CONFIG_EXDEBUG 0

#ifndef NULL
#define NULL ((void*) 0)
#endif

#ifndef MIN
#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#endif

#ifdef INT_MAX
#define MY_ATOI_INT_MAX (INT_MAX / 11)
  /* (avoids overflows in my_atoi() and is still big enough for any "real-life"
      case) */
#else
#define MY_ATOI_INT_MAX (99999999)
#endif

/* compare two numbers; in Perl we'd use the "<=>" operator... */
#define my_numcmp(num1, num2) \
  ( ((num1) > (num2)) ? (+1) : ( ((num1) < (num2)) ? (-1) : (0) ) )

#ifdef S_SPLINT_S

/* make splint happy; __sunused and __sallocator are splint-specific variants
   of __cunused and __callocator; this rubbish is necessary because splint and
   gcc expect such annotations in different places within a declaration... */

#define __sunused /*@unused@*/
#define __dnr /*@noreturn@*/
#define __sallocator /*@only@*/
#define my_enum1 typedef
#define my_enum2(ctype)

#else

/* make _me_ happy :-) */

#define __sunused
#define __dnr
#define __sallocator
#define my_enum1
#define my_enum2(ctype) ; typedef ctype

#endif

#if USING_CONFIGURE
#define __my_inline inline
#elif defined(__GNUC__)
#define __my_inline inline
#else
#define __my_inline /* nothing (we don't know for _sure_ how to do it) */
#endif

#ifdef DO_NOT_INLINE
#define my_inline /* nothing (to make the program smaller) */
#else
#define my_inline __my_inline
#endif

#if defined(__GNUC__) && ( (__GNUC__ > 2) || ( (__GNUC__ == 2) && (__GNUC_MINOR__ >= 5) ) )
#define __cunused __attribute__((__unused__))
#define my_spf_attr __attribute__((__format__(printf, 4, 5))) /* shows bugs */
#define does_not_return __attribute__((__noreturn__))
#if CONFIG_EXDEBUG
#define __init __attribute__((__section__ ("textinit")))
#endif
/* #define __callocator __attribute__((__malloc__)) */
#else
#define __cunused
#define my_spf_attr
#endif

#ifndef __init
#define __init
#endif
#ifndef __callocator
#define __callocator
#endif
#ifndef does_not_return
#define does_not_return __dnr
#endif
#define const_after_init
#if CONFIG_DEBUG
#define one_caller /* nothing (gcc/gdb and line numbers...) */
#else
#define one_caller __my_inline /* (function is called from only one place) */
#endif

/* Mark that we checked that the library function won't overflow the buffer: */
#define sprint_safe sprintf

/* Some snprintf() implementations in C libraries are rubbish. Just one example
   from a Linux manual page: "Linux libc4.[45] does not have a snprintf, but
   provides a libbsd that contains an snprintf equivalent to sprintf, i.e.,
   one that ignores the size argument. Thus, the use of snprintf with early
   libc4 leads to serious security problems." Congratulations... */
#define snprint_safe DOES_NOT_EXIST__DO_NOT_USE_SNPRINTF /* clear? :-) */

/* Some sscanf() implementations in C libraries are rubbish: they try to write
   into the <format> parameter, causing a crash if operating on read-only
   string constants. (Information found in "info gcc".) Thus: */
#if 1
#define PREPARE_SSCANF_FORMAT(bufvar, bufsize, formatstr) \
  char bufvar[bufsize]; strcpy(bufvar, formatstr); /* for buggy libraries */
#else
#define PREPARE_SSCANF_FORMAT(bufvar, bufsize, formatstr) \
  static const char bufvar[] = formatstr; /* for non-buggy libraries */
#endif

#define zero2one(x) ( ((x) == 0) ? (1) : (x) )
#define null2empty(str) ( (str == NULL) ? (strEmpty) : (str) )
#define __unconstify(_t, _v) ( (_t) (_v) ) /* discard "const", avoid warning */
#define unconstify(str) __unconstify(char*, str)

#define is_control_char(unsigned_char) \
  ( ((unsigned_char) < ' ') || ((unsigned_char) == 127) )
#define TO_EOS(str) do { while (*str != '\0') str++; } while (0)
  /* go to end of string */

#define ARRAY_ELEMNUM(arr) (sizeof(arr) / sizeof(arr[0]))

/* We need library function variants which are _guaranteed_ to be
   locale-independent, even if i18n is enabled. */
#define my_isdigit(ch) ( ((ch) >= '0') && ((ch) <= '9') )
#define my_islower(ch) ( ((ch) >= 'a') && ((ch) <= 'z') )
#define my_isupper(ch) ( ((ch) >= 'A') && ((ch) <= 'Z') )
#define my_isalpha(ch) (my_islower(ch) || my_isupper(ch))
#define my_isalnum(ch) (my_isalpha(ch) || my_isdigit(ch))
#define my_tolower(ch) ( my_isupper(ch) ? ((ch) + ('a' - 'A')) : (ch) )
#define my_toupper(ch) ( my_islower(ch) ? ((ch) - ('a' - 'A')) : (ch) )


/* Boolean */

/* It seems that "false", "False" and "FALSE" are already used by compilers and
   libraries; it can't be true... :-) Let's hope that our choice is safe. */
/*@-booltype tBoolean@*/ /*@-booltrue truE@*/ /*@-boolfalse falsE@*/
my_enum1 enum { falsE = 0, truE = 1 } my_enum2(unsigned char) tBoolean;
#define boolean2bool(x) ( ((x) == falsE) ? (0) : (1) )
#define cond2boolean(condition) ( (condition) ? (truE) : (falsE) )
#define cond2bool(condition) ( (condition) ? (1) : (0) )


/* Text/graphics configuration II */

#include "tgmode.inc"

#if MIGHT_DO_TERMIOS
extern tBoolean do_restore_termios;
extern struct termios saved_termios;
#endif

#if TGC_IS_GRAPHICS

/* We are Pixels of Borg - curses is irrelevant. */
#undef OPTION_CED
#define OPTION_CED 1
#undef OPTION_MENUS
#define OPTION_MENUS 2

#define TGSTR N_("graphics")

#if CONFIG_TG == TG_X
#include <X11/Xlib.h>
extern Display* xws_display;
typedef unsigned int tKey;
typedef int tCoordinate;
#else
#include <gtk/gtk.h>
typedef guint tKey; /*struct _GdkEventKey->keyval,/usr/include/gdk/gdktypes.h*/
typedef gint16 tCoordinate; /* struct _GdkPoint, gtk/gdk/gdktypes.h */
#endif

#else /* #if TGC_IS_GRAPHICS */

#define TGSTR N_("text")

typedef int tKey;
typedef signed short tCoordinate;

#endif /* #if TGC_IS_GRAPHICS */

#if (OPTION_TEXTMODEMOUSE) && ( ( (TGC_IS_CURSES) && (defined(NCURSES_MOUSE_VERSION)) ) || (CONFIG_TG == TG_XCURSES) )
/* "user wants" && "library can" */
#define CONFIG_DO_TEXTMODEMOUSE 1
#define TEXTMODEMOUSE_MASK (BUTTON1_CLICKED | BUTTON2_CLICKED |BUTTON3_CLICKED)
#else
#define CONFIG_DO_TEXTMODEMOUSE 0
#endif

extern tBoolean need_tglib_cleanup;
extern const_after_init tBoolean lfdmbs;


/* i18n configuration; this must be done _after_ OPTION_CED (re-)definition */

#if (OPTION_I18N) && (OPTION_CED == 0)
/* We need special code to make translated texts seven-bit clean: */

#define NEED_SPECIAL_GETTEXT 1
extern char* my_do_gettext(char** buffer, const char* str);
#define declare_local_i18n_buffer static char* local_i18n_buffer = NULL;
#define my_gettext(str) my_do_gettext(&local_i18n_buffer, str)
#define i18n_strdup(str) my_strdup(str) /* orig destroyed by i18n_cleanup */
#define i18n_cleanup \
  if (local_i18n_buffer != NULL) (void) my_do_gettext(&local_i18n_buffer,NULL);

#else

#define NEED_SPECIAL_GETTEXT 0
#define declare_local_i18n_buffer /* nothing */
#define my_gettext(str) gettext(str)
#define i18n_strdup(str) (str) /* no need to copy */
#define i18n_cleanup

#endif

#if OPTION_I18N

#if HAVE_LIBINTL_H
#include <libintl.h>
#endif

#undef _
#ifndef _
#define _(str) my_gettext(str)
#endif

#ifndef N_
#ifdef gettext_noop
#define N_(str) gettext_noop(str)
#else
#define N_(str) (str)
#endif
#endif
#define unconstify_or_(str) _(str)

#else /* #if OPTION_I18N */

#define _(str) (str)
#define N_(str) (str)
#define unconstify_or_(str) unconstify(str)

#endif /* #if OPTION_I18N */

#define localized_size(size) (size) /* IMPLEMENTME! */


/* Curses */

#if CONFIG_TG != TG_GTK
extern tBoolean use_colors;
#endif

#if TGC_IS_CURSES

my_enum1 enum
{ ccBlack = 0, ccRed = 1, ccGreen = 2, ccYellow = 3, ccBlue = 4, ccMagenta = 5,
  ccCyan = 6, ccWhite = 7
} my_enum2(unsigned char) tColorCode;

my_enum1 enum
{ /* cpnHardwired = 0, -- not usable for us */ cpnDefault = 1, cpnRed = 2,
  cpnGreen = 3, cpnYellow = 4, cpnBlue = 5
} my_enum2(short) tColorPairNumber;
#define cpnMax (5)

typedef chtype tColorBitmask;
extern tColorBitmask color_bitmask[cpnMax + 1];

#define my_color_attr(cpn) (color_bitmask[cpn]) /*only do this if use_colors!*/
#define my_set_color(cpn) \
  do { if (use_colors) { (void) attron(my_color_attr(cpn)); } } while (0)

#endif


/* Objects */

typedef size_t tObjectIdentifier;
#define oiUnknown (0)


/* HTML forms; active elements */

typedef unsigned int tHtmlInputLength;
#define MAX_HTML_INPUT_LENGTH (STRBUF_SIZE / 2)

enum
{ aekUnknown = 0, aekLink = 1, aekFormText = 2, aekFormPassword = 3,
  aekFormCheckbox = 4, aekFormRadio = 5, aekFormSubmit = 6, aekFormReset = 7,
  aekFormFile = 8, aekFormTextarea = 9, aekFormSelect = 10, aekFormButton = 11,
  aekFormImage = 12, aekFormHidden = 13
};
#define MAX_AEK (13) /* associated UI strings in main.c */
typedef unsigned char tActiveElementKind;
#define is_form_aek(kind) \
  ( ((kind) >= aekFormText) && ((kind) <= aekFormHidden) )
#define has_input_length(kind) \
  ( ((kind) == aekFormText) || ((kind) == aekFormPassword) )

enum
{ aefNone = 0, aefCheckedSelected = 0x01, aefDisabled = 0x02,
  aefReadonly = 0x04, aefMultiple = 0x08, aefButtonTag = 0x10
};
typedef unsigned char tActiveElementFlags;
#define aefChangeable (aefCheckedSelected)

typedef signed int tActiveElementNumber;
#define INVALID_AE (-1)

typedef signed int tHtmlFormNumber;
#define INVALID_HTML_FORM_NUMBER (-1)

#if CONFIG_JAVASCRIPT
struct tJavascriptEventHandler;
#endif

typedef struct
{ const char *data, /* an attribute value - href for links, name otherwise */
    *render; /* e.g. text to be rendered for this active element */
#if CONFIG_JAVASCRIPT
  const struct tJavascriptEventHandler* eh;
#endif
  tHtmlInputLength size, maxlength;
  tActiveElementKind kind;
  tActiveElementFlags flags;
} tActiveElementBase;
/* tActiveElementBase contains "constant" information which only depends on the
   parsed resource (HTML source code), e.g. the target of a link or the maximum
   length of a text-input element, while tActiveElement (in main.c) contains
   view-dependent information like layout coordinates and current contents of
   text-input elements. */
/* For aekFormSelect kind, <render> points to the list of options, and
   <maxlength> is the number of options. */

enum { hffNone = 0, hffMethodPost = 0x01, hffEncodingMultipart = 0x02 };
typedef unsigned char tHtmlFormFlags;

typedef struct
{ const char* action_uri;
  tActiveElementNumber first_ae, last_ae;
  tHtmlFormFlags flags;
} tHtmlForm;


/* To have or not to have, that is the question! :-) */

#if HAVE_MEMCPY
#define my_memcpy(dest, src, cnt) (void) memcpy(dest, src, cnt)
#elif HAVE_BCOPY
#define my_memcpy(dest, src, cnt) bcopy(src, dest, cnt)
#else
#define my_memcpy(dest, src, cnt) \
  do \
  { char* _d = (char*) (dest); \
    const char* _s = (const char*) (src); \
    size_t _c = (cnt); \
    while (_c-- > 0) *_d++ = *_s++; \
  } while (0)
#endif

#if HAVE_MEMSET
#define my_memset(start, value, length) memset(start, value, length)
#else
#define my_memset(start, value, length) \
  do \
  { char* _s = (char*) (start); \
    const char _v = (char) (value); \
    size_t count = (length); \
    while (count-- > 0) *_s++ = _v; \
  } while (0)
#endif

#if (!HAVE_MEMSET) && (HAVE_BZERO)
  /* Use bzero() because it might be much faster than our stupid byte-by-byte
     cleaning fallback; but use it only in this case -- it's marked "legacy",
     SUSv3 says: "The memset() function is preferred over this function." */
#define my_memclr(start, length) bzero(start, length)
#else
#define my_memclr(start, length) my_memset(start, 0, length)
#endif

#define my_memclr_var(var) my_memclr(&(var), sizeof(var))
#define my_memclr_arr(arr) my_memclr((arr), sizeof(arr))

#if HAVE_MMAP

#include <sys/mman.h>
#define my_mmap(len, fd) mmap(NULL, len, PROT_READ, MAP_PRIVATE, fd, 0)
#define my_munmap(start, len) (void) munmap(start, len)

#else /* #if HAVE_MMAP */

#ifndef MAP_FAILED
#define MAP_FAILED ((void*) (-1))
#endif

extern void* nonbroken_mmap(size_t, int);
#define my_mmap(len, fd) nonbroken_mmap(len, fd)
#define my_munmap(start, len) memory_deallocate(start)

#endif /* #if HAVE_MMAP */

#if defined(MADV_SEQUENTIAL)
#define my_madvise_sequential(ptr, size) \
  (void) madvise(ptr, size, MADV_SEQUENTIAL)
#elif defined(POSIX_MADV_SEQUENTIAL)
#define my_madvise_sequential(ptr, size) \
  (void) posix_madvise(ptr, size, POSIX_MADV_SEQUENTIAL)
#else
#define my_madvise_sequential(ptr, size) do { } while (0)
#endif

#if HAVE_STRCASECMP
#define my_strcasecmp(s1, s2) strcasecmp(s1, s2)
#else
extern int nonbroken_strcasecmp(const char*, const char*);
#define my_strcasecmp(s1, s2) nonbroken_strcasecmp(s1, s2)
#endif

#if 0
#if HAVE_STRNCASECMP
#define my_strncasecmp(s1, s2, n) strncasecmp(s1, s2, n)
#else
extern int nonbroken_strncasecmp(const char*, const char*, size_t);
#define my_strncasecmp(s1, s2, n) nonbroken_strncasecmp(s1, s2, n)
#endif
#endif

#if HAVE_STRCHR
#define my_strchr(str, ch) strchr(str, ch)
#else
extern char* nonbroken_strchr(const char*, int);
#define my_strchr(str, ch) nonbroken_strchr(str, ch)
#endif

#if HAVE_STRERROR
#define my_strerror(errnum) strerror(errnum) /* CHECKME: use strerror_r()? */
#define __my_strerror(errnum) my_strerror(errnum)
#else
#define my_strerror(errnum) _(strUnknown) /* CHECKME! */
#define __my_strerror(errnum) strUnknown /* CHECKME! */
#endif

#if HAVE_STRRCHR
#define my_strrchr(str, ch) strrchr(str, ch)
#else
extern char* nonbroken_strrchr(const char*, int);
#define my_strrchr(str, ch) nonbroken_strrchr(str, ch)
#endif

#if HAVE_STRSTR
#define my_strstr(haystack, needle) strstr(haystack, needle)
#else
extern char* nonbroken_strstr(const char*, const char*);
#define my_strstr(haystack, needle) nonbroken_strstr(haystack, needle)
#endif


/* Other stuff */

#if OPTION_LOCAL_CGI || OPTION_EXECEXT
#define MIGHT_FORK_EXEC 1
#define __initfe /* function not only used at initialization time */
#else
#define MIGHT_FORK_EXEC 0
#define __initfe __init /* function only used at initialization time */
#endif

#if MIGHT_FORK_EXEC && defined(FD_CLOEXEC)
#define make_fd_cloexec(fd) \
  do \
  { int flags = fcntl(fd, F_GETFD, 0); \
    if (flags != -1) (void) fcntl(fd, F_SETFD, flags | FD_CLOEXEC); \
  } while (0)
#else
#define make_fd_cloexec(fd) do { } while (0)
#endif

#define TLS_GNUTLS  (1)
#define TLS_OPENSSL (2)
#define TLS_BUILTIN (3)

#if OPTION_TLS
#define strTls ", tls"
#else
#define strTls ""
#endif

#if OPTION_TLS == TLS_GNUTLS
#define strTLS "TLS (GnuTLS)"
#elif OPTION_TLS == TLS_OPENSSL
#define strTLS "TLS (OpenSSL)"
#elif OPTION_TLS
#define strTLS "TLS"
#endif

my_enum1 enum
{ pmEnvironed = 0, pmDump = 1, pmDownload = 2, pmBatch = 3
#if CONFIG_CONSOLE
  , pmConsole = 4
#endif
} my_enum2(unsigned char) tProgramMode;
extern const_after_init tProgramMode program_mode;
#define is_environed (program_mode == pmEnvironed) /* using curses/GTK/... */

#if CONFIG_CONSOLE
#define is_promptable ( (is_environed) || (program_mode == pmConsole) )
#else
#define is_promptable (is_environed)
#endif

extern /*@null@*/ const char* initial_messages;
extern unsigned char launch_uri_count;
extern const char* launch_uri[2];

typedef unsigned char tRgbPart;

/* A fast, flexible binary search algorithm; the standard library function
   bsearch() only returns a useless _member_ of the array, while we want to get
   an array _index_ by which we can access key, value, member, neighbor member
   or whatever else we just need. */
typedef signed short tMbsIndex;
#define INVALID_INDEX ((tMbsIndex) (-1))
#define my_binary_search(_min_index, _max_index, comparison, action) \
  { tMbsIndex min = _min_index, max = _max_index, idx; \
    do \
    { int res; \
      idx = (min + max) / 2; \
      res = comparison; /* depends on <idx> */ \
      if (res < 0) max = idx - 1; \
      else if (res > 0) min = idx + 1; \
      else { perform_action: action; } \
    } while (min <= max); \
    idx = INVALID_INDEX; goto perform_action; /* not found */ \
  }

/* Simple bit manipulation helpers; <address> is the start address of a
   bitfield, <bit> is the number of the desired bit within the field. */
#define __my_byte(address, bit) (*((unsigned char*) ((address) + ((bit) / 8))))
#define __my_bit(bit) (1 << ((bit) % 8))
#define my_bit_set(address, bit) __my_byte(address, bit) |= __my_bit(bit)
#define my_bit_clear(address, bit) __my_byte(address, bit) &= ~__my_bit(bit)
#define my_bit_flip(address, bit) __my_byte(address, bit) ^= __my_bit(bit)
#define my_bit_test(address, bit) (__my_byte(address, bit) & __my_bit(bit))

#define list_reverse(_head, _type) \
  do \
  { _type *head = NULL, *temp = _head; \
    while (temp != NULL) \
    { _type* next = temp->next; \
      temp->next = head; \
      head = temp; \
      temp = next; \
    } \
    _head = head; \
  } while (0)

/* CHECKME: use a sizeof() test with configure to avoid (_bogus_!) compiler
   warnings on 64-bit machines! */
#define MY_INT_TO_POINTER(x) ( (void*) ((int) (x)) )
#define MY_POINTER_TO_INT(x) ((int) (x))

/* Suppress _bogus_ compiler warnings about uninitialized variables (in order
   to make all "real" warnings more visible): */
#if 1
#define SHUT_UP_COMPILER(value) = (value)
#else
#define SHUT_UP_COMPILER(value) /* nothing */
#endif

#if CONFIG_DEBUG
extern const_after_init int debugfd;
#define __debugmsg(msg, len) (void) my_write(debugfd, msg, len)
#define debugmsg(_dbgmsg) \
  do \
  { const char* dbgmsg = (_dbgmsg); /* (evaluate it only once) */ \
    __debugmsg(dbgmsg, strlen(dbgmsg)); \
  } while (0)
#else
#define __debugmsg(msg, len) do { } while (0)
#define debugmsg(msg) do { } while (0)
#endif

#define __exdebugmsg(msg, len) do { } while (0)
#define exdebugmsg(msg) do { } while (0)

#define __my_readwrite(op, fd, buf, count) \
  ssize_t retval; \
  unsigned char loopcount = 0; \
  do \
  { retval = op(fd, buf, count); \
  } while ( (retval == -1) && (errno == EINTR) && (++loopcount < 100) ); \
  return(retval);
#define my_write_str(fd, str) \
  do { const char* _x = (str); (void) my_write(fd, _x, strlen(_x)); } while (0)
#define my_write_crucial(fd, buf, __count) \
  do \
  { size_t _count = __count; /* (evaluate it only once) */ \
    ssize_t _err = my_write(fd, buf, _count); \
    if (_err != (ssize_t) _count) fatal_error_writefailed(_err); \
  } while (0)

#if CONFIG_DEBUG
#define __debug_my_close(fd) \
  do \
  { char __buf[50]; sprint_safe(__buf, "my_close(%d)\n", fd); debugmsg(__buf);\
  } while (0)
#else
#define __debug_my_close(fd) do { } while (0)
#endif

/* Each .c file shall get its own, private copy of the following functions.
   Since they're small and inlined, this shouldn't hurt; but anyway, let's
   thank the inventors of that stupid signal system - why couldn't they stick
   to files and processes... */
#define __define_filestuff \
static my_inline __sunused ssize_t my_read(int fd, /*@out@*/ void* buf, \
  size_t count) \
{ __my_readwrite(read, fd, buf, count) \
} \
static my_inline __sunused ssize_t __my_write(int fd, const void* buf, \
  size_t count) \
/* may write less than <count> bytes */ \
{ __my_readwrite(write, fd, buf, count) \
} \
static my_inline __sunused int my_create(const char* path, int flags, \
  mode_t mode) \
{ int retval; \
  unsigned char loopcount = 0; \
  do \
  { retval = open(path, flags | O_NOCTTY, mode); \
  } while ( (retval == -1) && (errno == EINTR) && (++loopcount < 100) ); \
  return(retval); \
} \
static my_inline __sunused int my_open(const char* path, int flags) \
{ int retval; \
  unsigned char loopcount = 0; \
  do \
  { retval = open(path, flags | O_NOCTTY); \
  } while ( (retval == -1) && (errno == EINTR) && (++loopcount < 100) ); \
  return(retval); \
} \
static my_inline __sunused void my_close(int fd) \
{ int err; \
  unsigned short loopcount = 0; \
  __debug_my_close(fd); \
  do \
  { err = close(fd); \
  } while ( (err == -1) && (errno == EINTR) && (++loopcount < 10000) ); \
  /* return(err); -- would be quite pointless */ \
} \
static my_inline __sunused int my_stat(const char* filename, \
  /*@out@*/ struct stat* sb) \
{ int retval; \
  unsigned char loopcount = 0; \
  do \
  { retval = stat(filename, sb); \
  } while ( (retval == -1) && (errno == EINTR) && (++loopcount < 100) ); \
  return(retval); \
} \
static my_inline __sunused int my_fstat(int fd, /*@out@*/ struct stat* sb) \
{ int retval; \
  unsigned char loopcount = 0; \
  do \
  { retval = fstat(fd, sb); \
  } while ( (retval == -1) && (errno == EINTR) && (++loopcount < 100) ); \
  return(retval); \
}

typedef unsigned short tPortnumber;
/* (The correct formal type would be in_port_t, but this seems to be
    not-so-portable; e.g. glibc before 2000-04-01 doesn't have it.) */

#define __dealloc(ptr) do { if (ptr != NULL) memory_deallocate(ptr); } while(0)
#define dealloc(ptr) \
  do { if (ptr != NULL) { memory_deallocate(ptr); ptr = NULL; } } while (0)
#define my_strdedup(dest, src) /* deallocate and duplicate */ \
  do { __dealloc(dest); (dest) = my_strdup(src); } while (0)


/* Run-time Configuration */

typedef struct tConfigProxy
{ struct tConfigProxy* next;
  char *hosts_pattern, *proxy_hostname, *auth_user, *auth_pass;
  tPortnumber hosts_portnumber, proxy_portnumber; /* (in network byte order) */
} tConfigProxy;

typedef struct tConfigProtocolVersion
{ struct tConfigProtocolVersion* next;
  char *hosts_pattern, *protstr;
  tPortnumber hosts_portnumber;
} tConfigProtocolVersion;

#if OPTION_COOKIES
enum { ccfNone = 0, ccfAllowed = 0x01 };
typedef unsigned char tConfigCookieFlags;

typedef struct tConfigCookie
{ struct tConfigCookie* next;
  const char* hosts_pattern;
  tConfigCookieFlags flags;
} tConfigCookie;
#endif /* #if OPTION_COOKIES */

#if OPTION_LOCAL_CGI
enum { clcfNone = 0, clcfAllowed = 0x01 };
typedef unsigned char tConfigLocalCgiFlags;

typedef struct tConfigLocalCgi
{ struct tConfigLocalCgi* next;
  const char* path_pattern;
  tConfigLocalCgiFlags flags;
} tConfigLocalCgi;
#endif /* #if OPTION_LOCAL_CGI */

typedef struct tConfigLogin
{ struct tConfigLogin* next;
  char *user, *password, *hosts_pattern;
  tPortnumber hosts_portnumber; /* (in network byte order) */
} tConfigLogin;

#if CONFIG_JUMPS
typedef struct tConfigJump
{ struct tConfigJump* next;
  const char* name; /* "name" is a shortcut for "shortcut" :-) */
  char *uri, **arg;
  unsigned short argcount, maxargcount;
} tConfigJump;
#endif

#if CONFIG_LOCALDIR > 1
typedef struct tConfigLocaldirsort
{ struct tConfigLocaldirsort* next;
  const char* path_pattern;
  const_after_init char* sorting;
} tConfigLocaldirsort;
typedef struct tConfigLocaldirformat
{ struct tConfigLocaldirformat* next;
  const char *path_pattern, *format;
} tConfigLocaldirformat;
#endif /* #if CONFIG_LOCALDIR > 1 */

#if CONFIG_FTP && OPTION_TLS
my_enum1 enum
{ ftmAutodetect = 0, ftmAuthTls = 1, ftmAuthTlsDataclear = 2, ftmAuthSsl = 3,
  ftmTls = 4
} my_enum2(unsigned char) tFtpTlsMethod;
typedef struct tConfigFtpTlsMethod
{ struct tConfigFtpTlsMethod* next;
  char* hosts_pattern;
  tPortnumber hosts_portnumber; /* (in network byte order) */
  tFtpTlsMethod ftm;
} tConfigFtpTlsMethod;
#endif

#if OPTION_EXECEXT
typedef struct tExecextParam
{ struct tExecextParam* next;
  char* param;
} tExecextParam;
#endif

enum
{ cfNone = 0, cfColorsOff = 0x01, cfColorsReverse = 0x02,
  cfUserDefinedConfigPath = 0x04,
  cfDontConfirmQuit = 0x10, cfDontConfirmClose = 0x20,
  cfDontConfirmOverwrite = 0x40, cfDontConfirmSubmit = 0x80,
  cfDontConfirmReset = 0x100, cfDontConfirmRepost = 0x200,
  cfDontConfirmHtml = 0x400, cfDontConfirmEnable = 0x800,
  cfUseDiskCache = 0x1000, cfMayWriteDiskCache = 0x2000
#if OPTION_EXECEXT & 1
  , cfExecextShellCustom = 0x4000
#endif
};
typedef unsigned short tConfigurationFlags;

typedef struct
{ const char* path; /* path to the directory which contains config file etc. */
  const char* cachepath; /* path to the disk cache directory */
  const char* user_agent; /* for HTTP header "User-Agent" */
  const char* languages; /* for HTTP header "Accept-Language" */
  const char *home_uri, *search_engine, *bookmarks; /* special URIs */
  const char* pm_uri; /* for pmDump/pmDownload */
  const_after_init tConfigProtocolVersion* http_version;
  const_after_init tConfigProxy* http_proxies;
#if OPTION_TLS
  const_after_init tConfigProxy* https_proxies;
#endif
#if CONFIG_FTP
  const_after_init tConfigLogin* ftp_login;
#if OPTION_TLS
  const_after_init tConfigFtpTlsMethod* ftp_tls_method;
#endif
#endif
#if CONFIG_JUMPS
  const_after_init tConfigJump* jumps;
#endif
#if CONFIG_LOCALDIR > 1
  const_after_init tConfigLocaldirsort* lds;
  const_after_init tConfigLocaldirformat* ldf;
#endif
#if OPTION_COOKIES
  const_after_init tConfigCookie* http_cookies;
#endif
#if OPTION_TLS && OPTION_COOKIES
  const_after_init tConfigCookie* https_cookies;
#endif
#if OPTION_LOCAL_CGI
  const_after_init tConfigLocalCgi* local_cgi;
#endif
#if OPTION_EXECEXT & 1
  const_after_init tExecextParam* execext_shell;
#endif
#if CONFIG_SESSIONS
  const char *session_default, *session_resume;
#endif
#if OPTION_NEWS
  const char* news_server;
#endif
  const char* batchfile;
  size_t ramcachesize, diskcachesize;
#if TGC_IS_GRAPHICS
  tCoordinate width, height; /* default window-contents width/height */
#endif
  tConfigurationFlags flags;
#if TGC_IS_CURSES
  char char_yes, char_no; /* (mostly for i18n) */
#endif
#if CONFIG_CONSOLE
  char console_backspace;
#endif
  char char_file, char_dir, char_link; /* (mostly for i18n) */
  unsigned char redirections; /* maximum number of automatic redirections */
} tConfiguration;

extern tConfiguration config;


/* Functions */

extern int streqcase3(const char*, const char*);
#define streqcase(str1, str2) (streqcase3(str1, str2) == 0)
extern int strneqcase3(const char*, const char*, size_t);
#define strneqcase(str1, str2, maxlen) (strneqcase3(str1, str2, maxlen) == 0)
extern void my_atoi(const char*, /*@reldef@*/ int*, /*@reldef@*/ const char**,
  int);
extern char* my_strnchr(const char*, int, int);
extern char* my_strncasestr(const char*, const char*, const size_t);
extern tBoolean is_suffix(const char* /*string*/, const char* /*suffix*/);
extern tBoolean my_pattern_matcher(const char*, const char*);
extern void does_not_return do_quit(void);
extern void does_not_return do_quit_sig(void);
extern void does_not_return fatal_error(int, const char*);
extern void does_not_return fatal_error_writefailed(int);
extern __sallocator /*@notnull@*/ /*@out@*/ void* __callocator
  __memory_allocate(size_t);
extern __sallocator /*@notnull@*/ void* __callocator memory_allocate(size_t);
extern __sallocator /*@notnull@*/ void* memory_reallocate(/*@only@*/ void*,
  size_t);
extern void memory_deallocate(/*@only@*/ /*@notnull@*/ const void*);
extern __sallocator char* __callocator my_strdup(const char*);
extern __sallocator char* __callocator my_strndup(const char*, size_t);
extern __sallocator char* __callocator my_strdup_tolower(const char*);
extern __sallocator char* __callocator my_strndup_tolower(const char*, size_t);
extern const char* get_homepath(void);
extern ssize_t my_write(int, const void*, size_t);
extern unsigned char my_mmap_file_readonly(const char*, /*@out@*/ void**,
  /*@out@*/ size_t*);
#if CONFIG_LOCALDIR > 1
extern void check_localdirsort(char*);
#endif

#define my_strcpy_conv(__dest, __src, __func) \
  do \
  { const char* _src = (__src); char ch, *_dest = (__dest); /* eval once */ \
    do { ch = *_src++; *_dest++ = (__func); } while (ch != '\0'); \
  } while (0)
#define my_strcpy_tolower(_dest, _src) \
  my_strcpy_conv(_dest, _src, my_tolower(ch))
#define my_strcpy_toupper(_dest, _src) \
  my_strcpy_conv(_dest, _src, my_toupper(ch))

extern tObjectIdentifier object_id_create(void);
extern void object_watch(tObjectIdentifier);
extern void object_unwatch(tObjectIdentifier);
extern void object_mark_vanished(tObjectIdentifier);
extern tBoolean object_did_vanish(tObjectIdentifier);

extern void my_spf_attr my_spf(char*, size_t, /*@out@*/ char**,
  const char*, ...);
#define my_spf_cleanup(staticbuf, spfbuf) \
  do { if (staticbuf != spfbuf) memory_deallocate(spfbuf); } while (0)
#define my_spf_use(spfbuf) (spfbuf) /* (just to clarify the code) */

#if MIGHT_FORK_EXEC
extern void reset_signals(void); /* see init.c */
#endif

#if CONFIG_KEYMAPS /* see main.c */
extern unsigned char keymap_command_key_register(const char*, const char*);
extern unsigned char keymap_lineinput_key_register(const char*, const char*);
#endif

extern void resource_quit(void); /* see resource.c */

#endif /* #ifndef __retawq_stuff_h__ */
