/* retawq/main.c - main routine, user interaction
   This file is part of retawq (<http://retawq.sourceforge.net/>), a network
   client created by Arne Thomassen; retawq is basically released under certain
   versions of the GNU General Public License and WITHOUT ANY WARRANTY.
   Read the file COPYING for license details, README for program information.
   Copyright (C) 2001-2004 Arne Thomassen <arne@arne-thomassen.de>
*/

#include "stuff.h"
#include "init.h"
#include "resource.h"
#include "parser.h"

#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#if CONFIG_TG == TG_X
#include <X11/keysym.h>
#elif CONFIG_TG == TG_GTK
#include <gdk/gdkkeysyms.h>
#endif


/* Strings */

static char strbuf[STRBUF_SIZE], strbuf2[STRBUF_SIZE], strbuf3[STRBUF_SIZE];

static const char strOn[] = "on", strPercsDash[] = "%s - ",
  strClosingBrace[] = ")", strDotSlash[] = "./", strSlashSlash[] = "//";
#define strDot (strDoubleDot + 1) /* ugly :-) */

static const char strBracedNewWindow[] = N_("(new window)"),
  strBack[] = N_("Back"), strForward[] = N_("Forward"),
  strReload[] = N_("Reload"), strEnforcedReload[] = N_("Enforced Reload"),
  strSearch[] = N_("Search"),
  strSource[] = N_("source"), strSourceCode[] = N_("Source Code"),
  strOpenInNewWindow[] = N_("Open in New Window"), strSave[] = N_("Save"),
  strOpenLinkInNewWindow[] = N_("Open Link in New Window"),
  strAddBookmark[] = N_("Add Bookmark"), strSaveAs[] = N_("Save As..."),
  strSelectionEmpty[] = N_("This selection list is empty!"),
  strContext[] = N_("Context"), strShowElementInfo[] = N_("Show Element Info"),
  strEnableThisElement[] = N_("Enable this Element"),
  strSubmitThisForm[] = N_("Submit this Form");
#if OPTION_COOKIES
static const char strCookies[] = N_("(cookies) ");
#endif

static const char* const strAek[MAX_AEK + 1] =
{ strUnknown, strLink, N_("text field"), N_("password field"),
  N_(strCheckbox), N_("radio button"), N_("submit button"), N_("reset button"),
  N_("file field"), N_("text-area field"), N_("selection"), N_(strButton),
  N_(strImage), N_("hidden field")
};

#if CONFIG_JUMPS
static /*@null@*/ const char* previous_jump = NULL;
#endif
static /*@null@*/ const char* search_string = NULL;

#if TGC_IS_GRAPHICS

static const char strCancel[] = N_("Cancel"), strDownload[] = N_("Download");
static const_after_init char *istrYesUc, *istrNoUc;

#if CONFIG_TG == TG_X

/* some minor, not-yet-complete abstraction for the future... */
typedef void tGraphicsWidget;
typedef void tGraphicsWindow;
typedef void tGraphicsLowlevelWindow;
typedef GC tGraphicsContext;
typedef XEvent tGraphicsEvent;
typedef XFontStruct tGraphicsFont;

Display* xws_display = NULL;

#define KEY_ESCAPE (XK_Escape)
#define KEY_CANCEL (XK_Cancel)
#define KEY_LEFT (XK_Left)
#define KEY_RIGHT (XK_Right)
#define KEY_END (XK_End)
#define KEY_NPAGE (XK_Page_Down)
#define KEY_PPAGE (XK_Page_Up)
#define KEY_DOWN (XK_Down)
#define KEY_UP (XK_Up)
#define KEY_HOME (XK_Home)
#define KEY_ENTER (XK_Return)
#define KEY_BACKSPACE (XK_BackSpace)
#define KEY_DC (XK_Delete)
#define KEY_IC (XK_Insert)

#else

/* some minor, not-yet-complete abstraction for the future... */
typedef GtkWidget tGraphicsWidget;
typedef GtkWindow tGraphicsWindow;
typedef GdkWindow tGraphicsLowlevelWindow;
typedef GdkGC tGraphicsContext;
typedef GdkEvent tGraphicsEvent;
typedef GdkFont tGraphicsFont;

static const char strGtkDestroy[] = "destroy", strGtkDelete[] = "delete_event",
  strGtkConfigure[] = "configure_event", strGtkClicked[] = "clicked",
  strGtkButton[] = "button_press_event", strGtkActivate[] = "activate",
  strGtkKey[] = "key_press_event";

FAIL!
#define use_colors (truE)
#define KEY_ESCAPE (GDK_Escape)
#define KEY_CANCEL (GDK_Cancel)
#define KEY_LEFT (GDK_Left)
#define KEY_RIGHT (GDK_Right)
#define KEY_END (GDK_End)
#define KEY_NPAGE (GDK_Page_Down)
#define KEY_PPAGE (GDK_Page_Up)
#define KEY_DOWN (GDK_Down)
#define KEY_UP (GDK_Up)
#define KEY_HOME (GDK_Home)
#define KEY_ENTER (GDK_Return)
#define KEY_BACKSPACE (GDK_BackSpace)
#define KEY_DC (GDK_Delete)
#define KEY_IC (GDK_Insert)

#endif

static const tGraphicsFont* default_font;

/* stubs'n'dummies - for now... */
#define COLS (80)
#define LINES (25)
#define A_NORMAL (0)
#define A_BOLD (0)
#define A_UNDERLINE (0)
#define A_REVERSE (0)
#define OK (0)
#define ERR (-1)
typedef unsigned long chtype;
typedef chtype attr_t;
typedef struct { short _curx, _cury; } tTextualWindowContents;
static my_inline int move(int y __cunused, int x __cunused) { return(OK); }
static my_inline int clrtoeol(void) { return(OK); }
static my_inline int addnstr(const char* s __cunused, int n __cunused)
{ return(OK); }
static my_inline int mvaddnstr(int y, int x, const char* s, int n)
{ move(y, x); addnstr(s, n); return(OK); }
static my_inline int addch(const chtype c __cunused) { return(OK); }
static my_inline int mvaddch(int y, int x, const chtype c)
{ move(y, x); addch(c); return(OK); }
static my_inline int addchnstr(const chtype* s __cunused, int n __cunused)
{ return(OK); }
static my_inline int COLOR_PAIR(int x __cunused) { return(0); }
static my_inline int color_set(short x __cunused, void* d __cunused)
{ return(OK); }
/* static my_inline int attron(attr_t a __cunused) { return(OK); }
   static my_inline int attroff(attr_t a __cunused) { return(OK); }
   static one_caller int refresh(void) { return(OK); } */
static my_inline chtype inch(void) { return(0); }

#else /* #if TGC_IS_GRAPHICS */

#if CONFIG_TG != TG_XCURSES
#undef KEY_ESCAPE
#define KEY_ESCAPE '\033'
#endif

#ifdef ACS_HLINE
#define __MY_HLINE (ACS_HLINE)
#else
#define __MY_HLINE ('-')
#endif

#ifdef ACS_VLINE
#define __MY_VLINE (ACS_VLINE)
#else
#define __MY_VLINE ('|')
#endif

#ifdef ACS_ULCORNER
#define __MY_UL (ACS_ULCORNER)
#else
#define __MY_UL ('+')
#endif

#ifdef ACS_URCORNER
#define __MY_UR (ACS_URCORNER)
#else
#define __MY_UR ('+')
#endif

#ifdef ACS_LLCORNER
#define __MY_LL (ACS_LLCORNER)
#else
#define __MY_LL ('+')
#endif

#ifdef ACS_LRCORNER
#define __MY_LR (ACS_LRCORNER)
#else
#define __MY_LR ('+')
#endif

#ifdef ACS_LTEE
#define __MY_LTEE (ACS_LTEE)
#else
#define __MY_LTEE ('+')
#endif

#ifdef ACS_RTEE
#define __MY_RTEE (ACS_RTEE)
#else
#define __MY_RTEE ('+')
#endif

#define VMIDDLE (MAX(((LINES - 2) / 2), 1))

#endif /* #if TGC_IS_GRAPHICS */

#ifndef KEY_TAB
#define KEY_TAB '\t'
#endif

#if OPTION_CED == 0
#define is_bad_uchar(c) ( ((c) < 32) || ((c) >= 127) )
#else
#define is_bad_uchar(c) ( ((c) < 32) || ((c) == 127) )
#endif

declare_local_i18n_buffer
__define_filestuff

static tResourceRequest* pm_request; /* for pmDump/pmDownload */

#if CONFIG_KEYMAPS
static const char strMouseFlip[] = "mouse-flip", strMouseOff[] = "mouse-off",
  strMouseOn[] = "mouse-on";
#endif

/* begin-autogenerated */
my_enum1 enum
{ pccUnknown = 0, pccDocumentBottom = 1, pccDocumentEnforceHtml = 2,
  pccDocumentEnforceSource = 3, pccDocumentReload = 4,
  pccDocumentReloadEnforced = 5, pccDocumentSave = 6, pccDocumentSearch = 7,
  pccDocumentSearchBackward = 8, pccDocumentSearchNext = 9,
  pccDocumentSearchPrevious = 10, pccDocumentTop = 11, pccDownload = 12,
  pccDownloadFromElement = 13, pccDump = 14, pccElementEnable = 15,
  pccElementInfo = 16, pccElementNext = 17, pccElementOpen = 18,
  pccElementOpenSplit = 19, pccElementPrevious = 20, pccExecextShell = 21,
  pccExecextShellFlip = 22, pccFormReset = 23, pccFormSubmit = 24,
  pccGoBookmarks = 25, pccGoHome = 26, pccGoSearch = 27, pccGoUri = 28,
  pccGoUriPreset = 29, pccJump = 30, pccJumpPreset = 31, pccLineDown = 32,
  pccLineUp = 33, pccLocalFileDirOpen = 34, pccMenuContextual = 35,
  pccMenuUriHistory = 36, pccMenuWindowlist = 37, pccMouseFlip = 38,
  pccMouseOff = 39, pccMouseOn = 40, pccPageDown = 41, pccPageUp = 42,
  pccQuit = 43, pccScreenSplit = 44, pccScreenSwitch = 45,
  pccScreenUnsplit = 46, pccSessionResume = 47, pccSessionSave = 48,
  pccStop = 49, pccViewBack = 50, pccViewForward = 51, pccWindowClose = 52,
  pccWindowNew = 53, pccWindowNewFromDocument = 54,
  pccWindowNewFromElement = 55, pccWindowNext = 56, pccWindowPrevious = 57
} my_enum2(unsigned char) tProgramCommandCode;

typedef struct
{ tKey key;
  tProgramCommandCode pcc;
} tKeymapCommandEntry;

static const tKeymapCommandEntry keymap_command_defaultkeys[] =
{ { '!', pccExecextShell },
  { '&', pccExecextShellFlip },
  { '.', pccStop },
  { '/', pccDocumentSearch },
  { '1', pccScreenUnsplit },
  { '2', pccScreenSplit },
  { '?', pccDocumentSearchBackward },
  { 'C', pccWindowClose },
  { 'D', pccDownloadFromElement },
  { 'E', pccElementEnable },
  { 'G', pccGoUriPreset },
  { 'H', pccDocumentEnforceHtml },
  { 'J', pccJumpPreset },
  { 'L', pccLineUp },
  { 'M', pccSessionResume },
  { 'N', pccWindowNewFromDocument },
  { 'O', pccWindowNewFromElement },
  { 'Q', pccQuit },
  { 'R', pccDocumentReloadEnforced },
  { 'S', pccSessionSave },
  { 'W', pccWindowPrevious },
  { 'Y', pccMouseFlip },
  { '\\', pccDocumentEnforceSource },
  { 'b', pccGoBookmarks },
  { 4, pccDownload },
  { 15, pccElementOpenSplit },
  { 23, pccMenuWindowlist },
  { KEY_DOWN, pccElementNext },
  { KEY_LEFT, pccViewBack },
  { KEY_RIGHT, pccViewForward },
  { KEY_UP, pccElementPrevious },
  { 'd', pccDump },
  { KEY_DC, pccLineDown },
  { 'e', pccGoSearch },
  { KEY_END, pccDocumentBottom },
  { KEY_ENTER, pccElementOpen },
  { 'g', pccGoUri },
  { 'h', pccGoHome },
  { KEY_HOME, pccDocumentTop },
  { 'i', pccElementInfo },
  { KEY_IC, pccLineUp },
  { 'j', pccJump },
  { 'l', pccLineDown },
  { 'm', pccMenuContextual },
  { 'n', pccWindowNew },
  { 'o', pccElementOpen },
  { KEY_NPAGE, pccPageDown },
  { KEY_PPAGE, pccPageUp },
  { 'r', pccDocumentReload },
  { 's', pccDocumentSave },
  { ' ', pccPageDown },
  { KEY_TAB, pccScreenSwitch },
  { 'u', pccMenuUriHistory },
  { 'w', pccWindowNext }
};

#if CONFIG_KEYMAPS

static const struct
{ const char* str; /* (sorted in strcmp() order) */
  tProgramCommandCode pcc; /* REMOVEME? */
} keymap_command_str2pcc[] =
{ { "document-bottom", pccDocumentBottom },
  { "document-enforce-html", pccDocumentEnforceHtml },
  { "document-enforce-source", pccDocumentEnforceSource },
  { "document-reload", pccDocumentReload },
  { "document-reload-enforced", pccDocumentReloadEnforced },
  { "document-save", pccDocumentSave },
  { "document-search", pccDocumentSearch },
  { "document-search-backward", pccDocumentSearchBackward },
  { "document-search-next", pccDocumentSearchNext },
  { "document-search-previous", pccDocumentSearchPrevious },
  { "document-top", pccDocumentTop },
  { "download", pccDownload },
  { "download-from-element", pccDownloadFromElement },
  { "dump", pccDump },
  { "element-enable", pccElementEnable },
  { "element-info", pccElementInfo },
  { "element-next", pccElementNext },
  { "element-open", pccElementOpen },
  { "element-open-split", pccElementOpenSplit },
  { "element-previous", pccElementPrevious },
  { "execext-shell", pccExecextShell },
  { "execext-shell-flip", pccExecextShellFlip },
  { "form-reset", pccFormReset },
  { "form-submit", pccFormSubmit },
  { "go-bookmarks", pccGoBookmarks },
  { "go-home", pccGoHome },
  { "go-search", pccGoSearch },
  { "go-url", pccGoUri },
  { "go-url-preset", pccGoUriPreset },
  { "jump", pccJump },
  { "jump-preset", pccJumpPreset },
  { "line-down", pccLineDown },
  { "line-up", pccLineUp },
  { "local-file-dir-open", pccLocalFileDirOpen },
  { "menu-contextual", pccMenuContextual },
  { "menu-url-history", pccMenuUriHistory },
  { "menu-windowlist", pccMenuWindowlist },
  { strMouseFlip, pccMouseFlip },
  { strMouseOff, pccMouseOff },
  { strMouseOn, pccMouseOn },
  { "page-down", pccPageDown },
  { "page-up", pccPageUp },
  { strQuit, pccQuit },
  { "screen-split", pccScreenSplit },
  { "screen-switch", pccScreenSwitch },
  { "screen-unsplit", pccScreenUnsplit },
  { "session-resume", pccSessionResume },
  { "session-save", pccSessionSave },
  { "stop", pccStop },
  { "view-back", pccViewBack },
  { "view-forward", pccViewForward },
  { "window-close", pccWindowClose },
  { "window-new", pccWindowNew },
  { "window-new-from-document", pccWindowNewFromDocument },
  { "window-new-from-element", pccWindowNewFromElement },
  { "window-next", pccWindowNext },
  { "window-previous", pccWindowPrevious }
};

#endif

my_enum1 enum
{ liacUnknown = 0, liacAreaSwitch = 1, liacCancel = 2, liacMouseFlip = 3,
  liacMouseOff = 4, liacMouseOn = 5, liacPass2user = 6, liacToEnd = 7,
  liacToLeft = 8, liacToRight = 9, liacToStart = 10
} my_enum2(unsigned char) tLineInputActionCode;

typedef struct
{ tKey key;
  tLineInputActionCode liac;
} tKeymapLineinputEntry;

static const tKeymapLineinputEntry keymap_lineinput_defaultkeys[] =
{ { 1, liacAreaSwitch },
  { 21, liacPass2user },
  { KEY_DOWN, liacToEnd },
  { KEY_LEFT, liacToLeft },
  { KEY_RIGHT, liacToRight },
  { KEY_UP, liacToStart },
  { KEY_END, liacToEnd },
  { KEY_CANCEL, liacCancel },
  { KEY_HOME, liacToStart },
  { KEY_NPAGE, liacToEnd },
  { KEY_PPAGE, liacToStart }
};

#if CONFIG_KEYMAPS

static const struct
{ const char* str; /* (sorted in strcmp() order) */
  tLineInputActionCode liac; /* REMOVEME? */
} keymap_lineinput_str2liac[] =
{ { "area-switch", liacAreaSwitch },
  { "cancel", liacCancel },
  { strMouseFlip, liacMouseFlip },
  { strMouseOff, liacMouseOff },
  { strMouseOn, liacMouseOn },
  { "pass2user", liacPass2user },
  { "to-end", liacToEnd },
  { "to-left", liacToLeft },
  { "to-right", liacToRight },
  { "to-start", liacToStart }
};

#endif
/* end-autogenerated */

static const_after_init tKeymapCommandEntry* keymap_command_keys = NULL;
static const_after_init size_t keymap_command_keys_num = 0,
  keymap_command_keys_maxnum = 0;

static int __init keymap_command_sorter(const void* _a, const void* _b)
{ const tKeymapCommandEntry *a = (const tKeymapCommandEntry*) _a,
    *b = (const tKeymapCommandEntry*) _b;
  tKey ak = a->key, bk = b->key;
  return(my_numcmp(ak, bk));
}

static tBoolean __init keymap_command_key_do_register(tKey key,
  tProgramCommandCode pcc)
/* returns whether it worked */
{ size_t count;
  for (count = 0; count < keymap_command_keys_num; count++) /* IMPROVEME? */
  { if (keymap_command_keys[count].key == key) return(falsE); } /* redefined */

  if (keymap_command_keys_num >= keymap_command_keys_maxnum)
  { keymap_command_keys_maxnum += 32;
    keymap_command_keys = (tKeymapCommandEntry*)
      memory_reallocate(keymap_command_keys, keymap_command_keys_maxnum *
      sizeof(tKeymapCommandEntry));
  }
  keymap_command_keys[keymap_command_keys_num].key = key;
  keymap_command_keys[keymap_command_keys_num].pcc = pcc;
  keymap_command_keys_num++;
  return(truE);
}

static const_after_init tKeymapLineinputEntry* keymap_lineinput_keys = NULL;
static const_after_init size_t keymap_lineinput_keys_num = 0,
  keymap_lineinput_keys_maxnum = 0;

static int __init keymap_lineinput_sorter(const void* _a, const void* _b)
{ const tKeymapLineinputEntry *a = (const tKeymapLineinputEntry*) _a,
    *b = (const tKeymapLineinputEntry*) _b;
  tKey ak = a->key, bk = b->key;
  return(my_numcmp(ak, bk));
}

static tBoolean __init keymap_lineinput_key_do_register(tKey key,
  tLineInputActionCode liac)
/* returns whether it worked */
{ size_t count;
  for (count = 0; count < keymap_lineinput_keys_num; count++) /* IMPROVEME? */
  { if (keymap_lineinput_keys[count].key == key) return(falsE); } /*redefined*/

  if (keymap_lineinput_keys_num >= keymap_lineinput_keys_maxnum)
  { keymap_lineinput_keys_maxnum += 16;
    keymap_lineinput_keys = (tKeymapLineinputEntry*)
      memory_reallocate(keymap_lineinput_keys, keymap_lineinput_keys_maxnum *
      sizeof(tKeymapLineinputEntry));
  }
  keymap_lineinput_keys[keymap_lineinput_keys_num].key = key;
  keymap_lineinput_keys[keymap_lineinput_keys_num].liac = liac;
  keymap_lineinput_keys_num++;
  return(truE);
}

#if CONFIG_KEYMAPS

static tKey __init keystr2key(const char* keystr)
{ tKey key = '\0';
  unsigned char ch = keystr[0];
  if ( (ch != '\0') && (keystr[1] == '\0') ) key = ch; /* most likely case? */
  else if (!strcmp(keystr, "delete")) key = KEY_DC;
  else if (!strcmp(keystr, "end")) key = KEY_END;
  /* else if (!strcmp(keystr, "enter")) key = KEY_ENTER; */
  else if (!strcmp(keystr, "insert")) key = KEY_IC;
  else if (!strcmp(keystr, "home")) key = KEY_HOME;
  else if (!strcmp(keystr, "page-down")) key = KEY_NPAGE;
  else if (!strcmp(keystr, "page-up")) key = KEY_PPAGE;
  else if (!strcmp(keystr, "escape")) key = KEY_CANCEL;
  else if (!strcmp(keystr, "space")) key = ' ';
  else if (!strcmp(keystr, "tab")) key = '\t';
  else if (!strncmp(keystr, "ctrl-", 5))
  { char cc = keystr[5]; /* "control character" */
    if ( (my_islower(cc)) && (keystr[6] == '\0') ) key = cc - 'a' + 1;
  }
  else if (!strncmp(keystr, "cursor-", 7))
  { const char* tmp = keystr + 7;
    if (!strcmp(tmp, "down")) key = KEY_DOWN;
    else if (!strcmp(tmp, "left")) key = KEY_LEFT;
    else if (!strcmp(tmp, "right")) key = KEY_RIGHT;
    else if (!strcmp(tmp, "up")) key = KEY_UP;
  }
#ifdef KEY_F
  else if (!strncmp(keystr, "fn-", 3))
  { const char* tmp = keystr + 3;
    if (my_isdigit(*tmp))
    { int num;
      my_atoi(tmp, &num, &tmp, 99);
      if ( (num >= 0) && (num <= 63) && (*tmp == '\0') ) key = KEY_F(num);
    }
  }
#endif
  return(key);
}

static one_caller tMbsIndex keymap_cmdstr_lookup(const char* str)
{ my_binary_search(0, ARRAY_ELEMNUM(keymap_command_str2pcc) - 1,
    strcmp(str, keymap_command_str2pcc[idx].str), return(idx))
}

unsigned char __init keymap_command_key_register(const char* keystr,
  const char* cmdstr)
/* return value: 0=fine, 1=bad key identifier; 2=repeatedly defined key; 3=bad
   command; IMPROVEME: binary search for <keystr>? */
{ tKey key = keystr2key(keystr);
  tMbsIndex idx;
  tProgramCommandCode pcc;

  if (key == '\0') return(1);

  /* interpret the command */
  idx = keymap_cmdstr_lookup(cmdstr);
  if (idx < 0) return(3);
  pcc = keymap_command_str2pcc[idx].pcc;

  /* register */
  if (!keymap_command_key_do_register(key, pcc)) return(2);
  return(0);
}

static one_caller tMbsIndex keymap_listr_lookup(const char* str)
{ my_binary_search(0, ARRAY_ELEMNUM(keymap_lineinput_str2liac) - 1,
    strcmp(str, keymap_lineinput_str2liac[idx].str), return(idx))
}

unsigned char __init keymap_lineinput_key_register(const char* keystr,
  const char* listr)
/* return value: 0=fine, 1=bad key identifier; 2=repeatedly defined key; 3=bad
   command; IMPROVEME: binary search for <keystr>? */
{ tKey key = keystr2key(keystr);
  tMbsIndex idx;
  tLineInputActionCode liac;

  if (key == '\0') return(1);

  /* interpret the action */
  idx = keymap_listr_lookup(listr);
  if (idx < 0) return(3);
  liac = keymap_lineinput_str2liac[idx].liac;

  /* register */
  if (!keymap_lineinput_key_do_register(key, liac)) return(2);
  return(0);
}

#endif /* #if CONFIG_KEYMAPS */

#if TGC_IS_GRAPHICS

#define is_khm_command (truE) /* only one key-handling mode available... */

#else

enum
{ khmCommand = 0, khmLineInput = 1
#if OPTION_MENUS
  , khmContextualMenu = 2
#endif
};
typedef unsigned char tKeyHandlingMode;
static tKeyHandlingMode key_handling_mode = khmCommand;
#define is_khm_command (key_handling_mode == khmCommand)

#define is_bottom_occupied (key_handling_mode == khmLineInput)

#endif

my_enum1 enum
{ prrfNone = 0, prrfRedrawAll = 0x01, prrfRedrawOne = 0x02, prrfSource = 0x04,
  prrfReload = 0x08, prrfEnforcedReload = 0x10, prrfIsHttpRedirection = 0x20,
  prrfIsFtpRedirection = 0x40, prrfUseSfbuf = 0x80, prrfPost = 0x100,
  prrfDownload = 0x200
} my_enum2(unsigned short) tPrrFlags; /* for prepare_resource_request() */
#define prrfIsRedirection (prrfIsHttpRedirection | prrfIsFtpRedirection)

static int prr_download_fd;

/* stuff for khmLineInput mode: */

my_enum1 enum
{ liekFail = 0, liekCancel = 1, liekKey = 2
} my_enum2(unsigned char) tLineInputEventKind;

typedef void (*tLineInputCallback)(void*, tLineInputEventKind, tKey key);

my_enum1 enum
{ liafNone = 0, liafFocusable = 0x01, liafEditable = 0x02,
  liafDisguising = 0x04, liafAllowEmptyText = 0x08
} my_enum2(unsigned char) tLineInputAreaFlags;

typedef struct
{ char* text;
  short len, maxlen, pos, first; /* CHECKME: size_t? */
  short row, colmin, colcurr, colmax, usable;
  tLineInputAreaFlags flags;
} tLineInputArea;
typedef unsigned char tLineInputAreaIndex;

struct tWindow;

static struct
{ tLineInputArea area[2];
#if TGC_IS_GRAPHICS
  struct tWindow* window;
  GtkEntry* entry;
#endif
  tLineInputCallback callback;
  void* callback_data;
  tPrrFlags prrf; /* rarely needed */
  tLineInputAreaIndex curr, num_areas;
} lid; /* aka "line input data" */

#define lid_area(what) (lid.area[lid.curr].what)

my_enum1 enum
{ eligGoToUri = 0, eligSaveAs = 1, eligDumpFilename = 2, eligDownloadUri = 3,
  eligDownloadFilename = 4, eligDocumentSearch = 5,
  eligDocumentSearchBackward = 6
#if CONFIG_USER_QUERY
  , eligUsername = 7, eligPassword = 8
#endif
#if !TGC_IS_GRAPHICS
  , eligFormText = 9, eligFormPassword = 10, eligFormFilename = 11
#endif
#if CONFIG_SESSIONS
  , eligSessionSave = 12, eligSessionResume = 13
#endif
#if CONFIG_JUMPS
  , eligJump = 14
#endif
#if OPTION_EXECEXT & 1
  , eligExecextShell = 15
#endif
} my_enum2(unsigned char) tEditableLineInputGoal;

#if CONFIG_USER_QUERY
#define is_disguising_elig(elig) \
  ( ((elig) == eligFormPassword) || ((elig) == eligPassword) )
#else
#define is_disguising_elig(elig) ((elig) == eligFormPassword)
#endif

#define is_emptyok_elig(elig) ( ((elig) == eligFormText) || \
  ((elig) == eligFormPassword) || ((elig) == eligFormFilename) )

my_enum1 enum
{ cgQuit = 0, cgClose = 1, cgOverwriteFile = 2, cgOverwriteDump = 3,
  cgOverwriteDownload = 4,
#if CONFIG_SESSIONS
  cgOverwriteSession = 5,
#endif
  cgSubmit = 6, cgReset = 7, cgRepost = 8, cgHtml = 9, cgEnable = 10
#if CONFIG_FTP && OPTION_TLS
  , cgFtpsDataclear = 11
#endif
} my_enum2(unsigned char) tConfirmationGoal;


static const char *khmli_download_uri = NULL, *khmli_filename = NULL;

#if OPTION_MENUS
#define URI_HISTORY_LEN (20)
static char* uri_history[URI_HISTORY_LEN];
static unsigned short uri_history_index = 0;
#endif

enum
{ wvfNone = 0, wvfScreenFull = 0x01, wvfDontDrawContents = 0x02,
  wvfScrollingUp = 0x04, wvfSource = 0x08, wvfHtmlTitleDone = 0x10,
  wvfHandledRedirection = 0x20, wvfPost = 0x40, wvfEnforceHtml = 0x80
};
/* wvfScreenFull implies two things: 1. the user may scroll down; 2. we need
   not redraw the view when the resource handler got more content for it. */
typedef unsigned char tWindowViewFlags;
typedef signed long tLinenumber;

typedef struct tActiveElementCoordinates
{ struct tActiveElementCoordinates* next;
  tLinenumber y;
  short x1, x2;
} tActiveElementCoordinates;
/* Such a list usually consists of only one entry, sometimes two (if a
   linebreak occurs within an active element). In very small table columns, the
   list might become longer - that's the main point with this structure. */

typedef struct
{ char* current_text; /* for text/password/... fields */
#if TGC_IS_GRAPHICS
  tGraphicsWidget* widget;
#else
  tActiveElementCoordinates* aec;
#endif
  tActiveElementFlags flags; /* only contains aefChangeable flags */
} tActiveElement;

typedef struct tWindowView /* a "view" of a resource within a window */
{ struct tWindow* window; /* the window to which the view belongs */
  struct tWindowView *prev, *next; /* list of views within that window */
  struct tResourceRequest* request; /*request/resource displayed in this view*/
  const char *title, *html_title, *last_info, *anchor;
  tActiveElement* active_element;
  tActiveElementNumber aenum, aemax;
#if !TGC_IS_GRAPHICS
  tActiveElementNumber aecur, former_ae;
#endif
  tLinenumber origin_y; /* line of content in top-left corner of window */
  tWindowViewFlags flags;
  unsigned char redirections; /* counter to avoid infinite loops */
} tWindowView;

typedef struct tWindow /* a "virtual window" */
{ struct tWindow *prev, *next;
  tWindowView *first_view, *current_view;
#if TGC_IS_GRAPHICS
  tGraphicsWindow* ww; /* "window widget" */
  tGraphicsWidget *contents, *location, *info, *message;
  tGraphicsLowlevelWindow* drawable; /* for the actual contents */
  tGraphicsContext* gc; /* ditto */
#endif
} tWindow;

static /*@null@*/ tWindow *windowlisthead = NULL, *windowlisttail = NULL;

#if TGC_IS_GRAPHICS
static tWindow* current_window = NULL;
#else
static /*@relnull@*/ tWindow* visible_window[2] = { NULL, NULL };
typedef unsigned short tVisibleWindowIndex;
static tVisibleWindowIndex current_window_index = 0;
#define current_window visible_window[current_window_index]
#endif /* #if TGC_IS_GRAPHICS */
#define is_normal_window(w) (1) /* (_currently_!) */

my_enum1 enum
{ wrtRedraw = 0, wrtRedrawRecursive = 1, wrtSearch = 2, wrtSearchBackward = 3,
  wrtToEnd = 4
} my_enum2(unsigned char) tWindowRedrawingTask;

/* some prototypes */
static void __window_redraw(const tWindow*, tWindowRedrawingTask);
#define window_redraw(window) __window_redraw(window, wrtRedraw)
static void window_hide(tWindow*, unsigned char);
static void handle_command_code(tProgramCommandCode);
static void start_save_as(const char*, tBoolean);
static void __line_input_start(/*@notnull@*/ const char*, const char*,
  tLineInputCallback, void*, tLineInputAreaFlags);
#define line_input_start(a, b, c, d, e, f, g, h) __line_input_start(a,b,c,d,e)
static void __line_input_estart(tEditableLineInputGoal, const char*,
  const char*, tUserQuery*);
#define line_input_estart(a, b, c, d, e, f, g) __line_input_estart(a, b, c, d)
static void test_redirection(tWindowView*, tResource*);

#if TGC_IS_GRAPHICS
static my_inline __sallocator char* __callocator
  my_strdup_ucfirst(const char* str)
/* duplicates a string, changing the first character to uppercase */
{ char *retval = my_strdup(str), ch = retval[0];
  retval[0] = my_toupper(ch);
  return(retval);
}
#endif

static __sallocator tResourceRequest* __callocator request_allocate(void)
{ tResourceRequest* retval =
    (tResourceRequest*) memory_allocate(sizeof(tResourceRequest));
  retval->id = object_id_create();
#if CONFIG_DEBUG
  sprint_safe(strbuf, "request_allocate(): %p\n", retval); debugmsg(strbuf);
#endif
  return(retval);
}

static void __request_deallocate(tResourceRequest* request)
{ __dealloc(request->uri); __dealloc(request->hostname);
  __dealloc(request->username); __dealloc(request->password);
  __dealloc(request->path); __dealloc(request->query);
  __dealloc(request->post);
#if CONFIG_DEBUG
  sprint_safe(strbuf, "__request_deallocate(): %p\n", request);
  debugmsg(strbuf);
#endif
  memory_deallocate(request);
}

static void request_deallocate(tResourceRequest* request)
{ resource_callback(request, mckRemovingRequest);
  object_mark_vanished(request->id);
  __request_deallocate(request);
}


/* Window and window-view handling; content display */

#if CONFIG_TG == TG_GTK

static unsigned int window_counter = 0;
#define connect_object(object, signal, handler, data) \
  gtk_signal_connect(object, signal, GTK_SIGNAL_FUNC(handler), data)
#define connect_widget(widget, signal, handler, data) \
  connect_object(GTK_OBJECT(widget), signal, handler, data)
#define connect_window(window, signal, handler, data) \
  connect_object(GTK_OBJECT(window), signal, handler, data)
#define show_widget(x) gtk_widget_show(GTK_WIDGET(x))
#define pack_box(container, element) \
  gtk_box_pack_start(GTK_BOX(container), element, FALSE, FALSE, 0)

typedef struct
{ tWindow* window;
  tProgramCommandCode pcc;
} tGtkWindowCommand;

static my_inline __sallocator tGtkWindowCommand* __callocator
  create_window_command(const tWindow* window, tProgramCommandCode pcc)
{ tGtkWindowCommand* retval =
    (tGtkWindowCommand*) __memory_allocate(sizeof(tGtkWindowCommand));
  retval->window = __unconstify(tWindow*, window); retval->pcc = pcc;
  return(retval);
}

static void graphics_handle_command_code(GtkWidget* widget, gpointer _data)
{ const tGtkWindowCommand* data = (tGtkWindowCommand*) _data;
  tWindow* window = data->window;
  const GtkWindow* ww = window->ww;
  tProgramCommandCode pcc = data->pcc;
  current_window = window; handle_command_code(pcc); current_window = NULL;
}

static gint graphics_propagate_event(tGraphicsWidget* w __cunused,
  tGraphicsEvent* event __cunused, gpointer data __cunused)
{ /* just enforce the propagation of signals to parent widgets: */
  return(FALSE);
}

static gint window_handle_delete(tGraphicsWidget* w __cunused,
  tGraphicsEvent* event __cunused, gpointer data __cunused)
{ /* don't handle it here, let GTK call window_handle_destroy()... */
  return(FALSE);
}

static gint window_handle_destroy(tGraphicsWidget* w __cunused, gpointer data)
{ window_hide((tWindow*) data, 2);
  return(TRUE);
}

static gint window_handle_configure(tGraphicsWidget* widget __cunused,
  tGraphicsEvent* event __cunused, gpointer data)
{ tWindow* w = (tWindow*) data;
  window_redraw(w);
  return(TRUE);
}

static tGraphicsWidget* create_menu(tGraphicsWidget* bar, char* _title)
{ tGraphicsWidget *menu = gtk_menu_new(),
    *title = gtk_menu_item_new_with_label(_title);
  show_widget(title);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(title), menu);
  gtk_menu_bar_append(GTK_MENU_BAR(bar), title);
  return(menu);
}

static void __create_item(const tWindow* window, GtkAccelGroup* accel,
  tGraphicsWidget* menu, const char* text, tProgramCommandCode pcc, tKey key)
{ tGraphicsWidget* item = gtk_menu_item_new_with_label(text);
  gtk_menu_append(GTK_MENU(menu), item);
  connect_widget(item, strGtkActivate, graphics_handle_command_code,
    (gpointer) create_window_command(window, pcc));
  if (key != '\0')
  { GdkModifierType t;
    if (my_isupper(key)) { t = GDK_CONTROL_MASK; key = my_tolower(key); }
    else t = GDK_MOD1_MASK;
    gtk_widget_add_accelerator(item, strGtkActivate, accel, key, t,
      GTK_ACCEL_VISIBLE);
  }
  show_widget(item);
}

#define create_item(m, t, pcc, k) __create_item(retval, accel, m, t, pcc, k)

static void create_button(const tWindow* window, tGraphicsWidget* container,
  char* text, tProgramCommandCode pcc)
{ tGraphicsWidget* button = gtk_button_new_with_label(text);
  connect_widget(button, strGtkClicked, graphics_handle_command_code,
    (gpointer) create_window_command(window, pcc));
  pack_box(container, button);
  show_widget(button);
}

static gint graphics_cm(tGraphicsWidget* w __cunused, tGraphicsEvent* _event,
  gpointer data)
/* shows a contextual menu in graphics mode */
{
  if (_event->type == GDK_BUTTON_PRESS)
  { GdkEventButton* event = (GdkEventButton*) _event;
    guint b = event->button;
    if (b == 3)
    { tGraphicsWidget *menu = gtk_menu_new(),
        *item = gtk_menu_item_new_with_label("TEST");
      tWindow* window = (tWindow*) data;
      show_widget(menu); show_widget(item);
      gtk_menu_append(GTK_MENU(menu), item);
      gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, b, event->time);
      return(TRUE);
    }
  }
  return(FALSE);
}


static const char* const retawq_logo_mini[] = {
"64 64 2 1",
" 	c None",
".	c #FF0000",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                        .....                   ",
"                         ....        ...........                ",
"            ..................     ..............               ",
"            ..................    ................              ",
"            ..................   ..................             ",
"            ..................   ..................             ",
"                 .............  ...... ............             ",
"                  ............ .....   ............             ",
"                  ............ ....    ............             ",
"                  .................    ............             ",
"                   ...............     ............             ",
"                   ...............     ............             ",
"                   ..............       ..........              ",
"                   ..............        ........               ",
"                   ..............          ....                 ",
"                   .............                                ",
"                   .............                                ",
"                   .............                                ",
"                   .............                                ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                   ............                                 ",
"                  .............                                 ",
"                  ..............                                ",
"             .........................                          ",
"            ..........................                          ",
"            ..........................                          ",
"            ..........................                          ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                ",
"                                                                "};

#endif /* #if CONFIG_TG == TG_GTK */

static void __init_ae(const tActiveElementBase* aeb, tActiveElement* ae,
  tBoolean is_first_init
#if TGC_IS_GRAPHICS
  , tWindowView* view
#endif
  )
{ const tActiveElementKind aek = aeb->kind;
  const char* text;

  if (is_first_init) my_memclr(ae, sizeof(tActiveElement));
  else dealloc(ae->current_text);
  ae->flags = aeb->flags & aefChangeable;

  if ((aek == aekFormText) || (aek == aekFormPassword) || (aek == aekFormFile))
  { if ( (text = aeb->render) != NULL ) ae->current_text = my_strdup(text);
  }
  else if (aek == aekFormSelect)
  { const tHtmlInputLength count = aeb->maxlength;
    if (count > 0) /* non-empty selection list */
    { char* t = ae->current_text = memory_allocate((count + 7) / 8);
      const tHtmlOption* o = (const tHtmlOption*) aeb->render;
      tHtmlInputLength c = 0;
      while (o != NULL)
      { if (o->flags & hofSelected) my_bit_set(t, c);
        o = o->next; c++;
      }
    }
  }

#if TGC_IS_GRAPHICS
  if (is_first_init) /* possibly create a widget */
  { tGraphicsWidget* w;
    tHtmlInputLength maxlength;
    switch (aek)
    { case aekFormText: case aekFormPassword: case aekFormTextarea:
        text = aeb->render;
        maxlength = (has_input_length(aek) ? aeb->maxlength : 0);
        /* FIXME: don't show passwords! */
        /* w = gtk_text_new(NULL, NULL); */
        w = ( (maxlength > 0) ? gtk_entry_new_with_max_length(maxlength) :
          gtk_entry_new() );
        if ( (text != NULL) && (text[0] != '\0') )
          gtk_entry_set_text(GTK_ENTRY(w), text);
        break;
      case aekFormCheckbox: w = gtk_check_button_new(); break;
      case aekFormRadio: w = gtk_radio_button_new(NULL); break;/*IMPLEMENTME!*/
      case aekFormFile: w = gtk_file_selection_new(strEmpty); break;
      default: w = NULL; break;
    }
    /* FIXME: check whether disabled, readonly, ...! */
    if (w != NULL)
    { pack_box(view->window->contents, w); show_widget(w); ae->widget = w; }
  }
#endif
}

#if TGC_IS_GRAPHICS
#define init_ae(a, b, c, d) __init_ae(a, b, c, d)
#else
#define init_ae(a, b, c, d) __init_ae(a, b, c)
#endif

static void deallocate_aec(/*const*/ tActiveElementCoordinates** _aec)
{ const tActiveElementCoordinates* aec = *_aec;
  if (aec == NULL) return; /* nothing to do */
  while (aec != NULL)
  { const tActiveElementCoordinates* aec_next = aec->next;
    memory_deallocate(aec); aec = aec_next;
  }
  *_aec = NULL;
}

static void deallocate_windowviewlist(const tWindowView* view)
/* deallocates the given list of views and drops associated request and
   resource data */
{ while (view != NULL)
  { const tWindowView* next = view->next;
    request_deallocate(view->request);
    if (view->aenum > 0)
    { tActiveElement* aes = view->active_element;
      tActiveElementNumber _ae;
      for (_ae = 0; _ae < view->aenum; _ae++)
      { __dealloc(aes[_ae].current_text);
#if TGC_IS_GRAPHICS
        /* FIXME: handle widget! */
#else
        deallocate_aec(&(aes[_ae].aec));
#endif
      }
      memory_deallocate(aes);
    }
    __dealloc(view->title); __dealloc(view->html_title);
    __dealloc(view->last_info); __dealloc(view->anchor);
    memory_deallocate(view);
    view = next;
  }
}

static void cut_windowviewlist(tWindow* window)
/* "cuts" the current view and all its successors out of the <window> and
   deallocates them */
{ tWindowView* view = window->current_view;
  if (view != NULL)
  { if (view->prev != NULL) view->prev->next = NULL;
    window->current_view = view->prev;
    if (window->first_view == view) window->first_view = NULL;
    deallocate_windowviewlist(view);
  }
}

static /* __sallocator -- not an "only" reference... */ tWindow* __callocator
  window_create(void)
{ tWindow* retval = (tWindow*) memory_allocate(sizeof(tWindow));
  if (windowlisthead == NULL)
  { windowlisthead
#if TGC_IS_CURSES
      = visible_window[0]
#endif
      = retval;
  }
  if (windowlisttail != NULL)
  { retval->prev = windowlisttail; windowlisttail->next = retval; }
  windowlisttail = retval;

  /* All the horrible rest is for graphics mode only... */
#if TGC_IS_GRAPHICS
  { tGraphicsWindow* w = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
    tGraphicsWidget *base=gtk_vbox_new(FALSE,0), *fluff=gtk_vbox_new(FALSE,0),
      *buttons = gtk_hbox_new(FALSE, 0), *texts = gtk_vbox_new(FALSE, 0),
      *contents = gtk_drawing_area_new(), *url0 = gtk_hbox_new(FALSE, 0),
      *url1 = gtk_label_new(_("Location: ")), *url2 = gtk_label_new(strEmpty),
      *info0 = gtk_hbox_new(FALSE, 0), *info1 = gtk_label_new(strEmpty),
      *info2 = gtk_label_new(strEmpty), *mbar = gtk_menu_bar_new(),
      *mFile = create_menu(mbar, _(strFileUc)), *mView = create_menu(mbar,
      _("View")), *mBookmark = create_menu(mbar, _("Bookmark")),
      *mWindow = create_menu(mbar, _("Window"));
    tGraphicsLowlevelWindow *some_parent, *drawable;
    GdkWindowAttr* drawattrs = memory_allocate(sizeof(GdkWindowAttr));
    GtkStyle* style;
    tGraphicsContext* gc;
    static tBoolean is_first_call = truE;
    static char *strHome, *strStop;
    static GdkPixmap* logo_mini_pixmap;
    static GdkBitmap* logo_mini_mask;
    GtkAccelGroup* accel = gtk_accel_group_new();

    if (is_first_call)
    { tGraphicsLowlevelWindow* template_window;
      GdkWindowAttr* attr = memory_allocate(sizeof(GdkWindowAttr));
      is_first_call = falsE;
      strHome = i18n_strdup(_("Home")); strStop = i18n_strdup(_("Stop"));
      attr->window_type = GDK_WINDOW_CHILD;
      template_window = gdk_window_new(NULL, attr, 0);
      logo_mini_pixmap = gdk_pixmap_create_from_xpm_d(template_window,
        &logo_mini_mask, NULL, (gchar**) retawq_logo_mini);
    }

    gtk_window_set_policy(w, TRUE /* CHECKME: documentation contradicts it! */,
      TRUE, FALSE);
    gtk_window_set_title(GTK_WINDOW(w), strVersion);
    gtk_window_add_accel_group(GTK_WINDOW(w), accel);

    gtk_drawing_area_size(GTK_DRAWING_AREA(contents), config.width,
      config.height);
    gtk_widget_set_events(contents, GDK_EXPOSURE_MASK);

    /* basic window structure */
    gtk_container_add(GTK_CONTAINER(w), base);
    pack_box(base, mbar); pack_box(base, fluff); pack_box(base, contents);
    pack_box(fluff, buttons); pack_box(fluff, texts);
    pack_box(texts, url0); pack_box(texts, info0);
    pack_box(url0, url1); pack_box(url0, url2);
    pack_box(info0, info1); pack_box(info0, info2);

    /* menus */
    create_item(mFile, _("New Window"), pccWindowNew, 'n');
    create_item(mFile, _(strOpenInNewWindow), pccWindowNewFromDocument, 'N');
    create_item(mFile, _("Open Local File..."), pccLocalFileDirOpen, 'O');
    create_item(mFile, _("Open URL..."), pccGoUri, 'o');
    create_item(mFile, _("Open Relative URL..."), pccGoUriPreset, '\0');
    create_item(mFile, _(strSaveAs), pccDocumentSave, 's');
    create_item(mFile, _("Close"), pccWindowClose, 'c');
    create_item(mFile, _("Quit"), pccQuit, 'q');
    create_item(mView, _(strBack), pccViewBack, '\0');
    create_item(mView, _(strForward), pccViewForward, '\0');
    create_item(mView, _(strReload), pccDocumentReload, 'R');
    if (config.home_uri != NULL) create_item(mView, strHome,pccGoHome, 'h');
    if (config.search_engine != NULL)
      create_item(mView, _(strSearch), pccGoSearch, 'e');
    create_item(mView, strStop, pccStop, '.');
    create_item(mView, _("Document Info"), pccUnknown, 'I');
    create_item(mView, _(strSourceCode), pccDocumentEnforceSource, '\\');
    create_item(mBookmark, _(strAddBookmark), pccUnknown, '\0');
    create_item(mWindow, "(Titles of all windows)", pccUnknown, '\0');

    /* buttons */
    create_button(retval, buttons, _(strBack), pccViewBack);
    create_button(retval, buttons, _(strForward), pccViewForward);
    create_button(retval, buttons, _(strReload), pccDocumentReload);
    if (config.home_uri != NULL)
      create_button(retval, buttons, strHome, pccGoHome);
    if (config.search_engine != NULL)
      create_button(retval, buttons, _(strSearch), pccGoSearch);
    create_button(retval, buttons, strStop, pccStop);

    /* show all */
    show_widget(url0); show_widget(url1); show_widget(url2);
    show_widget(info0); show_widget(info1); show_widget(info2);
    show_widget(mbar); show_widget(buttons); show_widget(texts);
    show_widget(fluff); show_widget(contents); show_widget(base);
    show_widget(w);

    /* GDK window and gc for contents */
    drawattrs->event_mask = gtk_widget_get_events(contents) | GDK_EXPOSURE_MASK
      | GDK_KEY_PRESS_MASK | GDK_BUTTON_PRESS_MASK;
    drawattrs->x = contents->allocation.x;
    drawattrs->y = contents->allocation.y;
    drawattrs->width = contents->allocation.width;
    drawattrs->height = contents->allocation.height;
    drawattrs->wclass = GDK_INPUT_OUTPUT;
    drawattrs->visual = gtk_widget_get_visual(contents);
    drawattrs->colormap = gtk_widget_get_colormap(contents);
    drawattrs->window_type = GDK_WINDOW_CHILD;
    some_parent = gtk_widget_get_parent_window(contents);
    drawable = gdk_window_new(some_parent, drawattrs,
      GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP);
    gdk_window_set_icon(some_parent, NULL, logo_mini_pixmap, logo_mini_mask);
    style = gtk_widget_get_style(contents);
    gc = gdk_gc_new(drawable);
    gdk_window_set_background(drawable, &(style->white));
    gdk_gc_set_foreground(gc, &(style->black));
    gdk_gc_set_background(gc, &(style->white));
    gdk_window_show(drawable);

    /* store */
    retval->ww = w;
    retval->contents = contents; retval->location = url2;
    retval->info = info1; retval->message = info2; window_counter++;
    retval->drawable = drawable; retval->gc = gc;

    /* connect signal handlers */
    connect_window(w, strGtkDestroy, window_handle_destroy, retval);
    connect_window(w, strGtkDelete, window_handle_delete, retval);
    connect_widget(contents, strGtkConfigure, window_handle_configure, retval);
    connect_widget(contents, strGtkKey, graphics_propagate_event, NULL);
    connect_widget(contents, strGtkButton, graphics_cm, retval);
  }
#endif
  return(retval);
}

#if TGC_IS_GRAPHICS
#define window_is_visible(window) (truE)
#else
static tBoolean window_is_visible(tWindow* window)
{ if (window == NULL) return(falsE); /* "should not happen" */
  if ( (window == visible_window[0]) || (window == visible_window[1]) )
    return(truE);
  return(falsE);
}
#endif

static my_inline void require_normal_current_window(void)
{
#if TGC_IS_CURSES
  if (!is_normal_window(current_window)) current_window = window_create();
#endif
}

static /* __sallocator -- not an "only" reference... */ tWindowView*
  __callocator window_append_view(tWindow* window, tResourceRequest* request)
/* prepares a new view in the given window and removes any old "tail" of views
   in that window */
{ tWindowView *current_view = window->current_view,
    *new_view = memory_allocate(sizeof(tWindowView));
  new_view->window = window; new_view->request = request;
#if !TGC_IS_GRAPHICS
  new_view->aecur = new_view->former_ae = INVALID_AE;
#endif
  request->window_view = new_view;

  if (current_view != NULL) /* wasn't a new (empty) window */
  { const tWindowView* tail = current_view->next;
    if (tail != NULL) deallocate_windowviewlist(tail); /* remove old tail */
    current_view->next = new_view;
    new_view->prev = current_view;
  }
  window->current_view = new_view;
  if (window->first_view == NULL) window->first_view = new_view;
  return(new_view);
}

#if TGC_IS_GRAPHICS

static my_inline tLinenumber current_view_height(void)
{ return(LINES); /* FIXME! */
}

static my_inline void calculate_minmaxrow(const tWindow* window __cunused,
  short* min, short* max)
{ *min = 0; *max = LINES - 1; /* FIXME! */
}

#else

static my_inline tLinenumber current_view_height(void)
{ short retval;
  if (visible_window[1] == NULL) /* currently only one window visible */
  { retval = LINES - 2; }
  else if (current_window_index == 0) /* upper half of screen */
  { retval = VMIDDLE /* - 1 + 1 */; }
  else /* lower half of screen */
  { retval = (LINES - 2) - (VMIDDLE + 1); }
  if (retval < 1) retval = 1;
  return((tLinenumber) retval);
}

static void calculate_minmaxrow(const tWindow* window, short* _min,short* _max)
{ short min = 0, max = MAX(LINES - 3, 0);
  if (window == visible_window[0])
  { if (visible_window[1] != NULL) /* only draw in top-half of screen */
      max = VMIDDLE - 1;
  }
  else if (window == visible_window[1]) /* only draw in bottom-half of screen*/
    min = VMIDDLE + 1;
  *_min = min; *_max = max;
}

static const tActiveElementCoordinates*
  find_visible_aec(const tWindowView* view, tActiveElementNumber _ae)
/* tries to find visible coordinates for the given active element */
{ const tActiveElementCoordinates *retval = NULL,
    *aec = view->active_element[_ae].aec;
  const tWindow* window;
  tLinenumber yvis1, yvis2;
  short minrow, maxrow;

  if (aec == NULL) goto out; /* no coordinates, nothing visible */
  window = view->window;
  calculate_minmaxrow(window, &minrow, &maxrow);
  yvis1 = view->origin_y; yvis2 = yvis1 + (maxrow - minrow);
  while (aec != NULL)
  { tLinenumber y = aec->y;
    if ( (y >= yvis1) && (y <= yvis2) ) { retval = aec; break; }
    aec = aec->next;
  }
  out:
  return(retval);
}

static my_inline tBoolean is_ae_visible(const tWindowView* view,
  tActiveElementNumber _ae)
{ return(cond2boolean(find_visible_aec(view, _ae) != NULL));
}

static tActiveElementNumber next_visible_ae(const tWindowView* view,
  tActiveElementNumber _ae)
{ tActiveElementNumber retval = INVALID_AE, l = view->aenum;
  while (++_ae < l) { if (is_ae_visible(view, _ae)) { retval = _ae; break; } }
  return(retval);
}

static tActiveElementNumber previous_visible_ae(const tWindowView* view,
  tActiveElementNumber _ae)
{ tActiveElementNumber retval = INVALID_AE;
  while (--_ae >= 0) { if (is_ae_visible(view, _ae)) { retval = _ae; break; } }
  return(retval);
}

#endif

static my_inline void cursor_reset_position(void)
{
#if TGC_IS_CURSES
  if (key_handling_mode == khmLineInput)
    (void) move(lid_area(row), lid_area(colcurr));
  else (void) move( (current_window_index == 1) ? (VMIDDLE + 1) : 0, 0);
#endif
}

#if !TGC_IS_GRAPHICS

static my_inline short line2row(const tWindowView* view, tLinenumber y)
{ short row = y - view->origin_y;
  if (view->window == visible_window[1]) row += VMIDDLE + 1;
  return(row);
}

static void do_activate_element(tWindowView* view, tActiveElementNumber _ae,
  tBoolean active)
{ const tActiveElementCoordinates* aec = view->active_element[_ae].aec;
  attr_t attribute;
  short minrow, maxrow, row;
  tLinenumber y1, y2;
  if (aec == NULL) goto finish;
  y1 = view->origin_y; y2 = y1 + current_view_height() - 1;
  attribute = (active ? A_REVERSE : A_UNDERLINE);
  calculate_minmaxrow(view->window, &minrow, &maxrow);
  while (aec != NULL)
  { tLinenumber y = aec->y;
    short x1, x2, count;
    if (y < y1) goto do_next;
    else if (y > y2) break; /* done */
    row = line2row(view, y);
    x1 = aec->x1; x2 = aec->x2; count = x2 - x1 + 1; (void) move(row, x1);
    while (count-- > 0)
      (void) addch((inch() & ~(A_REVERSE | A_UNDERLINE)) | attribute);
    do_next:
    aec = aec->next;
  }
  finish:
  if (active) view->aecur = view->former_ae = _ae;
  else view->aecur = INVALID_AE;
}

static my_inline void activate_element(tWindowView* view,
  tActiveElementNumber _ae)
{ tActiveElementNumber old = view->aecur;
  if (old != INVALID_AE) do_activate_element(view, old, falsE); /* deact. old*/
  do_activate_element(view, _ae, truE); /* activate new */
  cursor_reset_position();
}

#endif /* #if TGC_IS_GRAPHICS */

static void show_message(const char* msg, tBoolean is_error)
{
#if TGC_IS_GRAPHICS
  tGraphicsWidget* widget = current_window->message;
  gtk_label_set_text(GTK_LABEL(widget), msg);
#else
  if (!is_bottom_occupied)
  { if (is_error) my_set_color(cpnRed);
    /*(void)*/ mvaddnstr(LINES - 1, 0, msg, COLS - 1);
    if (is_error) my_set_color(cpnDefault);
    (void) clrtoeol(); cursor_reset_position();
  }
#endif
  debugmsg("UI message: "); debugmsg(msg); debugmsg(strNewline);
}

static void show_message_osfailure(int errnum)
{ const char* text;
  if (errnum == 0) text = _("Failed!");
  else
  { const char* e = my_strerror(errnum);
    if (strlen(e) > STRBUF_SIZE / 2) e = _(strUnknown); /* can't be serious */
    sprint_safe(strbuf, _("Failed! (error #%d, %s)"), errnum, e);
    text = strbuf;
  }
  show_message(text, truE);
}

#if (!OPTION_MENUS) || (!CONFIG_SESSIONS) || (!CONFIG_JUMPS) || (!CONFIG_DO_TEXTMODEMOUSE) || (!OPTION_EXECEXT)
static void fwdact(const char* str)
{ char* spfbuf;
  if (*str == 'F') str++;
  my_spf(strbuf, STRBUF_SIZE, &spfbuf, _(strFwdact), str, strEmpty);
  spfbuf[0] = my_toupper(spfbuf[0]);
  show_message(spfbuf, truE);
  my_spf_cleanup(strbuf, spfbuf);
}
#endif

#if !OPTION_MENUS
static my_inline void menus_were_disabled(void)
{ fwdact(_("Fmenus"));
}
#endif

static void redraw_message_line(void)
{
#if TGC_IS_CURSES
  { const char* str = NULL;
    tBoolean is_error = falsE;
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      if (view != NULL)
      { const char* li = view->last_info;
        if (li != NULL)
        { if (*li != 0) is_error = truE;
          str = li + 1;
        }
      }
    }
    show_message(null2empty(str), is_error);
  }
#endif
}

static my_inline void window_show_info(const __sunused tWindow* window
  __cunused)
{ /* IMPLEMENTME: use stuff from window_redraw()! */
  redraw_message_line();
}

#include "renderer.c"

typedef struct
{ char* line;
  int fd;
} tContentDumpData;

static void content_dump_line(tRendererData* data)
{ const tContentDumpData* cdd = data->line_callback_data;
  const tRendererElement* element = data->element;
  char* line = cdd->line;
  size_t count = 0, cnt;
  while (element != NULL)
  { const size_t textcount = element->textcount;
    if (textcount > 0)
    { if (element->is_spacer)
      { for (cnt = 0; cnt < textcount; cnt++) line[count++] = ' '; }
      else
      { const tRendererText* text = element->text;
        for (cnt = 0; cnt < textcount; cnt++) line[count++] = text[cnt];
      }
    }
    element = element->next;
  }
  line[count++] = '\n';
  if (my_write(cdd->fd, line, count) != (ssize_t) count)
  { if (is_environed) show_message_osfailure(errno);
    else fatal_error(errno, _("can't dump line"));
    data->flags |= rdfCallerDone;
  }
}

static void content_dump_to_fd(tResource* resource, int fd,
  tWindowViewFlags wvf)
{ tRendererData data;
  tContentDumpData cdd;
  const size_t line_width = 80;
  my_memclr_var(data);
  data.line_callback = content_dump_line; data.line_callback_data = &cdd;
  data.ra = raLayout; data.resource = resource; data.line_width = line_width;
  data.flags = rdfAlignment;
  if ( (wvf & wvfEnforceHtml) ||
       ( (!(wvf & wvfSource)) && (resource->kind == rckHtml) ) )
  { data.flags |= rdfHtml; }
  cdd.fd = fd; cdd.line = __memory_allocate(line_width + 5);
  renderer_run(&data);
  memory_deallocate(cdd.line);
}

static void content_do_dump(tResource* resource, const char* filename,
  tBoolean may_overwrite, tWindowViewFlags wvf)
{ int fd, cflags = O_CREAT | O_TRUNC | O_WRONLY;
  struct stat statbuf;
  if (!may_overwrite) cflags |= O_EXCL;
  fd = my_create(filename, cflags, S_IRUSR | S_IWUSR);
  if (fd < 0)
  { failed:
    if (is_environed) show_message_osfailure(errno);
    else fatal_error(errno, _("can't dump to file"));
    return;
  }
  if (my_fstat(fd, &statbuf) != 0)
  { int e; close_bad: e = errno; my_close(fd); errno = e; goto failed; }
  if (!S_ISREG(statbuf.st_mode))
  { if (S_ISDIR(statbuf.st_mode)) errno = EISDIR;
    goto close_bad;
  }
  content_dump_to_fd(resource, fd, wvf);
  my_close(fd);
}

static my_inline void content_dump(const char* filename,tBoolean may_overwrite)
{ const tWindowView* view = current_window->current_view;
  content_do_dump(view->request->resource, filename, may_overwrite,
    view->flags);
}

my_enum1 enum
{ wrfNone = 0, wrfFound = 0x01
} my_enum2(unsigned char) tWindowRedrawingFlags;

typedef struct
{ tLinenumber origin_y, currline, highest_backward_line;
  size_t searchlen;
  short row, maxrow, mincol;
  tWindowRedrawingTask task;
  tWindowRedrawingFlags flags;
} tWindowRedrawingData;

static void __window_redraw_line(tWindowRedrawingData* wrd,
  tRendererData* data)
/* actually draws a line on the screen */
{ const tRendererElement* element = data->element;
  move(wrd->row, wrd->mincol);
  while (element != NULL)
  { const tRendererText* text = element->text;
    const tRendererAttr* attr = element->attr;
    size_t count = 0, textcount = element->textcount;
    if (element->is_spacer) { while (count++ < textcount) addch(' '); }
    else
    { while (count < textcount)
      { tRendererAttr a = 0;
        if (text != NULL) a |= (tRendererAttr) text[count];
        if (attr != NULL) a |= attr[count];
        if (a == 0) a = (tRendererAttr) ' '; /* "should not happen" */
        addch(a); count++;
      }
    }
    element = element->next;
  }
  clrtoeol(); wrd->row++;
  if (wrd->row > wrd->maxrow) data->flags |= rdfCallerDone;
}

static void window_redraw_line__standard(tRendererData* data)
{ tWindowRedrawingData* wrd = data->line_callback_data;
  if (wrd->currline < wrd->origin_y)
  { wrd->currline++;
    if (wrd->currline >= wrd->origin_y) data->flags &= ~rdfVirtual;
  }
  else { wrd->currline++; __window_redraw_line(wrd, data); }
}

static tBoolean search_in_line(const tRendererElement* element,
  size_t searchlen)
/* returns whether it found something */
{ tBoolean retval = falsE;
  while (element != NULL)
  { const char* text;
    size_t textcount = element->textcount;
    if ( (!(element->is_spacer)) && ( (text = element->text) != NULL ) &&
         ( (textcount = element->textcount) >= searchlen ) &&
         (my_strncasestr(text, search_string, textcount)) )
    { retval = truE; break; }
    element = element->next;
  }
  return(retval);
}

static void window_redraw_line__search(tRendererData* data)
{ tWindowRedrawingData* wrd = data->line_callback_data;
  if (wrd->currline <= wrd->origin_y)
  { wrd->currline++;
    if (wrd->currline > wrd->origin_y) data->flags &= ~rdfVirtual;
  }
  else if (wrd->flags & wrfFound)
  { wrd->currline++; do_redraw: __window_redraw_line(wrd, data); }
  else
  { wrd->currline++;
    if (search_in_line(data->element, wrd->searchlen))
    { if (data->view != NULL) data->view->origin_y = wrd->currline - 1;
      wrd->flags |= wrfFound; goto do_redraw;
    }
  }
}

static void window_redraw_line__search_backward(tRendererData* data)
{ tWindowRedrawingData* wrd = data->line_callback_data;
  wrd->currline++;
  if (wrd->currline < wrd->origin_y)
  { if (search_in_line(data->element, wrd->searchlen))
    { wrd->highest_backward_line = wrd->currline - 1; wrd->flags |= wrfFound; }
  }
  else data->flags |= rdfCallerDone;
}

static void window_redraw_line__count(tRendererData* data)
{ tWindowRedrawingData* wrd = data->line_callback_data;
  wrd->currline++;
}

static void window_redraw_contents(tWindowView* view, tResource* resource,
  const short mincol, const short minrow, const short maxcol,
  const short maxrow, const tWindowRedrawingTask task)
{ tRendererData data;
  tWindowRedrawingData wrd;
  const tBoolean is_search =
    cond2boolean( (task == wrtSearch) || (task == wrtSearchBackward) );
#if !TGC_IS_GRAPHICS
  tActiveElementNumber _ae SHUT_UP_COMPILER(INVALID_AE);
#endif

  my_memclr_var(data); my_memclr_var(wrd);
  data.line_callback_data = &wrd;
  data.line_callback = window_redraw_line__standard; /* default callback */
  if (is_search)
  { wrd.searchlen = strlen(search_string);
    if (task == wrtSearch) data.line_callback = window_redraw_line__search;
    else if (task == wrtSearchBackward)
      data.line_callback = window_redraw_line__search_backward;
  }
  else if (task == wrtToEnd) data.line_callback = window_redraw_line__count;
  data.ra = raLayout; data.resource = resource; data.view = view;
  data.line_width = maxcol - mincol + 1;
  data.flags = rdfAttributes | rdfColors | rdfAlignment | rdfAe;
#if TGC_IS_GRAPHICS
  data.flags |= rdfGraphical;
#endif

  if (view == NULL) { /* wrd.origin_y = 0; -- superfluous */ }
  else
  { tWindowViewFlags wvf = view->flags;
    wrd.origin_y = view->origin_y;
#if !TGC_IS_GRAPHICS
    _ae = view->aecur;
    if (_ae != INVALID_AE) do_activate_element(view, _ae, falsE);
#endif
    if (wvf & (wvfEnforceHtml | wvfSource)) /*user explicitly wants something*/
    { if (wvf & wvfEnforceHtml) { set_html: data.flags |= rdfHtml; }
      goto kind_done;
    }
  }
  if ( (resource != NULL) && (resource->kind == rckHtml) )
    goto set_html; /* document kind knowledge, no explicit user preference */
  kind_done: {}

  wrd.row = minrow; wrd.maxrow = maxrow; wrd.mincol = mincol; wrd.task = task;
  switch (task)
  { case wrtSearch: case wrtToEnd: data.flags |= rdfVirtual; break;
    case wrtSearchBackward: /* nothing */ break;
    default: if (wrd.origin_y > 0) { data.flags |= rdfVirtual; } break;
  }

  renderer_run(&data);

  if (is_search)
  { tBoolean did_find = cond2boolean(wrd.flags & wrfFound);
    if ( (did_find) && (task == wrtSearchBackward) && (view != NULL) )
      view->origin_y = wrd.highest_backward_line;
    /* IMPROVEME: don't redraw always, only if necessary! */
    window_redraw_contents(view, resource, mincol, minrow, maxcol, maxrow,
      wrtRedrawRecursive); /* risk a recursion :-) */
    if (did_find) show_message(_("Found!"), falsE);
    else show_message(_("Not found!"), truE);
    goto out; /* rest was done in recursively called instance */
  }
  else if (task == wrtToEnd)
  { if (view != NULL)
    { short height = maxrow - minrow;
      tLinenumber num = wrd.currline;
      view->origin_y = ( (num > height) ? (num - height) : 0 );
    }
    window_redraw_contents(view, resource, mincol, minrow, maxcol, maxrow,
      wrtRedrawRecursive);
    goto out;
  }
  if (view != NULL)
  { if (wrd.row > wrd.maxrow) view->flags |= wvfScreenFull;
    else view->flags &= ~wvfScreenFull;
  }
  while (wrd.row <= wrd.maxrow) {move(wrd.row, mincol); clrtoeol(); wrd.row++;}

  if (view != NULL)
  {
#if !TGC_IS_GRAPHICS
    if (_ae == INVALID_AE) _ae = view->former_ae;
    if (_ae != INVALID_AE)
    { if (!is_ae_visible(view, _ae))
      { if (view->flags & wvfScrollingUp) _ae = previous_visible_ae(view, _ae);
        else _ae = next_visible_ae(view, _ae);
      }
    }
    else
    { _ae = next_visible_ae(view, -1); /* find the first visible ae */
      if ( (_ae != INVALID_AE) && (view->flags & wvfScrollingUp) )
      { /* proceed to the _last_ visible active element */
        tActiveElementNumber _ae2, l = view->aenum;
        for (_ae2 = _ae + 1; _ae2 < l; _ae2++)
        { if (is_ae_visible(view, _ae2)) _ae = _ae2; }
      }
    }
    if (_ae != INVALID_AE) activate_element(view, _ae);
#endif
    view->flags &= ~wvfScrollingUp;
  }
#if 0
  /* some texts for the very next versions...
     _("looking for fragment"), _("fragment not found"), _("account"),
     _("disk cache file"), _("Do not edit manually while retawq is running!"),
     _("Upload local file: "), _("Upload to FTP URL: "),
     _("Resume upload? [(r)esume, (o)verwrite%s]"), _(", (s)kip"),
     _("File size: local %d, remote %d")
  */
#endif
  out: {}
}

#if OPTION_TLS
static tBoolean tls_errtext(const tResource* resource, /*@out@*/ char* dest)
{ tBoolean retval;
  tTlsError te = resource->tls_error;
  if (!is_tls_error_expressive(te)) retval = falsE;
  else
  { sprint_safe(dest, _("TLS error - %s"), _(strTlsError[te])); retval=truE; }
  return(retval);
}
#endif

static void __window_redraw(const tWindow* window, tWindowRedrawingTask task)
{ short minrow, maxrow, mincol = 0, maxcol = COLS - 1, titlerow;
#if TGC_IS_CURSES
  short spacecount;
  tBoolean called_spf;
#else
  tGraphicsWindow* ww = window->ww;
  char* window_title = unconstify(strVersion);
#endif
  tWindowView* view = window->current_view;
  tResourceRequest* request = NULL;
  tResource* resource = NULL;
  const char *message, *temp0;
  char* temp;
  tBoolean is_error;

#if TGC_IS_GRAPHICS
  current_window = __unconstify(tWindow*, window);
#endif
#if CONFIG_DEBUG
  sprint_safe(strbuf, "window_redraw(): window=%p, view=%p\n", window, view);
  debugmsg(strbuf);
#endif

  /* Prepare */

  calculate_minmaxrow(window, &minrow, &maxrow);
  titlerow = maxrow + 1;
  if (view != NULL) { request = view->request; resource = request->resource; }

  /* Draw title line */

#if TGC_IS_GRAPHICS

  if (view == NULL) temp = unconstify_or_(strBracedNewWindow);
  else
  { char* s = strbuf;
    if (view->html_title != NULL) window_title = unconstify(view->html_title);
    if (view->title != NULL)
      gtk_label_set_text(GTK_LABEL(window->location), view->title);
#if OPTION_COOKIES
    if ( (resource != NULL) && (resource->flags & rfCookieAnything) )
      s += sprint_safe(s, strPercs, _(strCookies));
#endif
    if (view->flags & wvfHandledRedirection)
      s += sprint_safe(s, strPercsDash, _(strRedirection));
    else if (view->flags & wvfSource)
      s += sprint_safe(s, strPercsDash, _(strSource));
    *s = '\0';
    temp = strbuf;
  }
  gtk_window_set_title(ww, window_title);
  gtk_label_set_text(GTK_LABEL(window->info), temp);

#else /* #if TGC_IS_GRAPHICS */

  if (view == NULL)
  { temp = unconstify_or_(strBracedNewWindow); called_spf = falsE; }
  else
  { const char *html_title = view->html_title, *title = view->title;
#define dashify(cond, str) \
  ( (cond) ? str : strEmpty ), ( (cond) ? strSpacedDash : strEmpty )
    my_spf(strbuf, STRBUF_SIZE, &temp,
#if OPTION_COOKIES
/*A*/ "%s"
#endif
/*BCDE*/ "%s%s%s%s%s%s%s"
    ,
#if OPTION_COOKIES
/*A*/ ( ( (resource != NULL) && (resource->flags & rfCookieAnything) )
/*A*/   ? _(strCookies) : strEmpty ), /* RFC2965, 6.1, *2 */
#endif
/*B*/ dashify((view->flags & wvfHandledRedirection), _(strRedirection)),
/*C*/ dashify((view->flags & wvfSource), _(strSource)),
/*D*/ dashify((html_title != NULL), html_title),
/*E*/ null2empty(title)
    );
#undef dashify
    called_spf = truE;
  }
  (void) move(titlerow, 0); (void) attron(A_REVERSE);
  (void) addnstr(temp, COLS - 1);
  if (called_spf) my_spf_cleanup(strbuf, temp);
  spacecount = COLS - stdscr->_curx;
  while (spacecount-- > 0) (void) addch(' '); /* IMPROVEME? */
  (void) attroff(A_REVERSE);

#endif /* #if TGC_IS_GRAPHICS */

  /* Draw message */

#if TGC_IS_CURSES
  if ( (window != current_window) && (view == NULL) ) /* nothing to do */
    return;
#endif
  is_error = falsE;
  if (resource != NULL)
  { tResourceState rs = resource->state;
    tServerStatusCode ssc = resource->server_status_code;
#if OPTION_TLS
    tConnection* conn;
#endif

    if (rs == rsError)
    { tResourceError re = resource->error;
#if OPTION_TLS
      if ( (re == reTls) && (tls_errtext(resource, strbuf2)) ) temp = strbuf2;
      else
#endif
      temp = unconstify_or_(strResourceError[re]);
      is_error = truE;
    }
#if OPTION_TLS
    else if ( ( (conn = resource->cconn) != NULL ) &&
              (conn->flags & cnfTlsHandshaking) )
      temp = unconstify_or_("TLS handshaking");
#endif
    else if (rs == rsReading)
    { size_t count = resource->bytecount;
      if (count < 2) temp = unconstify_or_("Waiting for reply");
      else
      { size_t nominal = resource->nominal_contentlength;
        if ( (nominal != UNKNOWN_CONTENTLENGTH) && (count < nominal) )
        { sprint_safe(strbuf3, _("of %d "), localized_size(nominal));
          temp0 = strbuf3;
        }
        else temp0 = strEmpty;
        sprint_safe(strbuf2, _("Received %d %sbytes"), localized_size(count),
          temp0);
        temp = strbuf2;
      }
    }
    else if ( (rs == rsComplete) && ( (resource->content == NULL) ||
              (CONTENTBLOCK_USEDSIZE(resource->content) < 1) ) )
      temp = unconstify_or_("Document empty");
    else if (resource_ui_conn_ip(resource, strbuf2, STRBUF_SIZE / 2))
    { sprint_safe(strbuf3, _("Connecting to %s"), strbuf2); temp = strbuf3; }
    else temp = unconstify_or_(strResourceState[rs]);
#if CONFIG_DEBUG
    sprint_safe(strbuf,
      "%s [res=%p,rs=%d,re=%d,rch=%d,rf=%d,ths=%d,ssc=%d,ssi=*%s*,v=%d]", temp,
      resource, rs, resource->error, resource->handshake, resource->flags,
      resource->tlheaderstate, ssc, ssc2info(resource->protocol, ssc),
      resource->version);
    temp = strbuf;
#else
    if (ssc != 0)
    { tResourceProtocol rp = resource->protocol;
      const char *ssi = ssc2info(rp, ssc), *ssi_sep = ( (ssi == strEmpty) ?
        strEmpty : strSpacedDash ), *whose;
#if OPTION_LOCAL_CGI
      if (rp == rpLocalCgi) whose = _("script status: ");
      else
#endif
      { whose = _("server status: "); }
      sprint_safe(strbuf, "%s (%s%d%s%s)", temp, whose, ssc, ssi_sep, ssi);
      temp = strbuf;
    }
#endif
    message = temp;
  }
  else if (request != NULL)
  { tResourceRequestState rrs = request->state;
    switch (rrs)
    { case rrsError:
        message = unconstify_or_(strResourceError[request->error]);
        is_error = truE; break;
      case rrsUnknown: message = strEmpty; break;
      default: message = unconstify_or_(strResourceRequestState[rrs]); break;
    }
  }
  else
  { message = strEmpty;
  }
#if TGC_IS_CURSES
  if (window == current_window)
#endif
  { show_message(message, is_error); }
  if (view != NULL)
  { size_t msgsize = strlen(message) + 1;
    temp = __memory_allocate(msgsize + 1); /* (one extra byte for flags) */
    *temp = boolean2bool(is_error);
    my_memcpy(temp + 1, message, msgsize);
    __dealloc(view->last_info); view->last_info = temp;
  }

  /* Draw contents */

  window_redraw_contents(view, resource, mincol, minrow, maxcol, maxrow, task);

  /* Finish */

#if TGC_IS_GRAPHICS
  current_window = NULL;
#endif
  cursor_reset_position();
  if ( (view != NULL) && (resource != NULL) ) test_redirection(view, resource);
}

static void interpret_as_html(void)
{ tWindowView* view = current_window->current_view;
  if (view != NULL)
  { view->flags &= ~wvfSource; view->flags |= wvfEnforceHtml;
    window_redraw(current_window);
  }
}

static void window_redraw_all(void)
/* redraws all visible windows */
{
#if TGC_IS_CURSES
  window_redraw(visible_window[0]);
  if (visible_window[1] != NULL) window_redraw(visible_window[1]);
#endif
}

static void window_remove(tWindow* window)
{ tWindowView* view;

  /* detach the window from the window list */
  if (windowlisthead == window) windowlisthead = window->next;
  if (windowlisttail == window) windowlisttail = window->prev;
  if (window->prev != NULL) window->prev->next = window->next;
  if (window->next != NULL) window->next->prev = window->prev;

  /* deallocate all views of the window and the window structure itself */
  view = window->first_view;
  if (view != NULL) deallocate_windowviewlist(view);
  memory_deallocate(window);
#if CONFIG_DEBUG
  sprint_safe(strbuf, "removed window at %p\n", window);
  debugmsg(strbuf);
#endif
#if TGC_IS_GRAPHICS
  window_counter--;
  if (window_counter == 0) do_quit();
#endif
}

static void window_hide(tWindow* window, unsigned char removal)
{ /* <removal>: 0 = must not; 1 = may; 2 = must */
#if TGC_IS_GRAPHICS
  gtk_widget_hide(GTK_WIDGET(window->ww));
  if (removal) window_remove(window);
#else
  if (window == visible_window[0])
  { visible_window[0] = visible_window[1];
    if (visible_window[0] == NULL)
    { /* Make sure we always have at least one window on the screen */
      tWindow* w = window->next;
      if (w == NULL) w = windowlisthead;
      if ( (w == NULL) || (w == window) ) w = window_create();
      visible_window[0] = w;
    }
    goto xy;
  }
  else if (window == visible_window[1])
  { xy:
    if ( (removal == 2) ||
         ( (removal == 1) && (window->current_view == NULL) ) )
    { window_remove(window); }
    visible_window[1] = NULL;
    current_window_index = 0;
    window_redraw_all();
  }
  /* "else": nothing to do */
#endif
}

#if !CONFIG_JAVASCRIPT
#define javascript_handle_event(kind, _ae) do { } while (0)
#else
static void javascript_handle_event(tJavascriptEventKind kind,
  tActiveElementNumber _ae)
{ tWindowView* view = current_window->current_view;
  const tJavascriptEventHandler* eh = view->request->resource->aebase[_ae].eh;
  while (eh != NULL)
  { if (eh->kind == kind)
    { const tJavascriptCode* code = eh->code;
      if (code != NULL) javascript_execute(view, _ae, code);
      break;
    }
    eh = eh->next;
  }
}
#endif

static my_inline void request_set_error(tResourceRequest* request,
  tResourceError error)
{ request->state = rrsError;
  request->error = error;
}

static void queue_request(tResourceRequest* request,
  tResourceRequestAction action)
/* tries to feed the request to the resource handler */
{ request->action = action; request->state = rrsPreparedByMain;
#if CONFIG_DEBUG
  { char* spfbuf;
    my_spf(strbuf, STRBUF_SIZE, &spfbuf,
      "Queueing request: %d - *%s* - *%s* - *%s*\n", request->protocol,
      null2empty(request->uri), null2empty(request->query),
      null2empty(request->post));
    debugmsg(spfbuf);
    my_spf_cleanup(strbuf, spfbuf);
  }
#endif
  resource_callback(request, mckStart);
}

static const char* my_getcwd(void)
{
#if !HAVE_GETCWD
  return(strSlash); /* CHECKME! */
#else
  static const char* cwd = NULL;
  if (cwd == NULL) /* not yet calculated */
  { char buf[1024];
    if ( (getcwd(buf, sizeof(buf) - 5) != buf) || (buf[0] == '\0') )
      cwd = strSlash;
    else
    { if (buf[strlen(buf) - 1] != '/') strcat(buf, strSlash);
      cwd = my_strdup(buf);
    }
  }
  return(cwd);
#endif
}


/* submit_form() buffering and conversion */

static char* sfbuf;
static char sfconv[4];
static size_t sfbuf_size, sfbuf_maxsize;

static my_inline void sfbuf_reset(void)
{ sfbuf = NULL;
  sfbuf_size = sfbuf_maxsize = 0;
}

static void __sfbuf_add(const char ch)
/* append character unconverted */
{ if (sfbuf_maxsize <= sfbuf_size)
  { sfbuf_maxsize += 100;
    sfbuf = memory_reallocate(sfbuf, sfbuf_maxsize);
  }
  sfbuf[sfbuf_size++] = ch;
}

static my_inline void __sfbuf_add_str(const char* str)
/* append string unconverted */
{ char ch;
  while ( (ch = *str++) != '\0' ) __sfbuf_add(ch);
}

static void sfbuf_add_str(const char* str)
/* append string converted */
{ char ch;
  while ( (ch = *str++) != '\0' )
  { /* append character converted */
    { if (my_isalnum(ch)) __sfbuf_add(ch);
      else if (ch == ' ') __sfbuf_add('+');
      else if (ch == '\n') __sfbuf_add_str("%0d%0a");
      else if (ch != '\r')
      { sfconv[1] = strHexnum[(ch >> 4) & 15]; sfconv[2] = strHexnum[ch & 15];
        __sfbuf_add_str(sfconv);
      }
    }
  }
}


/* URI parsing */

static const struct
{ const char* scheme; /* (sorted in alphabetical order) */
  tResourceProtocol rp;
} scheme2rp[] =
{ { strAbout, rpAbout },
#if CONFIG_DEBUG
  { strCvs, rpCvs },
#endif
  { strFile, rpLocal },
  { strFinger, __rpFinger },
  { strFtp, __rpFtp },
  { strFtps, __rpFtps },
#if CONFIG_DEBUG
  { strGopher, rpGopher },
#endif
  { strHttp, rpHttp },
  { strHttps, __rpHttps },
#if CONFIG_DEBUG
  { strInfo, rpInfo },
#endif
  { strJavascript, __rpJavascript },
  { strLocal, rpLocal },
  { strLocalCgi, __rpLocalCgi },
  { strMailto, __rpMailto },
  { strNews, __rpNntp },
  { strNntp, __rpNntp },
  { strPop, __rpPop },
  { strPops, __rpPops }
};

static one_caller tMbsIndex do_lookup_rp(const char* scheme)
{ my_binary_search(0, ARRAY_ELEMNUM(scheme2rp) - 1,
    streqcase3(scheme, scheme2rp[idx].scheme), return(idx))
  /* case-insensitivity: RFC2396 (3.1) and RFC2616 (3.2.3) -- or was that
     casein-sensitivity? :-) */
}

static one_caller tResourceProtocol lookup_rp(const char* scheme)
{ tResourceProtocol retval;
  tMbsIndex idx = do_lookup_rp(scheme);
  if (idx < 0) retval = rpUnknown;
  else retval = scheme2rp[idx].rp;
  return(retval);
}

static char* parse_uri_part(const char* uri, unsigned char part)
{ static const char separator[] = ":/?#";
  const char* const sep = separator + part;
  char uch, sch; /* URI character, separator character */
  /* The following code does very much the same as strpbrk(), but that function
     isn't portable. */
  while ( (uch = *uri) != '\0' )
  { const char* temp = sep;
    while ( (sch = *temp++) != '\0' )
    { if (uch == sch) return(unconstify(uri)); }
    uri++;
  }
  return(NULL);
}

static void setup_authority(/*const*/ char** _uri,
  /*@out@*/ /*const*/ char** _authority)
{ /*const*/ char *uri = *_uri, *pos = parse_uri_part(uri, 1), *authority;
  if (pos != NULL) { authority = my_strndup(uri, pos - uri); uri = pos; }
  else { authority = my_strdup(uri); TO_EOS(uri); }
  *_uri = uri; *_authority = authority;
}

static __sallocator char* __callocator finalize_path(const char* _path)
/* transforms a path to a standard representation by removing "superfluous"
   components */
{ char ch, *path = my_strdup(_path), *temp, *dest, **component;
  size_t num, maxnum, count, orig_len = strlen(path);
  tBoolean is_first, found_tilde = falsE, is_abs=cond2boolean(path[0] == '/'),
    is_dir = cond2boolean( (orig_len > 0) && (path[orig_len - 1] == '/') );

  /* split the path into components */
  temp = path; component = NULL; num = maxnum = 0;
  splitloop:
  while (*temp == '/') temp++;
  if (*temp != '\0') /* got a component */
  { if (num >= maxnum)
    { maxnum += 16;
      component = memory_reallocate(component, maxnum * sizeof(char*));
    }
    component[num++] = temp;
    while ( (*temp != '/') && (*temp != '\0') ) temp++;
    if (*temp != '\0') { *temp++ = '\0'; goto splitloop; } /* look for more */
  }

  /* remove superfluous components */
#define rc(x) *(component[(x)]) = '\0'
  for (count = 0; count < num; count++)
  { const char* str = component[count];
    if (!strcmp(str, strDot)) rc(count);
    else if (!found_tilde)
    { if (!strcmp(str, strDoubleDot))
      { rc(count);
        if (count > 0) /* try to remove "the preceding" component */
        { size_t cnt = count - 1;
          while (1)
          { if (*(component[cnt]) != '\0') { rc(cnt); break; } /* done */
            if (cnt == 0) break; /* no removable preceding component */
            cnt--;
          }
        }
      }
      else if (!strcmp(str, "~"))
      { /* e.g. in FTP it should be possible to do "ftp://foo.org/~/../blah",
           which can only be interpreted by the server, not by us */
        found_tilde = truE;
      }
    }
  }
#undef rc

  /* build the new path */
  dest = path; is_first = truE;
  if (is_abs) *dest++ = '/';
  for (count = 0; count < num; count++)
  { const char* src = component[count];
    if (*src == '\0') continue; /* removed */
    if (is_first) is_first = falsE;
    else *dest++ = '/';
    while ( (ch = *src++) != '\0' ) *dest++ = ch;
  }
  if (is_dir)
  { if ( (dest == path) || ( (dest > path) && (*(dest - 1) != '/') ) )
      *dest++ = '/';
  }
  *dest = '\0';

  __dealloc(component);
  return(path);
}

#define __cleanup_path \
  if (must_dealloc_path) { memory_deallocate(path); must_dealloc_path = falsE;}

static void parse_uri(const char* _uri, tResourceRequest* request,
  const tResourceRequest* const orig_referrer, /*@out@*/ char** _fragment,
  const char* extra_query)
/* RFC2396 says: ^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?
   Simple in theory, but in practice it's much more complicated. Just to
   mention a few examples:
   - Colons can appear in four places in a URI: right after the scheme string,
     as a separator in "username:password", as a separator within a numerical
     IPv6 hostname, and as a portnumber indicator.
   - There aren't only absolute URIs to be handled, but also relative and
     abbreviated URIs.
   To distinguish all these correctly, we have to do much more than a trivial
   pattern matching. Let's thank all specification developers for such
   bogosities. (I've looked at several implementations, and none of them seems
   to handle all cases the right way. And the endless comments in this function
   might also be an indication of bad URI design concepts, similar to the HTTP
   cookie mess...)
*/
{ static const char strWwwDot[] = "www.", strFtpDot[] = "ftp.",
    strNewsDot[] = "news.";
  const tResourceRequest* referrer = orig_referrer;
  tResourceProtocol rp = rpUnknown, wouldbe_rp;
  char *uristart, *uri, *pos;
  char *scheme, *authority, *hostname, *port, *path, *query, *fragment;
  const char *username, *password;
  char *complete_query, *spfbuf_uri;
  tPortnumber portnumber = 0;
  tBoolean must_dealloc_path, must_dealloc_complete_query, ua, iph;

  scheme = authority = hostname = port = path = query = fragment = NULL;
  username = password = NULL;
  must_dealloc_path = must_dealloc_complete_query = falsE;

  /* Prepare the URI */

  { char* dest = uristart = uri = __memory_allocate(strlen(_uri) + 1);
    const char* src = _uri;
    const unsigned char* test = (const unsigned char*) _uri;
    unsigned char testch;
    while ( (testch = *test++) != '\0' )
    { const char urich = *src++; /* "My first name is Robert." :-) */
      *dest++ = (is_control_char(testch) ? '_' : urich);
    }
    *dest = '\0';
  }

  if (orig_referrer != NULL)
  { char ch = uri[0];
    if ( (ch == '\0') || (ch == '#') ) /* RFC2396, 4.2 */
    { const char* ref_uri = orig_referrer->uri;
      if ( (ref_uri != NULL) && (ref_uri[0] != '\0') )
      { char* new_uri;
        my_spf(NULL, 0, &new_uri, strPercsPercs, ref_uri, uri);
        memory_deallocate(uri); uri = my_spf_use(new_uri);
      }
    }
  }

  /* Look what's explicitly given */

  pos = parse_uri_part(uri, 0);
  if ( (pos != NULL) && (*pos == ':') && (pos > uri) )
  { /* RFC2396 says: scheme = alpha *( alpha | digit | "+" | "-" | "." )
       Test whether it actually _is_ a scheme name or rather an abbreviated URI
       like "foo.org:42" or a numerical IPv6 hostname beginning like
       "[42:43...]". To avoid problems, we don't allow any dots in a scheme
       name - RFC2396 allows dots, but nobody seems to use them. This way, only
       abbreviated URIs like "localhost:42" could still cause problems, and we
       check explicitly for "localhost". These are ugly special-case hacks, but
       all more elegant algorithms seem to be less correct. */
    const char* temp;
    char ch;
    size_t len;
    if ( (pos[1] == '/') && (pos[2] == '/') )
    { /* Found the sequence "://"; accept anything in front of this as a scheme
         string, no matter what it actually is. This mechanism allows users to
         override the below plausibility tests, so they can use even the
         strangest schemes if they want to. (Non-standard URI schemes will be
         configurable in a later version, and "all power to the user" is often
         a good idea anyway.)
      */
      goto do_scheme;
    }
    temp = uri;
    ch = *temp++;
    if (!my_isalpha(ch)) goto no_scheme;
    while (temp < pos)
    { ch = *temp++;
      if ( (!my_isalnum(ch)) && (ch != '+') && (ch != '-') ) goto no_scheme;
    }
    len = pos - uri;
    if ( (len == 9) && (strneqcase(uri, strLocalhost, 9)) ) goto no_scheme;
    do_scheme: scheme = uri; *pos++ = '\0'; uri = pos; /* scheme given */
    no_scheme: {}
  }

  if ( (uri[0] == '/') && (uri[1] == '/') ) /* authority given */
  { uri += 2; setup_authority(&uri, &authority);
  }

  path = uri; /* a path is always given (but may be empty, of course) */

  pos = parse_uri_part(uri, 2);
  if ( (pos != NULL) && (*pos == '?') ) /* query given */
  { *pos++ = '\0'; query = uri = pos; }

  pos = parse_uri_part(uri, 3);
  if ( (pos != NULL) && (*pos == '#') ) /* fragment given */
  { *pos++ = '\0'; fragment = uri = pos; }

  /* Interpret the whole stuff */

  /* calculate protocol and possibly authority: */
  if (scheme != NULL) rp = lookup_rp(scheme);
  else if (orig_referrer != NULL) rp = orig_referrer->protocol;
  else if (authority == NULL)
  { /* The URI can only be an abbreviated thing like "www.foo.org..." or a
       local path. */
    if ( (path[0] == '/') || ( (path[0] == '.') && (path[1] == '/') ) ||
         ( (path[0] == '~') && (path[1] == '/') ) )
    { /* local scheme, no authority necessary */
      set_local: rp = rpLocal;
    }
    else if (strneqcase(path, strWwwDot, 4))
    { set_http: rp = rpHttp;
      handle_abbr_authority: setup_authority(&path, &authority);
    }
    else if (strneqcase(path, strFtpDot, 4))
    { rp = __rpFtp; goto handle_abbr_authority; }
    else if ( (strneqcase(path, strNewsDot, 5)) /* CHECKME! ||
              ( (strneqcase(path, strNews, 4)) &&
                (parse_uri_part(path, 1) == path + 4) ) */ )
    { rp = __rpNntp; goto handle_abbr_authority; }
    else /* last resorts */
    { struct stat statbuf;
      /* might be a local regular file or directory */
      if (my_stat(path, &statbuf) == 0)
      { mode_t mode = statbuf.st_mode;
        if (S_ISREG(mode) || S_ISDIR(mode)) goto set_local;
      }
      /* nothing else worked, so assume it's an HTTP resource */
      goto set_http;
    }
  }

  /* split the authority: */
  if (authority != NULL)
  { pos = my_strchr(authority, '@');
    if (pos == NULL) hostname = authority;
    else
    { *pos++ = '\0'; hostname = pos; username = authority; /* username given */
      pos = my_strchr(username, ':');
      if (pos != NULL) { *pos++ = '\0'; password = pos; } /* password given */
    }

    pos = hostname; /* default start position for the below my_strchr() call */
    if (hostname[0] == '[')
    { /*looks like the beginning of a numeric IPv6 address notation (RFC2732)*/
      char* temp;
      if ( (hostname[1] != '\0') &&
           ( ( temp = my_strchr(hostname + 2, ']') ) != NULL ) )
      { /* yup, it is */
        *temp++ = '\0'; pos = temp; hostname++;
        request->flags |= rrfTryIpv6;
      }
      /* "else": it isn't; we assume that users know what they're doing; we
         don't know :-) so we pass the stuff unchanged */
    }
    pos = my_strchr(pos, ':');
    if (pos != NULL) { *pos++ = '\0'; port = pos; } /* port given */
    if (hostname[0] == '\0') hostname = NULL;
    else
    { if ( (rp == rpUnknown) && (scheme == NULL) && (orig_referrer == NULL) )
      { /* no explicit scheme, no referrer, but an explicit hostname; seems
           that the user entered something like "//www.foo.org" manually; very
           rare, but should be supported */
        if (strneqcase(hostname, strWwwDot, 4)) rp = rpHttp;
        else if (strneqcase(hostname, strFtpDot, 4)) rp = __rpFtp;
        else if ( (strneqcase(hostname, strNewsDot, 5)) ||
                  (streqcase(hostname, strNews)) )
          rp = __rpNntp;
      }
      hostname = my_strdup_tolower(hostname);
        /* case-insensitivity: RFC2396 (6) and RFC2616 (3.2.3) */
    }
  }

  wouldbe_rp = rp; rp = rp_data[wouldbe_rp].final_rp;
  ua = cond2boolean(rp_data[wouldbe_rp].flags & rpfUsesAuthority);
  iph = cond2boolean(rp_data[wouldbe_rp].flags & rpfIsPathHierarchical);

  /* calculate the portnumber: */
  if (ua)
  { if ( (port == NULL) || (port[0] == '\0') )
      portnumber = rp2portnumber(wouldbe_rp);
    else
    { int _portnumber;
      my_atoi(port, &_portnumber, NULL, 99999);
      if ( (_portnumber <= 0) || (_portnumber > 65535) )
      { request_set_error(request, rePortnumber); portnumber = 0; }
      else portnumber = (tPortnumber) htons((unsigned short) _portnumber);
    }
  }

#if CONFIG_FTP
  if (is_ftplike(wouldbe_rp))
  { /* check for ";type=X" ending of FTP path */
    const char *l_username, *l_password;
    size_t len = strlen(path);
    if ( (len >= 7) && (strneqcase(path + len - 7, ";type=", 6)) )
    { char typech = *(path + len - 1);
      if (typech == 'a') request->flags |= rrfFtpTypeA;
      else if (typech == 'd') request->flags |= rrfFtpTypeD;
      else if (typech == 'i') request->flags |= rrfFtpTypeI;
      else goto no_ftp_type;
      *(path + len - 7) = '\0';
      no_ftp_type: {}
    }
    /* handle username and password */
    l_username = "anonymous"; l_password = "guest";
    if ( (wouldbe_rp == rpFtp) && ((username == NULL) || (password == NULL)) )
    { const tConfigLogin* l = ( (hostname != NULL) ? config.ftp_login : NULL );
      tPortnumber portnumber1 = portnumber;
      while (l != NULL)
      { tPortnumber portnumber2 = l->hosts_portnumber;
        if ( ( (portnumber2 == 0) || (portnumber2 == portnumber1) ) &&
             (my_pattern_matcher(l->hosts_pattern, hostname)) ) /* found */
        { l_username = l->user; l_password = l->password; break; }
        l = l->next;
      }
    }
    if (username == NULL) username = l_username;
    if ( (password == NULL) && (username != NULL) && (l_username != NULL) &&
         (!strcmp(username, l_username)) )
      password = l_password;
  }
#endif

  /* Complete the data */

  if ( (query != NULL) && (query[0] == '\0') ) query = NULL;
  if ( (fragment != NULL) && (fragment[0] == '\0') ) fragment = NULL;

  if ( (query != NULL) && (extra_query != NULL) )
  { my_spf(NULL, 0, &complete_query, "%s&%s", query, extra_query);
    must_dealloc_complete_query = truE;
  }
  else if (query != NULL) complete_query = query;
  else if (extra_query != NULL) complete_query = unconstify(extra_query);
  else complete_query = NULL;

  if ( (wouldbe_rp == rpLocal) || (wouldbe_rp == __rpLocalCgi) )
  { if ( (path[0] == '~') && (path[1] == '/') )
    { my_spf(NULL, 0, &path, strPercsPercs, get_homepath(), path + 2);
      must_dealloc_path = truE;
    }
  }

  if ( (referrer != NULL) && (rp != referrer->protocol) )
    referrer = NULL; /* can't use it */

  if (referrer != NULL)
  { const char* refpath;
    /* duplicate any missing information by taking it from the referrer: */
    if (ua)
    { if (hostname == NULL) hostname = my_strdup(referrer->hostname);
      else if (strcmp(hostname, referrer->hostname))
      { referrer = NULL; goto done_with_referrer; } /* can't use it */
    }
    if ( (iph) && (path[0] != '/') && ( (refpath = referrer->path) != NULL ) )
    { const char* temp = my_strrchr(refpath, '/');
      if (temp == NULL) /* "can't happen" */
      { __cleanup_path path = unconstify(strSlash);
      }
      else
      { size_t refpathsize = temp - refpath + 1, pathsize = strlen(path) + 1,
          size = refpathsize + pathsize;
        char* dest = __memory_allocate(size);
        my_memcpy(dest, refpath, refpathsize);
        my_memcpy(dest + refpathsize, path, pathsize);
        __cleanup_path path = dest; must_dealloc_path = truE;
      }
    }
  }
  done_with_referrer: {}

  if ( (path[0] != '/') && (is_locallike(wouldbe_rp)) )
  { char *spfbuf, *pathtemp = path;
    const char* prefix = my_getcwd();
    while (!strncmp(pathtemp, strDotSlash, 2)) pathtemp += 2; /* REMOVEME! */
    my_spf(NULL, 0, &spfbuf, strPercsPercs, prefix, pathtemp);
    __cleanup_path path = my_spf_use(spfbuf); must_dealloc_path = truE;
  }

  if ( (iph) && (path[0] == '\0') )
  { __cleanup_path path = unconstify(strSlash); }

  if (iph)
  { char* finpath = finalize_path(path);
    __cleanup_path path = finpath; must_dealloc_path = truE;
  }

  switch (rp)
  { case rpAbout:
      if (path[0] == '\0') { __cleanup_path path = unconstify(strRetawq); }
      break;
#if OPTION_NEWS
    case rpNntp:
      if (hostname == NULL)
      { const char* n = config.news_server;
        if (n != NULL) hostname = my_strdup(n);
      }
      break;
#endif
  }

  /* Compose the final URI and store everything in the request structure */

  if (ua) sprint_safe(strbuf, strPercd, ntohs(portnumber));

  my_spf(NULL, 0, &spfbuf_uri
    ,
/*A*/ "%s:" /* scheme */
/*B*/ "%s%s%s%s" /* double slash, hostname, colon, portnumber (if ua) */
/*C*/ "%s%s" /* slash (conditional), path */
/*D*/ "%s%s" /* question mark, query (if any query) */
    ,
/*A*/ ( ( (scheme != NULL) && (!is_rp_nice(wouldbe_rp)) ) ? scheme :
/*A*/   rp_data[wouldbe_rp].scheme ),
/*B*/ (ua ? strSlashSlash : strEmpty),
/*B*/ ( (ua && (hostname != NULL)) ? hostname : strEmpty ),
/*B*/ (ua ? ":" : strEmpty), (ua ? strbuf : strEmpty), /* portnumber */
/*C*/ ( (ua && (path[0] != '/')) ? strSlash : strEmpty ), path,
/*D*/ ( (complete_query != NULL) ? strQm : strEmpty ),
/*D*/ ( (complete_query != NULL) ? complete_query : strEmpty )
    );

  request->protocol = rp;
  request->uri = my_spf_use(spfbuf_uri);
  request->hostname = hostname;
  if (path != NULL)
  { if (must_dealloc_path) { request->path = path; must_dealloc_path = falsE; }
    else request->path = my_strdup(path);
  }
  if (complete_query != NULL)
  { if (must_dealloc_complete_query)
    { request->query = complete_query; must_dealloc_complete_query = falsE; }
    else request->query = my_strdup(complete_query);
  }
  if (username != NULL) request->username = my_strdup(username);
  if (password != NULL) request->password = my_strdup(password);
  if (ua) request->textual_portnumber = portnumber;

  if (_fragment != NULL) *_fragment = (fragment ? my_strdup(fragment) : NULL);

  if (rp == rpUnknown) request_set_error(request, reProtocol);
  else if (rp == rpDisabled) request_set_error(request, reProtDisabled);
  else if (ua && (hostname == NULL)) request_set_error(request, reHostname);

  memory_deallocate(uristart);
  __dealloc(authority);
  if (must_dealloc_path) memory_deallocate(path);
  if (must_dealloc_complete_query) memory_deallocate(complete_query);
}

#undef __cleanup_path

static void prepare_resource_request(tWindow* window, const char* uri,
  tPrrFlags prrf, const tWindowView* _referrer,
  /*@out@*/ tResourceRequest** _result)
{ tResourceRequest* request = request_allocate();
  const tResourceRequest* referrer = (_referrer ? _referrer->request : NULL);
  tWindowView *current_view, *new_view;
  const char *sfbuf_final, *extra_query = NULL;
  char* anchor;

  if (_result != NULL) *_result = request;
  if (sfbuf == NULL) sfbuf_final = NULL;
  else
  { sfbuf_final = sfbuf;
    if (sfbuf_final[0] == '&') sfbuf_final++;
    if (sfbuf_final[0] == '\0') sfbuf_final = NULL; /* no "real" data */
  }
  if (prrf & prrfPost)
  { request->flags |= rrfPost;
    if ( (prrf & prrfUseSfbuf) && (sfbuf_final != NULL) )
      request->post = my_strdup(sfbuf_final);
  }
  else if ( (prrf & prrfUseSfbuf) && (sfbuf_final != NULL) )
  { /* not a post request, but still some submit-form data => query */
    extra_query = sfbuf_final;
  }
  parse_uri(uri, request, referrer, &anchor, extra_query);
  if (sfbuf != NULL) { memory_deallocate(sfbuf); sfbuf_reset(); }
  if ( ( (program_mode == pmDownload) || (program_mode == pmDump) ) &&
         (request->uri != NULL) && (lfdmbs) )
  { my_write_str(2, "URL: "); my_write_str(2, request->uri);
    my_write_str(2, strNewline);
  }

  if ( (prrf & prrfIsRedirection) && (referrer != NULL) )
  { /* check whether old and redirected URI are the same */
    const char* requri = request->uri;
    if (requri != NULL)
    { const tResource* refres = referrer->resource;
      if ( ( (refres != NULL) && (refres->uri != NULL) &&
             (!strcmp(requri, refres->uri)) ) ||
           ( (referrer->uri != NULL) && (!strcmp(requri, referrer->uri)) ) )
      { __request_deallocate(request); __dealloc(anchor);
        if (_result != NULL) *_result = NULL;
        return;
      }
    }
  }

  if (prrf & prrfDownload)
  { request->flags |= rrfDownload; request->download_fd = prr_download_fd; }

  current_view = (window ? window->current_view : NULL);
  new_view = (window ? window_append_view(window, request) : NULL);
  if (new_view != NULL)
  { new_view->anchor = anchor;
    if (prrf & prrfPost) new_view->flags |= wvfPost;
    if (prrf & prrfSource)
    { new_view->flags |= wvfSource;
      if ( (current_view != NULL) && (current_view->html_title != NULL) )
        new_view->html_title = my_strdup(current_view->html_title);
    }
  }
  else __dealloc(anchor);

  if ( (prrf & prrfIsRedirection) ||
       ( (_referrer != NULL) && (_referrer->redirections > 0) ) )
  { request->flags |= rrfIsRedirection; }

  if ( (prrf & prrfIsRedirection) && (_referrer != NULL) && (new_view != NULL))
  { unsigned char count = new_view->redirections = _referrer->redirections + 1;
    if (count > config.redirections)
    { redirection_error:
      if (request->state != rrsError) request_set_error(request,reRedirection);
      goto finish;
    }
    if ( (prrf & prrfIsHttpRedirection) && (referrer != NULL) )
    { const tResource* res = referrer->resource;
      if ( (res != NULL) && (res->flags & rfPost) )
        goto redirection_error; /* RFC2616, 9.3 */
    }
  }

  finish:
  if ( (request->uri != NULL) && (new_view != NULL) )
    new_view->title = my_strdup(request->uri);
#if CONFIG_JAVASCRIPT
  if ( (request->protocol == rpJavascript) && (request->state != rrsError) )
    request_set_error(request, reProtDisabled);
#endif
  if (is_environed)
  { if (prrf & prrfRedrawAll) window_redraw_all();
    else if (prrf & prrfRedrawOne) window_redraw(window);
  }
  if (request->state != rrsError)
  { tResourceRequestAction action;
    if (prrf & prrfEnforcedReload) action = rraEnforcedReload;
    else if (prrf & prrfReload) action = rraReload;
    else
    { action = rraLoad;
#if OPTION_MENUS
      my_strdedup(uri_history[uri_history_index], request->uri);
      if (uri_history_index < URI_HISTORY_LEN - 1) uri_history_index++;
      else uri_history_index = 0;
      /* IMPLEMENTME: if the URI already is in the list, just "move it" to the
         top of the list! */
#endif
    }
    queue_request(request, action);
  }
}

static void reload_document(void)
{ const tWindowView* view = current_window->current_view;
  const char* uri = my_strdup(view->title);
  const tBoolean do_post = cond2boolean(lid.prrf & prrfPost);
  tPrrFlags prrf = lid.prrf | prrfRedrawOne;
  if (do_post)
  { sfbuf = my_strdup(null2empty(view->request->post));
    sfbuf_size = sfbuf_maxsize = strlen(sfbuf) + 1;
    prrf |= prrfUseSfbuf;
  }
  cut_windowviewlist(current_window);
  prepare_resource_request(current_window, uri, prrf, NULL, NULL);
  memory_deallocate(uri);
}

#if TGC_IS_CURSES
static my_inline void terminal_too_small(void)
{ show_message(_("Terminal too small"), truE);
}
#endif


/* HTML form submission and reset */

static tHtmlFormNumber ae2form(const tResource* resource,
  tActiveElementNumber _ae)
/* searches for the number of the HTML form to which the <_ae> belongs */
{ tHtmlFormNumber hfn, hfnum = resource->hfnum;
  if ( (_ae != INVALID_AE) && (hfnum > 0) )
  { const tHtmlForm* form = resource->form;
    for (hfn = 0; hfn < hfnum; hfn++)
    { tActiveElementNumber a1 = form[hfn].first_ae, a2 = form[hfn].last_ae;
      if ( (a1 != INVALID_AE) && (a2 != INVALID_AE) && (_ae>=a1) && (_ae<=a2) )
        return(hfn); /* found */
    }
  }
  return(INVALID_HTML_FORM_NUMBER); /* not found */
}

static void submit_form(void)
{ const tWindowView* view = current_window->current_view;
  const tResource* resource = view->request->resource;
  tActiveElementNumber _ae, a1, a2;
#if TGC_IS_GRAPHICS
  tActiveElementNumber aecur = INVALID_AE; /* FIXME! */
  tHtmlFormNumber hfn = INVALID_HTML_FORM_NUMBER; /* FIXME! */
#else
  tActiveElementNumber aecur = view->aecur;
  tHtmlFormNumber hfn = ae2form(resource, aecur);
#endif
  const tHtmlOption* o;
  tHtmlOptionNumber onum;
  const tActiveElementBase* aebase;
  const tActiveElement* aes;
  tPrrFlags prrf;

  if (hfn == INVALID_HTML_FORM_NUMBER)
  { cant_submit:
    show_message(_("Can't submit - not inside a form?"), truE);
    return;
  }
  a1 = resource->form[hfn].first_ae; a2 = resource->form[hfn].last_ae;
  if ( (a1 == INVALID_AE) || (a2 == INVALID_AE) ) goto cant_submit;

  javascript_handle_event(jekSubmit, aecur);
  sfbuf_reset();
  aebase = resource->aebase; aes = view->active_element;
  for (_ae = a1; _ae <= a2; _ae++)
  { tActiveElementKind kind;
    const char *name, *value, *bitfield;
    tBoolean is_first, is_multiple;
    if (aebase[_ae].flags & aefDisabled) continue;
    kind = aebase[_ae].kind;
    if ( (!is_form_aek(kind)) || ((name = aebase[_ae].data) == NULL)
        || (name[0] == '\0') )
    { continue; } /* non-form element or no element name - unusable here */
    switch (kind)
    {case aekFormText: case aekFormPassword: case aekFormTextarea:
      value = aes[_ae].current_text;
      append:
      if (value != NULL)
      { __sfbuf_add('&'); sfbuf_add_str(name);
        __sfbuf_add('='); sfbuf_add_str(value);
      }
      break;
     case aekFormCheckbox:
      if (aes[_ae].flags & aefCheckedSelected) { value = strOn; goto append; }
      break;
     case aekFormRadio:
      if (aes[_ae].flags & aefCheckedSelected)
      { value = aebase[_ae].render; goto append; }
      break;
     case aekFormSubmit: case aekFormImage:
      if (_ae == aecur)
      { if (kind == aekFormImage)
        { __sfbuf_add('&'); sfbuf_add_str(name); __sfbuf_add_str(".x=0&");
          sfbuf_add_str(name); __sfbuf_add_str(".y=0"); continue;
        }
        else if (aebase[_ae].flags & aefButtonTag) value = aebase[_ae].render;
        else value = strOn;
        goto append;
      }
      break;
     case aekFormHidden:
      value = aebase[_ae].render; goto append; /*@notreached@*/ break;
     case aekFormSelect:
      bitfield = aes[_ae].current_text;
      o = (const tHtmlOption*) aebase[_ae].render; onum = 0; is_first = truE;
      is_multiple = cond2boolean(aebase[_ae].flags & aefMultiple);
      while (o != NULL)
      { if (my_bit_test(bitfield, onum)) /* option is selected */
        { if (!is_first) __sfbuf_add(',');
          else
          { __sfbuf_add('&'); sfbuf_add_str(name); __sfbuf_add('=');
            is_first = falsE;
          }
          sfbuf_add_str(o->value);
          if (!is_multiple) break; /* done with collecting */
        }
        o = o->next; onum++;
      }
      break;
     /* case aekFormReset: break; -- nothing to do, element can't succeed */
     /* case aekFormFile: break; -- IMPLEMENTME! */
    }
  }
  __sfbuf_add('\0');
  prrf = prrfRedrawOne | prrfUseSfbuf;
  if (resource->form[hfn].flags & hffMethodPost) prrf |= prrfPost;
  prepare_resource_request(current_window, resource->form[hfn].action_uri,
    prrf, view, NULL);
}

static void reset_form(void)
{ tWindowView* view = current_window->current_view;
  tResource* resource = view->request->resource;
  const tActiveElementBase* aebase = resource->aebase;
#if TGC_IS_GRAPHICS
  tActiveElementNumber aecur = INVALID_AE; /* FIXME! */
  tHtmlFormNumber hfn = INVALID_HTML_FORM_NUMBER; /* FIXME! */
#else
  tActiveElementNumber aecur = view->aecur;
  tHtmlFormNumber hfn = ae2form(resource, aecur);
#endif
  if (hfn == INVALID_HTML_FORM_NUMBER)
  { cant_reset:
    show_message(_("Can't reset - not inside a form?"), truE);
  }
  else
  { tActiveElementNumber _ae, a1 = resource->form[hfn].first_ae,
      a2 = resource->form[hfn].last_ae;
    if ( (a1 == INVALID_AE) || (a2 == INVALID_AE) ) goto cant_reset;
    else
    { tActiveElement* aes;
      javascript_handle_event(jekReset, aecur);
      aes = view->active_element;
      for (_ae = a1; _ae <= a2; _ae++)
        init_ae(&(aebase[_ae]), &(aes[_ae]), falsE, view);
      window_redraw(current_window);
    }
  }
}


/* Downloads */

static void content_download_to_fd(const char* uri, int fd)
{ prr_download_fd = fd;
  prepare_resource_request(NULL, uri, prrfDownload, NULL,
    ((program_mode == pmDownload) ? (&pm_request) : NULL));
}

static void __content_download(const char* filename, tBoolean may_overwrite)
{ const char* uri = khmli_download_uri;
  int fd, cflags = O_CREAT | O_TRUNC | O_WRONLY;
  struct stat statbuf;
  if (!may_overwrite) cflags |= O_EXCL;
  fd = my_create(filename, cflags, S_IRUSR | S_IWUSR);
  if (fd < 0)
  { failed:
    if (is_environed) show_message_osfailure(errno);
    else fatal_error(errno, _("can't download to file"));
    return;
  }
  if (my_fstat(fd, &statbuf) != 0)
  { int e; close_bad: e = errno; my_close(fd); errno = e; goto failed; }
  if (!S_ISREG(statbuf.st_mode))
  { if (S_ISDIR(statbuf.st_mode)) errno = EISDIR;
    goto close_bad;
  }
  content_download_to_fd(uri, fd);
}

static my_inline void content_download(tBoolean may_overwrite)
{ __content_download(khmli_filename, may_overwrite);
}


/* Sessions */

#if CONFIG_SESSIONS

static const char strSessionMarker[] = "|#|x:";
#define smlp (3) /* session marker letter position */
#define psm(buf, ch) /* prepare session marker */ \
  do { strcpy((buf), strSessionMarker); *((buf) + smlp) = (ch); } while (0)
#define bdp(buf) ((buf) + strlen(strSessionMarker)) /* buffer data position */

static void session_save(const char* filename, tBoolean may_overwrite)
{ int fd, cflags = O_CREAT | O_TRUNC | O_WRONLY;
  struct stat statbuf;
  tWindow* window;
  char flagsbuf[50], linenumbuf[50], widthbuf[50], winbuf[50],
    htmltitlemarker[10], urimarker[10];

  /* create the file */
  if (!may_overwrite) cflags |= O_EXCL;
  fd = my_create(filename, cflags, S_IRUSR | S_IWUSR);
  if (fd < 0) { failed: show_message_osfailure(errno); return; }
  if (my_fstat(fd, &statbuf) != 0)
  { int e; close_bad: e = errno; my_close(fd); errno = e; goto failed; }
  if (!S_ISREG(statbuf.st_mode))
  { if (S_ISDIR(statbuf.st_mode)) errno = EISDIR;
    goto close_bad;
  }

  /* save the session */
  my_write_str(fd,
    _("# Session file for retawq (http://retawq.sourceforge.net/)\n"));
  psm(flagsbuf, 'f'); psm(linenumbuf, 'l'); psm(htmltitlemarker, 'h');
  psm(urimarker, 'u');
  psm(widthbuf, 'w'); *(bdp(widthbuf)) = strTG[0];
  winbuf[0] = 'W'; winbuf[1] = ':';
  window = windowlisthead;
  while (window != NULL)
  { tWindowView *view = window->first_view, *currview = window->current_view;
    tBoolean is_fuiw = truE; /* first URI in window? */
    while (view != NULL)
    { const tResourceRequest* request;
      tWindowViewFlags wvf = view->flags;
      const char *uri, *html_title, *flagsptr, *linenumptr;
      char* spfbuf;
      tLinenumber linenum;
      tBoolean is_currview = cond2boolean(view == currview);

      if (wvf & wvfPost) goto nextview; /* can't handle this correctly */

      request = view->request;
      if (request != NULL)
      { const tResource* resource = request->resource;
        if ( (resource != NULL) && ( (uri = resource->uri) != NULL ) &&
             (uri[0] != '\0') )
        { goto write_uri; }
        if ( ( (uri = request->uri) != NULL ) && (uri[0] != '\0') )
          goto write_uri;
      }
      if ( ( (uri = view->title) != NULL ) && (uri[0] != '\0') )
        goto write_uri;
      goto nextview; /* no usable URI found */

      write_uri:
      if (is_fuiw)
      { char* temp = winbuf + 2;
#if TGC_IS_CURSES
        unsigned char num;
        for (num = 0; num < 2; num++)
        { if (window == visible_window[num])
          { psm(temp, 'v'); temp = bdp(temp); *temp++ = '0' + num; break; }
        }
#endif
        *temp++ = '\n'; *temp = '\0';
        my_write_str(fd, winbuf);
        is_fuiw = falsE;
      }

      if ( (wvf & (wvfSource | wvfEnforceHtml)) || (is_currview) )
      { char* temp = bdp(flagsbuf);
        if (wvf & wvfSource) *temp++ = 's';
        if (wvf & wvfEnforceHtml) *temp++ = 'h';
        if (is_currview) *temp++ = 'c';
        *temp = '\0';
        flagsptr = flagsbuf;
      }
      else flagsptr = strEmpty;

      linenum = view->origin_y;
      if (linenum <= 0) linenumptr = strEmpty;
      else
      { sprint_safe(bdp(linenumbuf), strPercd, view->origin_y);
        linenumptr = linenumbuf;
      }

      html_title = ( (wvf & wvfSource) ? view->html_title : NULL );
      if ( (html_title != NULL) && (html_title[0] == '\0') )
        html_title = NULL;

      my_spf(strbuf, STRBUF_SIZE, &spfbuf, "U:%s%s%s%s%s%s\n",
        flagsptr, linenumptr, urimarker, uri,
        ( (html_title != NULL) ? htmltitlemarker : strEmpty ),
        null2empty(html_title));
      my_write_str(fd, spfbuf);
      my_spf_cleanup(strbuf, spfbuf);

      nextview:
      view = view->next;
    }
    window = window->next;
  }

  /* finish */
  my_close(fd);
}

#undef smlp
#undef psm
#undef bdp

static tLinenumber session_linenum;
static tWindow* session_window;
#if TGC_IS_CURSES
static tWindow* session_viswin[2];
#endif
static tWindowView* session_currview;
#define SESSIONPART_MAXNUM (10)
static const char* session_part[SESSIONPART_MAXNUM];
static unsigned char session_numparts;

static void session_resume_split(char* line)
{ char* temp;
  session_numparts = 0;
  loop:
  temp = my_strstr(line, "|#|");
  if (temp != NULL)
  { char ch = temp[3];
    if (!my_islower(ch)) { incandloop: line = temp + 1; goto loop; }
    if (temp[4] != ':') goto incandloop;
    temp[0] = '\0';
    session_part[session_numparts++] = temp + 3;
    if (session_numparts < SESSIONPART_MAXNUM)
    { line = temp + 5; goto loop; } /* look for more */
  }
}

static void session_resume_finish_window(void)
{ if ( (session_window != NULL) && (session_currview != NULL) )
    session_window->current_view = session_currview;
}

static void session_resume_window(void)
{ session_resume_finish_window();
  session_window = window_create();
  session_currview = NULL;
}

static void session_resume_line(const char* line, const char* limit)
{ size_t len = limit - line;
  char buf[STRBUF_SIZE], *temp;
  unsigned char count;
  session_linenum++;
  if (len <= 0) return; /* empty line */
  if (len > STRBUF_SIZE - 5) return; /* very long line, can't be serious */
  my_memcpy(buf, line, len); buf[len] = '\0'; /* (for simplicity) */
  if (buf[len - 1] == '\r') /* (e.g. manually written file?) */
    buf[len - 1] = '\0';
  debugmsg("session_resume_line(): *"); debugmsg(buf); debugmsg("*\n");
  temp = buf;
  while ( (*temp == ' ') || (*temp == '\t') ) temp++; /* skip whitespace */
  if (*temp == '#') return; /* comment line */
  if (!strncmp(temp, "W:", 2))
  { session_resume_window();
#if TGC_IS_CURSES
    session_resume_split(temp + 2);
    for (count = 0; count < session_numparts; count++)
    { const char *str = session_part[count], *data = str + 2;
      int num;
      debugmsg("session part W: *"); debugmsg(str); debugmsg("*\n");
      if (str[0] != 'v') continue; /* unknown part */
      if (data[0] == '\0') continue; /* empty part */
      my_atoi(data, &num, &data, 9);
      if ( (num >= 0) && (num <= 1) && (data[0] == '\0') )
        session_viswin[num] = session_window;
    }
#endif
  }
  else if (!strncmp(temp, "U:", 2))
  { const char *uri, *flagstr, *html_title;
    session_resume_split(temp + 2);
    uri = flagstr = html_title = NULL;
    for (count = 0; count < session_numparts; count++)
    { const char *str = session_part[count], *data = str + 2;
      debugmsg("session part U: *"); debugmsg(str); debugmsg("*\n");
      if (data[0] == '\0') continue; /* empty part */
      switch (str[0])
      { case 'u': uri = data; break;
        case 'f': flagstr = data; break;
        case 'h': html_title = data; break;
        /* "default": unknown part; IMPLEMENTME: print warning/error? */
      }
    }
    if (uri != NULL) /* got useful information */
    { tResourceRequest* request;
      tPrrFlags prrf = prrfRedrawOne;
      tBoolean enforce_html, is_currview;
      enforce_html = is_currview = falsE;
      if (flagstr != NULL)
      { const char* f = flagstr;
        char ch;
        while ( (ch = *f++) != '\0' )
        { if (ch == 's') prrf |= prrfSource;
          else if (ch == 'h') enforce_html = truE;
          else if (ch == 'c') is_currview = truE;
        }
      }
      if (session_window == NULL) session_resume_window();
      prepare_resource_request(session_window, uri, prrf, NULL, &request);
      if (request != NULL)
      { tWindowView* view = request->window_view;
        if (enforce_html) view->flags |= wvfEnforceHtml;
        if (is_currview) session_currview = view;
        if (html_title != NULL) my_strdedup(view->html_title, html_title);
      }
    }
  }
  /* "else": IMPLEMENTME: bad line, print a message, stop resumption! */
}

static void session_resume(const char* filename)
{ size_t size;
  void* filebuf;
  char *buf, *end, *temp;
  if (filename[0] == '\0') filename = config.session_default;
  debugmsg("session_resume(): *"); debugmsg(filename); debugmsg("*\n");
  switch (my_mmap_file_readonly(filename, &filebuf, &size))
  { case 0: show_message_osfailure(errno); return; /*@notreached@*/ break;
    case 1: goto out; /*@notreached@*/ break;
  }
  buf = (char*) filebuf; end = buf + size - 1;
  session_window = NULL;
#if TGC_IS_CURSES
  session_viswin[0] = session_viswin[1] = NULL;
#endif
  session_linenum = 0; session_currview = NULL;
  temp = buf;
  while (temp <= end)
  { if (*temp == '\n') { session_resume_line(buf, temp); buf = temp + 1; }
    temp++;
  }
  my_munmap(filebuf, size);
  session_resume_finish_window();
#if TGC_IS_CURSES
  { tWindow *viswin0 = session_viswin[0], *viswin1 = session_viswin[1];
    if ( (viswin0 == NULL) && (viswin1 != NULL) )
    { viswin0 = viswin1; viswin1 = NULL; }
    if (viswin0 != NULL) visible_window[0] = viswin0;
    if ( (viswin1 != NULL) && (viswin1 != visible_window[0]) )
      visible_window[1] = viswin1;
  }
#endif
  window_redraw_all();

  out:
  show_message(_("Done"), falsE);
}

#endif /* #if CONFIG_SESSIONS */


/* Contextual menus */

#if OPTION_MENUS

#if TGC_IS_CURSES

static void my_hline(short x1, short x2, short y)
/* draws a horizontal line */
{ (void) move(y, x1);
  while (x1 <= x2) { (void) addch(__MY_HLINE); x1++; }
}

static void my_vline(short y1, short y2, short x)
/* draws a vertical line */
{ while (y1 <= y2) { mvaddch(y1, x, __MY_VLINE); y1++; }
}

static void draw_box(short x1, short y1, short x2, short y2)
{ my_set_color(cpnBlue); /* attron(A_BOLD); */
  my_hline(x1 + 1, x2 - 1, y1); my_hline(x1 + 1, x2 - 1, y2);
  my_vline(y1 + 1, y2 - 1, x1); my_vline(y1 + 1, y2 - 1, x2);
  mvaddch(y1, x1, __MY_UL); mvaddch(y1, x2, __MY_UR);
  mvaddch(y2, x1, __MY_LL); mvaddch(y2, x2, __MY_LR);
  /* attroff(A_BOLD); */ my_set_color(cpnDefault);
}

typedef void* tCmCallbackData;
typedef void (*tCmCallbackFunction)(tCmCallbackData);
typedef signed int tCmNumber; /* ("signed" for simplicity only) */

typedef struct
{ const char* render;
  tCmCallbackFunction function;
  tCmCallbackData data;
} tCmEntry;

static tCmEntry* cm_info;
static tCmNumber cm_num, cm_maxnum, cm_current,
  cm_topmost, cm_onscreen_num; /* what's displayed on the screen */
static tCoordinate cm_x1, cm_y1, cm_x2, cm_y2;

static char* cm_select_bitfield;
static tBoolean cm_select_is_multiple;
static tHtmlOptionNumber cm_select_old; /* formerly selected option */

static void cm_init(void)
{ cm_info = NULL;
  cm_num = cm_maxnum = cm_current = cm_topmost = 0;
}

static void __cm_add(const char* str, tCmCallbackFunction func,
  tCmCallbackData data)
{ if (cm_num >= cm_maxnum)
  { cm_maxnum += 20;
    cm_info = memory_reallocate(cm_info, cm_maxnum * sizeof(tCmEntry));
  }
  cm_info[cm_num].render = my_strdup(str);
  cm_info[cm_num].function = func; cm_info[cm_num].data = data;
  cm_num++;
}

#define cm_add(str, func, val) \
  __cm_add(unconstify_or_(str), func, (tCmCallbackData) MY_INT_TO_POINTER(val))

static my_inline void cm_add_separator(void)
{ if (cm_num > 0) cm_add(strMinus, NULL, NULL);
}

static void cm_activate(tBoolean active)
/* indicates on screen what's currently selected ("menu cursor") */
{ short x, x0 = cm_x1 + 1, y = (cm_y1 + 1) + (cm_current - cm_topmost);
  (void) move(y, x0);
  for (x = x0; x < cm_x2; x++)
  { chtype ch = inch();
    if (active) ch |= A_REVERSE;
    else ch &= ~A_REVERSE;
    (void) addch(ch);
  }
  (void) move(LINES - 1, 0);
}

static void __cm_draw(void)
/* draws the menu texts */
{ tCmNumber num;
  short x = cm_x1 + 1, width = cm_x2 - cm_x1 - 1;
  for (num = 0; num < cm_onscreen_num; num++)
  { short y = cm_y1 + num + 1;
    const char* str = cm_info[cm_topmost + num].render;
    if ( (str[0] == '-') && (str[1] == '\0') )
    { /* It's a separator */
      my_hline(x, x + width - 1, y);
    }
    else
    { short count = width - strlen(str);
      /*(void)*/ mvaddnstr(y, x, str, width);
      while (count-- > 0) (void) addch(' ');
    }
  }
}

static one_caller tBoolean cm_draw(short x, short y, const char* title)
/* calculates and draws the menu rectangle; returns whether it worked */
{ short titlelen = ( (title != NULL) ? strlen(title) : 0 ),
    titlewidth = ( (titlelen > 0) ? (titlelen + 2) : 0 ),
    width = titlewidth, cols = COLS, lines = LINES, maxwidth = cols - 10;
  tCmNumber num, onscreen_maxnum = lines - 4;
  if ( (maxwidth <= 5) || (onscreen_maxnum <= 2) ) return(falsE);
  for (num = 0; num < cm_num; num++)
  { short len = strlen(cm_info[num].render);
    if (width < len) width = len;
  }
  if (width > maxwidth) width = maxwidth;
  cm_onscreen_num = MIN(cm_num, onscreen_maxnum);

  if ( (x < 0) || (x >= cols) ) x = 0;
  if ( (y < 0) || (y >= lines) ) y = 0;

  if (x + width + 1 >= cols - 1) x = cols - width - 3;
  if (y + cm_onscreen_num + 1 >= lines - 2) y = lines - cm_onscreen_num - 4;

  cm_x1 = x; cm_y1 = y; cm_x2 = x + width + 1; cm_y2 = y + cm_onscreen_num + 1;
  draw_box(cm_x1, cm_y1, cm_x2, cm_y2);
  if (titlelen > 0)
  { short maxlen = cm_x2 - cm_x1 - 3, len = MIN(maxlen, titlelen),
      offset = (maxlen - len) / 2;
    if (offset < 0) offset = 0;
    my_set_color(cpnBlue);
    /*(void)*/ mvaddnstr(cm_y1, cm_x1 + 2 + offset, title, len);
    my_set_color(cpnDefault);
  }
  __cm_draw(); cm_activate(truE);
  return(truE);
}

static void cm_start(short x, short y, const char* title, const char* errstr)
{ if (cm_num > 0)
  { if (cm_draw(x, y, title)) key_handling_mode = khmContextualMenu;
    else terminal_too_small();
  }
  else
  { /* The menu would be empty, so we won't show it. */
    if (errstr != NULL) show_message(errstr, truE);
  }
}

static void cm_handle_command_code(tCmCallbackData data)
{ handle_command_code((tProgramCommandCode) MY_POINTER_TO_INT(data));
}

static void cm_setup(short x, short y, tActiveElementNumber _ae)
/* sets up a contextual menu with ae-related entries (unless ae is invalid)
   and some general entries */
{ const tWindowView* view = current_window->current_view;
  cm_init();

  if (view != NULL)
  { const tResourceRequest* request = view->request;
    const tResource* resource = ((request != NULL) ? request->resource : NULL);
    const char* uri = ( (request != NULL) ? request->uri : NULL );

    /* ae-specific entries */
    if (_ae != INVALID_AE)
    { const tActiveElementBase* aeb = ((resource != NULL) ? resource->aebase :
        NULL);
      tActiveElementKind aek = ( (aeb != NULL) ? aeb[_ae].kind : aekUnknown );
      const tActiveElementCoordinates* aec = find_visible_aec(view, _ae);
      if (aec != NULL) { x = aec->x1; y = line2row(view, aec->y) + 1; }
      cm_add(strShowElementInfo, cm_handle_command_code, pccElementInfo);
      if ( (aeb != NULL) && (aeb[_ae].flags & aefDisabled) )
        cm_add(strEnableThisElement, cm_handle_command_code, pccElementEnable);
      if (aek == aekLink)
      { cm_add(strOpenLinkInNewWindow, cm_handle_command_code,
          pccWindowNewFromElement);
      }
      else if (is_form_aek(aek))
      { tHtmlFormNumber hfn = ae2form(resource, _ae);
        if (hfn != INVALID_HTML_FORM_NUMBER)
        { cm_add(strSubmitThisForm, cm_handle_command_code, pccFormSubmit);
          /* IMPLEMENTME: "Show Form Info" */
        }
      }
      cm_add_separator();
    }

    /* general entries */
    if (view->prev != NULL)
      cm_add(strBack, cm_handle_command_code, pccViewBack);
    if (view->next != NULL)
      cm_add(strForward, cm_handle_command_code, pccViewForward);
    if ( (uri != NULL) && (uri[0] != '\0') )
    { cm_add(strReload, cm_handle_command_code, pccDocumentReload);
      cm_add(strEnforcedReload, cm_handle_command_code,
        pccDocumentReloadEnforced);
      cm_add(strSaveAs, cm_handle_command_code, pccDocumentSave);
      cm_add(strOpenInNewWindow, cm_handle_command_code,
        pccWindowNewFromDocument);
#if 0
      cm_add(strAddBookmark, cm_handle_command_code, pccBookmarkAdd);
#endif
    }
    cm_add(strSourceCode, cm_handle_command_code, pccDocumentEnforceSource);
  }
  cm_start(x, y, _(strContext), _("No contextual menu available here!"));
}

static void cm_handle_select_tag(tCmCallbackData data)
/* callback function for menus for the HTML <select> tag */
{ tHtmlOptionNumber num = (tHtmlOptionNumber) MY_POINTER_TO_INT(data);
  if (cm_select_is_multiple) my_bit_flip(cm_select_bitfield, num);
  else
  { my_bit_clear(cm_select_bitfield, cm_select_old);
    my_bit_set(cm_select_bitfield, num);
  }
  window_redraw(current_window);
}

static void cm_setup_select_tag(const tActiveElementBase* aebase,
  const tWindowView* view, tActiveElementNumber _ae)
/* sets up a menu for the HTML <select> tag */
{ tActiveElement* aes = view->active_element;
  const tHtmlOption* o = (const tHtmlOption*) (aebase[_ae].render);
  const tActiveElementCoordinates* aec = find_visible_aec(view, _ae);
  tHtmlOptionNumber num = 0;
  short x, y;
  cm_init();
  cm_select_is_multiple = cond2boolean(aebase[_ae].flags & aefMultiple);
  cm_select_bitfield = aes[_ae].current_text;
  cm_select_old = 0;
  while (o != NULL)
  { tBoolean is_selected = cond2boolean(my_bit_test(cm_select_bitfield, num));
    char ch, *spfbuf;
    if (is_selected) ch = '+';
    else if (o->flags & hofDisabled) ch = '-'; /* CHECKME! */
    else ch = ' ';
    my_spf(strbuf, STRBUF_SIZE, &spfbuf, "%c%s", ch, o->render);
    __cm_add(spfbuf, cm_handle_select_tag,
      (tCmCallbackData) MY_INT_TO_POINTER(num));
    my_spf_cleanup(strbuf, spfbuf);
    if (is_selected) cm_select_old = num;
    o = o->next; num++;
  }
  if (aec != NULL) { x = aec->x1; y = line2row(view, aec->y) + 1; }
  else { x = y = 0; }
  cm_start(x, y, _("Options"), _(strSelectionEmpty));
}

static void cm_handle_uri_history(tCmCallbackData data)
{ unsigned short i = (unsigned short) MY_POINTER_TO_INT(data);
  prepare_resource_request(current_window, uri_history[i], prrfRedrawOne, NULL,
    NULL);
}

static void cm_setup_uri_history(void)
{ unsigned short count = 0, i = uri_history_index;
  cm_init();
  while (count < URI_HISTORY_LEN)
  { char* text;
    if (i > 0) i--;
    else i = URI_HISTORY_LEN - 1;
    text = uri_history[i];
    if (text == NULL) break; /* reached the "end" of the list */
    __cm_add(text, cm_handle_uri_history,
      (tCmCallbackData) MY_INT_TO_POINTER(i));
    count++;
  }
  cm_start(0, 0, _("History"), _("No URL history available yet!"));
}

static void cm_handle_window_list(tCmCallbackData data)
{ tWindow* window = (tWindow*) data;
  if (current_window != window) /* CHECKME! */
  { current_window = window;
    window_redraw(current_window);
  }
}

static void cm_setup_window_list(void)
{ const tWindow* w = windowlisthead;
  cm_init();
  while (w != NULL)
  { const tWindowView* view = w->current_view;
    const char *text, *html_title, *uri;
    if (view == NULL) text = _(strBracedNewWindow);
    else if ( ( (html_title = view->html_title) != NULL ) &&
              (html_title[0] != '\0') )
    { text = html_title; }
    else if ( ( (uri = view->title) != NULL ) && (uri[0] != '\0') )
    { text = uri; }
    else text = unconstify_or_(strUnknown);
    __cm_add(text, cm_handle_window_list, (tCmCallbackData) w);
    w = w->next;
  }
  cm_start(0, 0, _("Window List"), _("No open window!"));
}

static void cm_remove(void)
/* This is only for "internal" use by cm_....() functions; external users
   should call cm_cancel() instead. */
{ tCmNumber num;
  for (num = 0; num < cm_num; num++) __dealloc(cm_info[num].render);
  memory_deallocate(cm_info);
  key_handling_mode = khmCommand; window_redraw_all();
}

static void cm_change_current(tCmNumber num)
/* moves the "menu cursor" */
{ tCmNumber old_topmost;
  if (cm_current == num) return; /* no "real" change, nothing to do */
  cm_activate(falsE); /* deactivate old */
  cm_current = num;

  old_topmost = cm_topmost;
  if (cm_topmost > num) cm_topmost = num;
  else if (cm_topmost <= num - cm_onscreen_num)
    cm_topmost = num - cm_onscreen_num + 1;
  if (cm_topmost != old_topmost) __cm_draw(); /* need to "scroll" */
  cm_activate(truE);
}

static my_inline void cm_line_up(void)
{ if (cm_current > 0) cm_change_current(cm_current - 1);
  else if (cm_num > 0) cm_change_current(cm_num - 1);
}

static my_inline void cm_line_down(void)
{ if (cm_current < cm_num - 1) cm_change_current(cm_current + 1);
  else cm_change_current(0);
}

static my_inline void cm_page_up(void)
{ tCmNumber num = cm_current - (cm_y2 - cm_y1 - 1);
  if (num < 0) num = 0;
  cm_change_current(num);
}

static my_inline void cm_page_down(void)
{ tCmNumber num = cm_current + (cm_y2 - cm_y1 - 1);
  if (num > cm_num - 1) num = cm_num - 1;
  cm_change_current(num);
}

static my_inline void cm_cancel(void)
/* The user cancelled the menu explicitly or resized the terminal. */
{ cm_remove();
}

static void cm_apply(void)
/* The user pressed some KEY_ENTER variant. */
{ tCmCallbackFunction function = cm_info[cm_current].function;
  if (function != NULL)
  { tCmCallbackData data = cm_info[cm_current].data;
    cm_remove();
    (function)(data);
  }
}

#if CONFIG_DO_TEXTMODEMOUSE
static my_inline void cm_handle_mouseclick(tCoordinate x, tCoordinate y)
{ if ( (x > cm_x1) && (x < cm_x2) && (y > cm_y1) && (y < cm_y2) )
  { /* The user clicked inside the menu box. */
    tCmNumber num = ((tCmNumber) (y - cm_y1 - 1)) + cm_topmost;
    cm_change_current(num);
    cm_apply();
  }
  else cm_remove();
}
#endif

#endif /* #if TGC_IS_CURSES */

#endif /* #if OPTION_MENUS */


/* Line input */

#if TGC_IS_GRAPHICS

static tBoolean current_window_exists(void)
/* returns whether current_window still exists; in graphics mode, it's possible
   that a user starts a dialog which is related to a document window, then
   closes the window and finally clicks the "action" button of the dialog, so
   the callback function would refer to a non-existing window. This problem
   could be solved by using _modal_ dialogs, but I hate modality because it
   takes away flexibility from users - often unnecessarily. */
{ if (current_window != NULL)
  { const tWindow* w = windowlisthead;
    while (w != NULL)
    { if (w == current_window) return(truE);
      w = w->next;
    }
  }
  return(falsE);
}

/* "require current_window" */
#define __rcw(solution) if (!current_window_exists()) solution;
#define rcw __rcw(return) /* the simple solution */
#define rcwc __rcw(current_window = window_create()) /* sometimes nicer */

#else

#define __rcw(solution) /* nothing; current_window _is_ valid  */
#define rcw
#define rcwc

#endif


#if TGC_IS_GRAPHICS

#if 0
typedef struct
{ tWindow* window;
  GtkEntry* entry;
  tEditableLineInputGoal elig;
} tLigData;

static one_caller __sallocator tLigData* __callocator
  lig_data_create(tWindow* window, GtkWidget* _entry,
  tEditableLineInputGoal elig)
{ tLigData* retval = (tLigData*) __memory_allocate(sizeof(tLigData));
  GtkEntry* entry = ( (_entry != NULL) ? GTK_ENTRY(_entry) : NULL );
  retval->window = window; retval->entry = entry; retval->elig = elig;
  return(retval);
}

static void graphics_handle_lig_action(tGraphicsWidget* w __cunused,
  gpointer _data)
{ const tLigData* data = (tLigData*) _data;
  tEditableLineInputGoal elig = data->elig;
  current_window = data->window;
  if (is_confirmation_lig(_lig)) handle_lig_confirmation(_lig);
  else
  { GtkEntry* entry = data->entry;
    const char* text = ( (entry != NULL) ? gtk_entry_get_text(entry) : NULL );
    handle_lig_other(_lig, text);
  }
  current_window = NULL;
}
#endif

static void graphics_window_destroy(tGraphicsWidget* w __cunused,
  gpointer data)
/* low-level callback for window destruction triggered by buttons; for
   _document_ windows, use window_hide()/_remove() instead! */
{ gtk_widget_destroy((tGraphicsWidget*) data); /* (yes!) */
}

static GtkWidget* dialog_create_button(GtkDialog* dialog, const char* text)
{ GtkWidget* button = gtk_button_new_with_label(text);
  pack_box(dialog->action_area, button);
  show_widget(button);
  return(button);
}

#define connect_button(b, h, d) connect_widget(b, strGtkClicked, h, d)

#if 0
static void start_line_input_FAIL(tLineInputGoal _lig, const char* msg,
  const char* initstr, const char* wtitle, const char* wact,
  const char* wcancel)
{ GtkDialog* dialog = GTK_DIALOG(gtk_dialog_new());
  GtkWidget *button, *inputbox;
  if (wtitle != NULL) gtk_window_set_title(GTK_WINDOW(dialog), wtitle);
  if (is_confirmation_lig(_lig))
  { GtkWidget* label = gtk_label_new(msg);
    show_widget(label); pack_box(dialog->vbox, label);
    inputbox = NULL;
  }
  else
  { GtkWidget *hbox = gtk_hbox_new(FALSE, 0), *label = gtk_label_new(msg);
    inputbox = gtk_entry_new();
    if (initstr != NULL) gtk_entry_set_text(inputbox, (gchar*) initstr);
    show_widget(hbox); show_widget(label); show_widget(inputbox);
    pack_box(hbox, label); pack_box(hbox, inputbox);
    pack_box(dialog->vbox, hbox);
  }

  button = dialog_create_button(dialog, wact);
  connect_button(button, graphics_handle_lig_action,
    lig_data_create(current_window, inputbox, _lig));
  connect_button(button, graphics_window_destroy, dialog);

  button = dialog_create_button(dialog, wcancel);
  connect_button(button, graphics_window_destroy, dialog);

  show_widget(dialog);
}
#endif

static void __line_input_start(/*@notnull@*/ const char* msg,
  const char* initstr, tLineInputCallback func, void* data,
  tLineInputAreaFlags liaf)
{ /* FIXME! */
}

typedef struct
{ GtkFileSelection* sel;
  tWindow* window;
} tFileselData;

static one_caller __sallocator tFileselData* __callocator
  filesel_data_create(GtkWidget* sel, tWindow* window)
{ tFileselData* retval = (tFileselData*) __memory_allocate(sizeof(*retval));
  retval->sel = GTK_FILE_SELECTION(sel); retval->window = window;
  return(retval);
}

static void filesel_handle(GtkFileSelection* widget __cunused, gpointer _data)
{ const tFileselData* data = (tFileselData*) _data;
  const char* file = gtk_file_selection_get_filename(data->sel);
  if ( (file != NULL) && (file[0] != '\0') )
    prepare_resource_request(data->window, file, prrfRedrawOne, NULL, NULL);
}

#else /* #if TGC_IS_GRAPHICS */

static void line_input_redraw(void)
{ tLineInputAreaIndex count;
  for (count = 0; count < lid.num_areas; count++)
  { tLineInputArea* lia = &(lid.area[count]);
    const tLineInputAreaFlags liaf = lia->flags;
    const short len = lia->len, first = lia->first, usable = lia->usable,
      row = lia->row, colmin = lia->colmin;
    short cnt = len - first;
    const char* src = lid.area[count].text + first;
    if (cnt > usable) cnt = usable;
    (void) move(row, colmin);
    if (cnt > 0)
    { if (liaf & liafDisguising) { while (cnt-- > 0) (void) addch('*'); }
      else (void) addnstr(src, cnt);
    }
    (void) clrtoeol(); /* IMPROVEME! */
  }
  (void) move(lid_area(row), lid_area(colcurr));
}

static void line_input_layout(void)
{ tLineInputAreaIndex numa = lid.num_areas, count;
  const short line = LINES - 1, usable = COLS - 1;
  short col = 0;
  for (count = 0; count < numa; count++)
  { lid.area[count].row = line;
    if (lid.area[count].flags & liafEditable)
      lid.area[count].flags |= liafFocusable;
  }

  if (usable < 5 * numa)
  { key_handling_mode = khmCommand; terminal_too_small();
    (lid.callback)(lid.callback_data, liekFail, 0);
    return;
  }

  for (count = 0; count < numa; count++)
  { short len = lid.area[count].len, maxlen = usable - col - 5*(numa-count-1),
      used;
    if (len > maxlen) len = maxlen;
    used = ((lid.area[count].flags & liafEditable) ? maxlen : len);
    lid.area[count].colmin = lid.area[count].colcurr = col;
    lid.area[count].colmax = col + used; lid.area[count].usable = used;
    col += used;
  }

  line_input_redraw();
}

static my_inline void line_input_resize(void)
{ short count;
  for (count = 0; count < lid.num_areas; count++)
  { lid.area[count].pos = 0; lid.area[count].first = 0; } /* KISS */
  line_input_layout();
}

static one_caller void __line_input_to_eos(void)
/* ("eos": "end of string", as usual) */
{ if (lid_area(len) >= lid_area(usable))
  { lid_area(first) = lid_area(len) - lid_area(usable);
    lid_area(colcurr) = lid_area(colmax);
  }
  else
  { lid_area(first) = 0;
    lid_area(colcurr) = lid_area(colmin) + lid_area(len);
  }
}

static void line_input_to_eos(void)
{ if (lid_area(pos) < lid_area(len))
  { lid_area(pos) = lid_area(len); __line_input_to_eos(); line_input_redraw();}
}

static one_caller tMbsIndex do_lookup_lineinput_key(tKey key)
{ if (keymap_lineinput_keys_num <= 0) return(INVALID_INDEX); /*"can't happen"*/
  my_binary_search(0, keymap_lineinput_keys_num - 1,
    my_numcmp(key, keymap_lineinput_keys[idx].key), return(idx))
}

static tLineInputActionCode lookup_lineinput_key(tKey key)
{ tMbsIndex idx = do_lookup_lineinput_key(key);
  return( (idx < 0) ? liacUnknown : keymap_lineinput_keys[idx].liac );
}

#if CONFIG_DO_TEXTMODEMOUSE
static void mouse_flip(unsigned char what)
{ static tBoolean is_on = truE;
  tBoolean new_state;
  switch (what)
  { default: /* "can't happen" */ /*@fallthrough@*/
    case 0: new_state = cond2boolean(!is_on); break;
    case 1: new_state = falsE; break;
    case 2: new_state = truE; break;
  }
  if (is_on != new_state) /* must do something */
  { is_on = new_state;
    mousemask(is_on ? TEXTMODEMOUSE_MASK : 0, NULL);
    show_message(is_on ? _("mouse on") : _("mouse off"), falsE);
  }
}
#else
static const char strFtmm[] = N_("Ftext-mode mouse");
#define mouse_flip(dummy) fwdact(_(strFtmm))
#endif

static one_caller void line_input_handle_key(tKey key)
{ tLineInputActionCode liac = lookup_lineinput_key(key);
  switch (liac)
  { case liacCancel:
      (lid.callback)(lid.callback_data, liekCancel, 0);
      return; /*@notreached@*/ break;
    case liacAreaSwitch:
      if (lid.num_areas == 2) { lid.curr = 1 - lid.curr; line_input_redraw(); }
      return; /*@notreached@*/ break;
    case liacMouseFlip: case liacMouseOff: case liacMouseOn:
      mouse_flip(liac - liacMouseFlip); return; /*@notreached@*/ break;
  }

  if (lid_area(flags) & liafEditable) /* the callback will wanna do the work */
  { do_call: (lid.callback)(lid.callback_data, liekKey, key); return; }

  switch (liac)
  { case liacToLeft:
      if (lid_area(first) > 0)
      { lid_area(first)--; do_redraw: line_input_redraw(); }
      break;
    case liacToRight:
      if (lid_area(first) + lid_area(usable) < lid_area(len))
      { lid_area(first)++; goto do_redraw; }
      break;
    case liacToStart:
      if (lid_area(first) > 0) { lid_area(first) = 0; goto do_redraw; }
      break;
    case liacToEnd:
      if (lid_area(first) + lid_area(usable) < lid_area(len))
      { lid_area(first) = lid_area(len) - lid_area(usable); goto do_redraw; }
      break;
    default:
      if (lid.curr == lid.num_areas - 1) goto do_call;
      break;
  }
}

static void lid_prepare_text(const char* text, tLineInputAreaIndex idx)
{ size_t len = strlen(text);
  if (len > 1024) len = 1024; /* can't be serious */
  if (len > 0)
  { char* t = __memory_allocate(len); /* not storing trailing '\0' */
    my_memcpy(t, text, len); lid.area[idx].text = t;
  }
  lid.area[idx].len = lid.area[idx].maxlen = len;
}

static void __line_input_start(/*@notnull@*/ const char* msg,
  const char* initstr, tLineInputCallback func, void* data,
  tLineInputAreaFlags liaf)
{ tLineInputAreaIndex count = 0, curr;
  my_memclr_var(lid); lid.callback = func; lid.callback_data = data;
  lid_prepare_text(msg, count++);
  if (liaf & liafEditable)
  { if (initstr != NULL) lid_prepare_text(initstr, count);
    count++;
  }
  curr = count - 1; lid.area[curr].flags = liaf; lid.curr = curr;
  lid.num_areas = count; key_handling_mode = khmLineInput;
  line_input_layout();
  if ( (key_handling_mode == khmLineInput) && (liaf & liafEditable) )
    line_input_to_eos(); /* IMPROVEME! */
}

#endif /* #if TGC_IS_GRAPHICS */

static void line_input_ensure_room(void)
{ if (lid_area(len) >= lid_area(maxlen))
  { lid_area(maxlen) += 100;
    lid_area(text) = memory_reallocate(lid_area(text), lid_area(maxlen));
  }
}

static void line_input_finish(void)
{
#if TGC_IS_CURSES
  key_handling_mode = khmCommand;
#endif
  redraw_message_line();
  __dealloc(lid.area[0].text); __dealloc(lid.area[1].text);
}

static const char* request2filename(const tResourceRequest* request)
/* returns a dup()ed filename for the <request> or NULL */
{ const char *retval = NULL, *temp;
  const tResourceProtocol rp = request->protocol;
  if ( (is_rp_nice(rp)) && (rp_data[rp].flags & rpfIsPathHierarchical) &&
       ( (temp = request->path) != NULL ) )
  { const char* slash = my_strrchr(temp, '/');
    if (slash == NULL) retval = temp;
    else
    { slash++;
      if (*slash != '\0') retval = slash;
      /* "else": it's a directory */
    }
  }
  if (retval != NULL)
  { char ch, *ptr = __memory_allocate(strlen(retval) + 1), *ptr2 = ptr;
    while ( (ch = *retval++) != '\0' )
    { if ( (!my_isalnum(ch)) && (ch != '.') && (ch != '-') ) ch = '_';
      *ptr++ = ch;
    }
    *ptr = '\0'; retval = ptr2;
  }
  return(retval);
}

static void ask_download_to(void)
{ tResourceRequest* request = request_allocate();
  const char* filename;
  parse_uri(khmli_download_uri, request, NULL, NULL, NULL);
  filename = request2filename(request);
  __request_deallocate(request);
  line_input_estart(eligDownloadFilename, _("Download to: "), filename, NULL,
    _(strDownload), _(strDownload), _(strCancel));
  __dealloc(filename);
}

static one_caller void ask_yesno_act(tConfirmationGoal cg)
{ const tWindowView* view;
  switch (cg)
  { case cgQuit: do_quit(); /*@notreached@*/ break;
    case cgClose: rcw window_hide(current_window, 2); break;
#if TGC_IS_CURSES
    case cgOverwriteFile: rcw start_save_as(khmli_filename, truE); break;
    case cgOverwriteDump: rcw content_dump(khmli_filename, truE); break;
    case cgOverwriteDownload: content_download(truE); break;
#if CONFIG_SESSIONS
    case cgOverwriteSession: session_save(khmli_filename, truE); break;
#endif
#endif
    case cgSubmit: rcw submit_form(); break;
    case cgReset: rcw reset_form(); break;
    case cgRepost: rcw reload_document(); break;
    case cgHtml: rcw interpret_as_html(); break;
    case cgEnable: rcw
      view = current_window->current_view;
#if TGC_IS_GRAPHICS
      /* FIXME! */
#else
      view->request->resource->aebase[view->aecur].flags &= ~aefDisabled;
#endif
      break;
  }
}

typedef struct
{
#if CONFIG_USER_QUERY
  tUserQuery* query;
#endif
  tConfirmationGoal cg;
} tAskYesnoData;

static void ask_yesno_callback(void* _data, tLineInputEventKind liek, tKey key)
{ tAskYesnoData* data = (tAskYesnoData*) _data;
#if CONFIG_USER_QUERY
  tUserQuery* query = data->query;
#endif
  tConfirmationGoal cg = data->cg;
  switch (liek)
  { case liekFail:
#if CONFIG_USER_QUERY
      if (query != NULL) query->mif |= mifQueryFailed;
#endif
      goto do_cancel0; /*@notreached@*/ break;
    case liekCancel:
      do_cancel: {}
#if CONFIG_USER_QUERY
      if (query != NULL) query->mif |= mifUserCancelled;
#endif
      do_cancel0: line_input_finish(); memory_deallocate(data);
#if CONFIG_USER_QUERY
      if (query != NULL)
      { tObjectIdentifier id;
        handle_conf_query:
        id = query->id;
        if (id != oiUnknown)
        { tBoolean did_vanish = object_did_vanish(id);
          object_unwatch(id);
          if (did_vanish) query->mif |= mifObjectVanished;
        }
        resource_callback(query, mckUserResponse);
      }
#endif
      break;
#if TGC_IS_CURSES
    case liekKey:
      if ( (key == config.char_yes) || (key == KEY_ENTER) )
      { line_input_finish(); memory_deallocate(data);
#if CONFIG_USER_QUERY
        if (query != NULL) { goto handle_conf_query; } else
#endif
        { ask_yesno_act(cg); }
      }
      else if (key == config.char_no) goto do_cancel;
      break;
#endif
  }
}

static void ask_yesno(const char* msg, tConfirmationGoal cg, tUserQuery* query)
{ tAskYesnoData* data = __memory_allocate(sizeof(tAskYesnoData));
#if CONFIG_USER_QUERY
  data->query = query;
#endif
  data->cg = cg;
#if TGC_IS_GRAPHICS
#define buf msg /* (no string change necessary) */
#else
  sprint_safe(strbuf3, "%s (%c/%c) ", msg, config.char_yes, config.char_no);
#define buf strbuf3
#endif
  line_input_start(buf, NULL, ask_yesno_callback, data, liafNone,
    _("Confirmation Required"), istrYesUc, istrNoUc);
#undef buf
}

static tBoolean test_local_file_overwrite(/*const*/ char** _filename,
  tConfirmationGoal cg, tBoolean* _may_overwrite)
/* returns whether a local file may be overwritten right now; if not, this
   function either prepares a "may overwrite?" question for the user or
   produces an error message */
{ const char* filename = *_filename;
  struct stat statbuf;
  int err;

  *_may_overwrite = falsE;
  err = my_stat(filename, &statbuf);
  if (err != 0)
  { if (errno == ENOENT) { allow: return(truE); } /* nothing there */
    else { failed: show_message_osfailure(errno); }
  }
  else /* already there */
  { if (!S_ISREG(statbuf.st_mode)) /* only overwrite _files_ */
    { if (S_ISDIR(statbuf.st_mode)) errno = EISDIR;
      goto failed;
    }
    else if (config.flags & cfDontConfirmOverwrite)
    { *_may_overwrite = truE; goto allow; }
    else
    { __dealloc(khmli_filename); khmli_filename = filename; *_filename = NULL;
      ask_yesno(_("Overwrite existing?"), cg, NULL);
    }
  }
  return(falsE);
}

static void start_save_as(const char* filename, tBoolean may_overwrite)
{ tResourceRequest* new_request = request_allocate();
#if TGC_IS_GRAPHICS
  tResourceRequest* old_request = NULL; /* FIXME! */
#else
  tResourceRequest* old_request = current_window->current_view->request;
#endif
  tWindowView* new_view = window_append_view(current_window, new_request);
  new_view->title = my_strdup(filename);
  new_request->uri = my_strdup(filename);
  new_request->resource = old_request->resource;
  if (may_overwrite) new_request->flags |= rrfMayOverwrite;
  resource_callback(new_request, mckStartSaveAs);
  /* IMPROVEME: simply use the resource, no new request structure at all? */
}

typedef struct
{
#if CONFIG_USER_QUERY
  tUserQuery* query;
#endif
  tEditableLineInputGoal elig;
} tEstartData;

#if CONFIG_USER_QUERY

static void do_start_user_query_login(tUserQuery* query)
{ tResource* resource = query->resource;
  tMissingInformationFlags mif = query->mif;
  tBoolean pr = cond2boolean(mif & mifProxyrelated);
  const char *ps = (pr ? _("proxy") : _("server")), *lf, *what, *sep,
    *old_username, *userat; /* user_at? use_rat? :-) */
  char schemebuf[MAXSCHEMESTRSIZE], * spfbuf;
  tEditableLineInputGoal elig;
  if (mif & mifPriorLoginAttemptFailed)
  { lf = (pr ? _("Proxy login failed") : _(strLoginFailed));
    sep = strSpacedDash;
  }
  else lf = sep = strEmpty;

  rp2scheme(resource->protocol, schemebuf); old_username = userat = strEmpty;
  if (mif & mifUsername) { elig = eligUsername; what = _("username"); }
  else if (mif & mifPassword)
  { const char* u = resource->username;
    if (u != NULL) { old_username = u; userat = "@"; }
    elig = eligPassword; what = _("password");
  }
  else
  { /* bug */
    return;
  }

  my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("%s%s%s %s for %s %s%s%s:%d: "), lf,
    sep, schemebuf, what, ps, old_username, userat, query->hostname,
    ntohs(query->portnumber));
  line_input_estart(elig, spfbuf, NULL, query, _("User Input"), _("Apply"),
    _(strCancel));
  my_spf_cleanup(strbuf, spfbuf);
}

static void start_user_query(tUserQuery* query)
{
#if CONFIG_FTP && OPTION_TLS
  if (query->mif & mifFtpsDataclear)
  { const char* hostname = query->hostname;
    char schemebuf[MAXSCHEMESTRSIZE];
    tResource* resource = query->resource;
    if (strlen(hostname) > 1000) hostname = "[?]"; /* can't be serious */
    rp2scheme(resource->protocol, schemebuf);
    sprint_safe(strbuf,
      _("%s server %s:%d can't protect data - allow cleartext?"),
      schemebuf, hostname, ntohs(query->portnumber));
    ask_yesno(strbuf, cgFtpsDataclear, query);
  }
  else
#endif
  { do_start_user_query_login(query); }
#if TGC_IS_CURSES
  if (key_handling_mode == khmLineInput)
  { if (query->id != oiUnknown) object_watch(query->id);
  }
#endif
}

static void start_user_query_username(void)
/* go from "asking for password" back to "ask for username"; CHECKME: dirty! */
{ const tEstartData* data = (const tEstartData*) (lid.callback_data);
  tUserQuery* query = data->query;
  tObjectIdentifier id = query->id;
  if ( (id != oiUnknown) && (object_did_vanish(id)) )
  { (lid.callback)(lid.callback_data, liekFail, 0); return; }
  memory_deallocate(data); line_input_finish();
  query->mif &= ~mifPassword; query->mif |= mifUsername;
  do_start_user_query_login(query);
}

#endif /* #if CONFIG_USER_QUERY */

#if CONFIG_JUMPS

static tConfigJump* lookup_jump_shortcut(const char* shortcut)
{ tConfigJump* j = config.jumps;
  while (j != NULL)
  { const char* name = j->name;
    if ( (name != NULL) && (!strcmp(name, shortcut)) ) break; /* found */
    j = j->next;
  }
  return(j);
}

static one_caller void handle_lig_jump(const char* text)
{ char *text2 SHUT_UP_COMPILER(NULL), *uri SHUT_UP_COMPILER(NULL);
  const tConfigJump* j;
  tBoolean cleanup_text2, cleanup_uri;
  cleanup_text2 = cleanup_uri = falsE;
  my_strdedup(previous_jump, text);
  if (my_strchr(text, ' ') == NULL) /* no arguments given, just the shortcut */
  { j = lookup_jump_shortcut(text);
    if (j != NULL)
    { uri = j->uri;
      if ( (uri != NULL) && (uri[0] != '\0') )
      { prr:
        rcwc
        prepare_resource_request(current_window, uri, prrfRedrawOne, NULL,
          NULL);
      }
      else { bad_uri: show_message(_(strResourceError[reUri]), truE); }
    }
    else
    { char* spfbuf;
      not_found:
      my_spf(strbuf, STRBUF_SIZE, &spfbuf,
        _("Can't find the jump shortcut \"%s\""), text);
      show_message(spfbuf, truE);
      my_spf_cleanup(strbuf, spfbuf);
    }
  }
  else /* need to extract shortcut and replace markers with argument values */
  { char *ptr, *ptr2;
    unsigned short count;
    tBoolean is_last;
    text2 = my_strdup(text); cleanup_text2 = truE;
    ptr = text2; ptr2 = my_strchr(ptr, ' ');
    if (ptr2 == NULL) /* "can't happen" */
    { not_found2:
      text = ptr; goto not_found;
    }
    *ptr2 = '\0';
    j = lookup_jump_shortcut(ptr);
    if (j == NULL) goto not_found2;
    uri = j->uri;
    if ( (uri == NULL) || (uri[0] == '\0') ) goto bad_uri;
    uri = my_strdup(uri); cleanup_uri = truE;
    /* Try to replace all markers in the URI with arguments: */
    is_last = falsE;
    for (count = 0; count < j->argcount; count++)
    { const char *marker, *match, *src;
      char *new_uri, *dest;
      size_t markerlen, size;
      if (is_last) break; /* no more arguments given */
      spaceloop:
      ptr = ptr2 + 1;
      if (*ptr == '\0') break;
      ptr2 = my_strchr(ptr, ' ');
      if (ptr2 != NULL)
      { if (ptr2 == ptr) goto spaceloop; /* skip accidental space sequence */
        *ptr2 = '\0';
      }
      else
      { ptr2 = ptr;
        while (*ptr2 != '\0') ptr2++; /* set it onto the trailing '\0' byte */
        is_last = truE;
      }
      marker = j->arg[count];
      if ( (marker == NULL) || (marker[0] == '\0') ) /* "can't happen" */
        continue;
      match = my_strstr(uri, marker);
      if (match == NULL)
      { char* spfbuf;
        my_spf(strbuf, STRBUF_SIZE, &spfbuf,
          _("Can't find marker \"%s\" in URL pattern"), marker);
        show_message(spfbuf, truE);
        my_spf_cleanup(strbuf, spfbuf);
        goto cleanup;
      }
      markerlen = strlen(marker);
      size = strlen(uri) - markerlen + strlen(ptr) + 1;
      new_uri = __memory_allocate(size);
      src = uri; dest = new_uri;
      while (src < match) *dest++ = *src++;
      while (ptr < ptr2) *dest++ = *ptr++;
      src += markerlen;
      while ( (*dest++ = *src++) != '\0' ) { /* nothing */ }
      memory_deallocate(uri); uri = new_uri;
    }
    goto prr;
  }
  cleanup:
  if (cleanup_text2) memory_deallocate(text2);
  if (cleanup_uri) memory_deallocate(uri);
}

#endif /* #if CONFIG_JUMPS */

#if OPTION_EXECEXT & 1

static one_caller void handle_lig_execext_shell(const char* text)
{ tExecextShellData data;
  tExecextShellFlags esf = esfNone;
  tResource* resource = NULL;
  const char *cont = NULL, *cont2;
  char* spfbuf_header = unconstify(strEmpty);
  size_t header_len, content_size, writedata_size;
  tBoolean must_dealloc_command, want_header, want_content, interpret_percents;
  must_dealloc_command = want_header = want_content = interpret_percents=falsE;
  my_memclr_var(data);
  if (*text == ':') /* options given */
  { char ch;
    while ( (ch = *++text) != '\0' )
    { switch (ch)
      { case 'o': esf |= esfReadStdout; break;
        case 'e': esf |= esfReadStderr; break;
        case 'H': esf |= esfEnforceHtml; break;
        case 'h': want_header = truE; break;
        case 'c': want_content = truE; break;
        case 'i': interpret_percents = truE; break;
        case ' ': text++; goto end_of_options; /*@notreached@*/ break;
      }
    }
    end_of_options: {}
  }
  if (*text == '\0') {show_message(_("No shell command given"), truE); return;}
  if (current_window == NULL) esf &= ~(esfReadStdout | esfReadStderr);
  data.esf = esf;

  { const tWindowView* view;
    const tResourceRequest* req;
    if ( (current_window != NULL) && ( (view = current_window->current_view)
         != NULL ) && ( (req = view->request) != NULL ) )
      resource = req->resource;
  }

  if (!interpret_percents) { simple_command: data.command = text; }
  else
  { const char *temp = my_strstr(text, "%U"), *uri;
    char* convcomm;
    if (temp == NULL) goto simple_command;
    if ( (resource != NULL ) && ( (uri = resource->uri) != NULL ) &&
         (*uri != '\0') )
    { const size_t len_before = temp - text, len_uri = strlen(uri),
        len_after = strlen(temp + 2);
      char* ptr = convcomm = __memory_allocate(len_before + len_uri + len_after
        + 1);
      if (len_before > 0) { my_memcpy(ptr, text, len_before); ptr+=len_before;}
      my_memcpy(ptr, uri, len_uri); ptr += len_uri;
      if (len_after > 0) { my_memcpy(ptr, temp+2, len_after); ptr+=len_after; }
      *ptr = '\0';
    }
    else /* just remove the "%U" */
    { size_t len_before = temp - text;
      convcomm = __memory_allocate(strlen(text) - 2 + 1);
      if (len_before > 0) my_memcpy(convcomm, text, len_before);
      strcpy(convcomm + len_before, temp + 2);
    }
    data.command = convcomm; must_dealloc_command = truE;
  }

  if (data.esf & (esfReadStdout | esfReadStderr))
  { tResourceRequest* request = data.request = request_allocate();
    tWindowView* view;
    char* spfbuf;
    request->uri = my_strdup("execext-shell:"); /* dummy - CHECKME! */
    request->protocol = rpExecextShell;
    view = window_append_view(current_window, request);
    my_spf(NULL, 0, &spfbuf, _("Result of \"%s\""), text);
    view->html_title = my_spf_use(spfbuf);
    view->title = my_strdup(request->uri);
    if (esf & esfEnforceHtml) view->flags |= wvfEnforceHtml;
  }

  header_len = content_size = writedata_size = 0;
  if (want_header)
  { my_spf(strbuf, STRBUF_SIZE, &spfbuf_header, "V:%s\n\n", strSoftwareId);
    header_len = strlen(spfbuf_header);
  }
  if (want_content)
  { if ( (resource != NULL ) && ( (cont = resource->content) != NULL ) )
    { cont2 = cont;
      while (cont2 != NULL)
      { content_size += CONTENTBLOCK_USEDSIZE(cont2);
        cont2 = CONTENTBLOCK_NEXT(cont2);
      }
    }
    if (content_size <= 0) want_content = falsE; /* no content there */
  }
  writedata_size = header_len + content_size;
  if (writedata_size > 0) /* user wants to write something (header/content) */
  { char* writedata = __memory_allocate(writedata_size), *temp;
    data.writedata = writedata; data.writedata_size = writedata_size;
    if (header_len > 0) my_memcpy(writedata, spfbuf_header, header_len);
    temp = writedata + header_len; cont2 = cont;
    while (cont2 != NULL)
    { const size_t usedsize = CONTENTBLOCK_USEDSIZE(cont2);
      if (usedsize > 0)
      { my_memcpy(temp, CONTENTBLOCK_DATA(cont2), usedsize); temp += usedsize;}
      cont2 = CONTENTBLOCK_NEXT(cont2);
    }
  }
  if (spfbuf_header != strEmpty) my_spf_cleanup(strbuf, spfbuf_header);

  resource_callback((void*)(&data), mckExecextShell);
  if (must_dealloc_command) memory_deallocate(data.command);
  __dealloc(data.writedata);
}

#endif /* #if OPTION_EXECEXT & 1 */

static one_caller void line_input_estart_act(tEditableLineInputGoal elig,
  /*const*/ char** _text)
{ const char* text = *_text;
  tWindowRedrawingTask wrt;
  switch (elig)
  { case eligGoToUri:
      rcwc
      prepare_resource_request(current_window, text, prrfRedrawOne, NULL,NULL);
      break;
    case eligSaveAs:
      { tBoolean may_overwrite;
        rcw
        if (test_local_file_overwrite(_text, cgOverwriteFile, &may_overwrite))
          start_save_as(text, may_overwrite);
      }
      break;
    case eligDumpFilename:
      { tBoolean may_overwrite;
        if (test_local_file_overwrite(_text, cgOverwriteDump, &may_overwrite))
          content_dump(text, may_overwrite);
      }
      break;
    case eligDownloadUri:
      my_strdedup(khmli_download_uri, text); ask_download_to();
      break;
    case eligDownloadFilename:
      { tBoolean may_overwrite;
        if(test_local_file_overwrite(_text,cgOverwriteDownload,&may_overwrite))
          __content_download(text, may_overwrite);
      }
      break;
    case eligDocumentSearch:
      wrt = wrtSearch;
      handle_search:
      __dealloc(search_string); search_string = text; *_text = NULL;
      __window_redraw(current_window, wrt);
      break;
    case eligDocumentSearchBackward:
      wrt = wrtSearchBackward; goto handle_search; /*@notreached@*/ break;
#if CONFIG_JUMPS
    case eligJump: handle_lig_jump(text); break;
#endif
#if TGC_IS_CURSES
    case eligFormText: case eligFormPassword: case eligFormFilename:
      { tWindowView* view;
        tActiveElement* aes;
        tActiveElementNumber _ae;
        view = current_window->current_view;
        aes = view->active_element; _ae = view->aecur;
        if (text[0] != '\0')
          my_strdedup(aes[_ae].current_text, text);
        else dealloc(aes[_ae].current_text);
        /* IMPLEMENTME: "javascript_handle_event(jekChange, _ae);" if
           appropriate */
        javascript_handle_event(jekBlur, _ae);
        window_redraw(current_window);
      }
      break;
#endif
#if CONFIG_SESSIONS
    case eligSessionSave:
      { tBoolean may_overwrite;
        if (test_local_file_overwrite(_text,cgOverwriteSession,&may_overwrite))
          session_save(text, may_overwrite);
      }
      break;
    case eligSessionResume: session_resume(text); break;
#endif
#if OPTION_EXECEXT & 1
    case eligExecextShell: handle_lig_execext_shell(text); break;
#endif
  }
}

#if TGC_IS_CURSES
static my_inline void __curleft(void)
{ (void) move(stdscr->_cury, stdscr->_curx - 1);
}
static my_inline void __curright(void)
{ (void) move(stdscr->_cury, stdscr->_curx + 1);
}
#endif

static void line_input_estart_callback(void* _data, tLineInputEventKind liek,
  tKey key)
{ tEstartData* data = (tEstartData*) _data;
#if CONFIG_USER_QUERY
  tUserQuery* query = data->query;
#endif
  tEditableLineInputGoal elig = data->elig;
  tLineInputActionCode liac;
  switch (liek)
  { case liekFail:
#if CONFIG_USER_QUERY
      if (query != NULL) query->mif |= mifQueryFailed;
#endif
      goto do_cancel0; /*@notreached@*/ break;
    case liekCancel:
      do_cancel: {}
#if CONFIG_USER_QUERY
      if (query != NULL) query->mif |= mifUserCancelled;
#endif
      do_cancel0: line_input_finish(); memory_deallocate(data);
#if CONFIG_USER_QUERY
      if (query != NULL)
      { tObjectIdentifier id = query->id;
        if (id != oiUnknown)
        { tBoolean did_vanish = object_did_vanish(id);
          object_unwatch(id);
          if (did_vanish) query->mif |= mifObjectVanished;
        }
        resource_callback(query, mckUserResponse);
      }
#endif
      break;
#if TGC_IS_CURSES
    case liekKey:
      liac = lookup_lineinput_key(key);
      switch (liac)
      { case liacUnknown: break; /* the most likely case first */
        case liacCancel: goto do_cancel; /*@notreached@*/ break;
        case liacToLeft:
          if (lid_area(pos) > 0)
          { lid_area(pos)--;
            if (lid_area(colcurr) > lid_area(colmin))
            { __curleft(); lid_area(colcurr)--; }
            else { lid_area(first)--; line_input_redraw(); }
          }
          return; /*@notreached@*/ break;
        case liacToRight:
          if (lid_area(pos) < lid_area(len))
          { lid_area(pos)++;
            if (lid_area(colcurr) < lid_area(colmax))
            { __curright(); lid_area(colcurr)++; }
            else { lid_area(first)++; line_input_redraw(); }
          }
          return; /*@notreached@*/ break;
        case liacToStart:
          if (lid_area(pos) > 0)
          { lid_area(pos) = lid_area(first) = 0;
            lid_area(colcurr) = lid_area(colmin);
            line_input_redraw();
          }
          return; /*@notreached@*/ break;
        case liacToEnd: line_input_to_eos(); return; /*@notreached@*/ break;
        case liacPass2user:
#if CONFIG_USER_QUERY
          if ( (elig == eligPassword) && (query != NULL) )
            start_user_query_username();
#endif
          return; /*@notreached@*/ break;
      }

      if (key == KEY_ENTER)
      { if ( (lid_area(len) > 0) || (lid_area(flags) & liafAllowEmptyText) )
        { char* text;
          line_input_ensure_room(); text = lid_area(text);
          text[lid_area(len)] = '\0'; lid_area(text) = NULL;
          line_input_finish(); memory_deallocate(data);
#if CONFIG_DEBUG
          sprint_safe(strbuf, "line input: *%s*\n", text);
          debugmsg(strbuf);
#endif
#if CONFIG_USER_QUERY
          if (query != NULL)
          { tObjectIdentifier id = query->id;
            tResource* resource;
            if (id != oiUnknown)
            { tBoolean did_vanish = object_did_vanish(id);
              object_unwatch(id);
              if (did_vanish)
              { query->mif |= mifObjectVanished; goto do_resp; }
            }
            resource = query->resource; query->mif &= ~mifSet;
            if (elig == eligUsername)
            { resource->username = text; text = NULL;
              query->mif |= mifUsername;
            }
            else if (elig == eligPassword)
            { resource->password = text; text = NULL;
              query->mif |= mifPassword;
            }
            do_resp: resource_callback(query, mckUserResponse);
          }
          else
#endif
          { require_normal_current_window(); /* CHECKME! */
            line_input_estart_act(elig, &text);
          }
          __dealloc(text);
        }
      }
      else if (key == KEY_BACKSPACE)
      { if (lid_area(pos) > 0)
        { char* text = lid_area(text);
          short count;
          /* update the string */
          for (count = lid_area(pos); count < lid_area(len); count++)
            text[count - 1] = text[count];
          lid_area(pos)--; lid_area(len)--;
          /* update the screen */
          if (lid_area(colcurr) <= lid_area(colmin)) lid_area(first)--;
          else { lid_area(colcurr)--; line_input_redraw(); }
        }
      }
      else if (key == KEY_DC)
      { if (lid_area(pos) < lid_area(len))
        { char* text = lid_area(text);
          short count;
          /* update the string */
          for (count = lid_area(pos); count < lid_area(len) - 1; count++)
            text[count] = text[count + 1];
          lid_area(len)--;
          /* update the screen */
          line_input_redraw();
        }
      }
      else if ( (key >= ' ') && (key <= 255) && (key != 127) &&
                (lid_area(len) < 1234) ) /* CHECKME: OPTION_CED? */
      { char ch = (char) key, *text;
        short count;
        /* update the string */
        line_input_ensure_room();
        text = lid_area(text);
        for (count = lid_area(len); count > lid_area(pos); count--)
          text[count] = text[count - 1];
        text[lid_area(pos)++] = ch; lid_area(len)++;
        /* update the screen */
        if (lid_area(colcurr) < lid_area(colmax)) lid_area(colcurr)++;
        else lid_area(first)++;
        line_input_redraw();
      }
      break;
#endif /* #if TGC_IS_CURSES */
  }
}

static void __line_input_estart(tEditableLineInputGoal elig, const char* msg,
  const char* initstr, tUserQuery* query)
{ tLineInputAreaFlags liaf = liafEditable;
  tEstartData* data = __memory_allocate(sizeof(tEstartData));
#if CONFIG_USER_QUERY
  data->query = query;
#endif
  data->elig = elig;
#if TGC_IS_CURSES
  if (is_disguising_elig(elig)) liaf |= liafDisguising;
  if (is_emptyok_elig(elig)) liaf |= liafAllowEmptyText;
#endif
  __line_input_start(msg, initstr, line_input_estart_callback, data, liaf);
}


/* Keys and command codes */

static one_caller void remap_key(tKey* _key)
{ switch (*_key)
  { case KEY_ESCAPE: *_key = KEY_CANCEL; break;
    case '\015': *_key = KEY_ENTER; break;
    case '\010': *_key = KEY_BACKSPACE; break;
    case 127: *_key = KEY_DC; break;
  }
}

static void handle_command_code(tProgramCommandCode code)
{ tLinenumber scrolloff;
  const char *temp, *text, *msg;
  tEditableLineInputGoal elig;
  switch (code)
  {case pccViewBack:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      if ( (view != NULL) && (view->prev != NULL) )
      { current_window->current_view = view->prev;
        window_redraw(current_window);
      }
    }
    break;
   case pccViewForward:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      if ( (view != NULL) && (view->next != NULL) )
      { current_window->current_view = view->next;
        window_redraw(current_window);
      }
    }
    break;
   case pccPageDown:
    scrolloff = current_view_height();
    check_scroll_down:
    if (is_normal_window(current_window))
    { tWindowView* view;
      scroll_down:
      view = current_window->current_view;
      if ( (view != NULL) && (view->flags & wvfScreenFull) )
      { view->origin_y += scrolloff; view->flags &= ~wvfScrollingUp;
#if TGC_IS_CURSES
        view->aecur = INVALID_AE;
#endif
        window_redraw(current_window);
      }
    }
    break;
   case pccPageUp:
    scrolloff = -current_view_height();
    check_scroll_up:
    if (is_normal_window(current_window))
    { tWindowView* view;
      scroll_up:
      view = current_window->current_view;
      if (view != NULL)
      { tLinenumber y = view->origin_y + scrolloff;
        if (y < 0) y = 0;
        if (view->origin_y > y) /* actually need to scroll */
        { view->origin_y = y; view->flags |= wvfScrollingUp;
#if TGC_IS_CURSES
          view->aecur = INVALID_AE;
#endif
          window_redraw(current_window);
        }
      }
    }
    break;
   case pccLineDown:
    scrolloff = 1; goto check_scroll_down; /*@notreached@*/ break;
   case pccLineUp:
    scrolloff = -1; goto check_scroll_up; /*@notreached@*/ break;
#if TGC_IS_GRAPHICS
#define connect_sel(b, h, d) \
  connect_object(GTK_OBJECT(GTK_FILE_SELECTION(sel)->b), strGtkClicked, h, d)
   case pccLocalFileDirOpen:
    { GtkWidget* sel = gtk_file_selection_new(_("Open Local File/Directory"));
      connect_sel(ok_button, filesel_handle,
        filesel_data_create(sel, current_window));
      connect_sel(ok_button, graphics_window_destroy, sel);
      connect_sel(cancel_button, graphics_window_destroy, sel);
      show_widget(sel);
    }
    break;
#else
   case pccElementNext:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      if (view != NULL)
      { tActiveElementNumber _ae = view->aecur, _ae2;
        if ( (_ae != INVALID_AE) &&
             ( (_ae2 = next_visible_ae(view, _ae)) != INVALID_AE ) )
        { activate_element(view, _ae2); }
        else { scrolloff = current_view_height(); goto scroll_down; }
      }
    }
    break;
   case pccElementPrevious:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      if (view != NULL)
      { tActiveElementNumber _ae = view->aecur, _ae2;
        if ( (_ae != INVALID_AE) &&
             ( (_ae2 = previous_visible_ae(view, _ae)) != INVALID_AE ) )
        { activate_element(view, _ae2); }
        else { scrolloff = -current_view_height(); goto scroll_up; }
      }
    }
    break;
   case pccElementOpen: case pccWindowNewFromElement: case pccElementOpenSplit:
    { tWindowView* view;
      tActiveElementNumber _ae;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (_ae = view->aecur) != INVALID_AE ) &&
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) )
      { tActiveElementBase* aebase = resource->aebase;
        tActiveElementKind kind;
        tActiveElement* aes;
        if (aebase[_ae].flags & aefDisabled)
        { show_message(_("This element is disabled!"), truE);
          return;
        }
        javascript_handle_event(jekFocus, _ae);
        aes = view->active_element;
        kind = aebase[_ae].kind;
        switch (kind)
        {case aekLink:
          temp = aebase[_ae].data;
          if ( (temp != NULL) && (temp[0] != '\0') )
          { tWindow* destwin;
            tPrrFlags prrf = prrfRedrawOne;
            if (code == pccElementOpen) destwin = current_window;
            else
            { destwin = window_create();
              if (code == pccElementOpenSplit)
              { visible_window[1 - current_window_index] = destwin;
                prrf = prrfRedrawAll;
              }
              else current_window = destwin;
            }
            prepare_resource_request(destwin, temp, prrf, view, NULL);
          }
          break;
         case aekFormSubmit: case aekFormImage:
          goto handle_submit; /*@notreached@*/ break;
         case aekFormReset:
          goto handle_reset; /*@notreached@*/ break;
         case aekFormText:
          elig = eligFormText;
          start_input:
          if (aebase[_ae].flags & aefReadonly)
          { /* IMPLEMENTME: the user should be able to see the _whole_ contents
               of the element somehow, not only what is displayed within the
               normal web page layout! (E.g. the text might be clipped.) */
            readonly:
            show_message(_("This element is read-only!"), truE);
            return;
          }
          else
          { sprint_safe(strbuf3, "%s: ", _(strAek[kind]));
            __line_input_estart(elig, strbuf3, aes[_ae].current_text, NULL);
            /* CHECKME: respect the "maxlength" attribute for eligFormText and
               eligFormPassword? */
          }
          break;
         case aekFormPassword:
          elig = eligFormPassword; goto start_input; /*@notreached@*/ break;
         case aekFormFile:
          elig = eligFormFilename; goto start_input; /*@notreached@*/ break;
         case aekFormCheckbox:
          if (aebase[_ae].flags & aefReadonly) goto readonly;
          else
          { view->active_element[_ae].flags ^= aefCheckedSelected;
            window_redraw(current_window); /* IMPROVEME? */
          }
          break;
         case aekFormRadio:
          if (aebase[_ae].flags & aefReadonly) goto readonly;
          else if (!(aes[_ae].flags & aefCheckedSelected))
          { tHtmlFormNumber fn = ae2form(resource, _ae);
            const char* name = aebase[_ae].data;
            aes[_ae].flags |= aefCheckedSelected;
            /* Uncheck all other radio buttons of the same name in this form */
            if ( (fn != INVALID_HTML_FORM_NUMBER) && (name != NULL) )
            { tActiveElementNumber a, a1 = resource->form[fn].first_ae,
                a2 = resource->form[fn].last_ae;
              if ( (a1 != INVALID_AE) && (a2 != INVALID_AE) )
              { for (a = a1; a <= a2; a++)
                { if ( (a != _ae) && (aebase[a].kind == aekFormRadio) )
                  { const char* n = aebase[a].data;
                    if ( (n != NULL) && (!my_strcasecmp(name, n)) )
                      aes[a].flags &= ~aefCheckedSelected;
                  }
                }
              }
            }
            window_redraw(current_window); /* IMPROVEME? */
          }
          break;
         case aekFormSelect:
          if (aebase[_ae].maxlength > 0)
          {
#if OPTION_MENUS
            cm_setup_select_tag(aebase, view, _ae);
#else
            menus_were_disabled();
#endif
          }
          else show_message(_(strSelectionEmpty), truE);
          break;
        } /* switch */
      }
    }
    break;
   case pccElementInfo:
    { tWindowView* view;
      tActiveElementNumber _ae;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (_ae = view->aecur) != INVALID_AE ) &&
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) )
      { tActiveElementFlags flags = resource->aebase[_ae].flags;
        const char *dat = null2empty(resource->aebase[_ae].data),
          *disabl = ( (flags & aefDisabled) ? (_("disabled ")) : strEmpty ),
          *readonl = ( (flags & aefReadonly) ? (_("read-only ")) : strEmpty ),
          *mult = ( (flags & aefMultiple) ? (_("multi-option ")) : strEmpty );
        char* spfbuf;
        my_spf(strbuf, STRBUF_SIZE, &spfbuf, "%s%s%s%s%s%s", disabl, readonl,
          mult, _(strAek[resource->aebase[_ae].kind]),
          ((dat[0] != '\0') ? ", " : strEmpty), dat);
        show_message(spfbuf, falsE);
        my_spf_cleanup(strbuf, spfbuf);
      }
    }
    break;
   case pccElementEnable:
    { tWindowView* view;
      tActiveElementNumber _ae;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (_ae = view->aecur) != INVALID_AE ) &&
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) &&
           (resource->aebase[_ae].flags & aefDisabled) )
      { if (config.flags & cfDontConfirmEnable)
          resource->aebase[_ae].flags &= ~aefDisabled;
        else ask_yesno(_("Really enable this element?"), cgEnable, NULL);
      }
    }
    break;
   case pccScreenUnsplit:
    if (visible_window[1] != NULL)
      window_hide(visible_window[1 - current_window_index], 1);
    break;
   case pccScreenSplit:
    if (visible_window[1] == NULL)
    { tWindowView* view = current_window->current_view;
      if (view != NULL)
      { tActiveElementNumber ae = view->aecur;
        if (ae != INVALID_AE) do_activate_element(view, ae, falsE);
      }
      visible_window[1] = window_create();
      current_window_index = 1;
      window_redraw_all();
    }
    break;
   case pccScreenSwitch:
    if (visible_window[1] != NULL)
    { current_window_index = 1 - current_window_index;
      window_show_info(current_window);
    }
    break;
   case pccWindowNext:
    { tWindow* window = current_window;
      tBoolean is_first_try = truE;
      findloop1:
      window = window->next;
      if (window == NULL) window = windowlisthead;
      if (window != NULL)
      { if (window_is_visible(window))
        { if (is_first_try) { is_first_try = falsE; goto findloop1; }
          else return;
        }
        current_window = window;
        window_redraw(current_window);
      }
    }
    break;
   case pccWindowPrevious:
    { tWindow* window = current_window;
      tBoolean is_first_try = truE;
      findloop2:
      window = window->prev;
      if (window == NULL) window = windowlisttail;
      if (window != NULL)
      { if (window_is_visible(window))
        { if (is_first_try) { is_first_try = falsE; goto findloop2; }
          else return;
        }
        current_window = window;
        window_redraw(current_window);
      }
    }
    break;
#if OPTION_MENUS
   case pccMenuWindowlist:
     cm_setup_window_list(); break;
   case pccMenuContextual:
    { tWindowView* view = current_window->current_view;
      tActiveElementNumber _ae = ((view != NULL) ? (view->aecur) : INVALID_AE);
      cm_setup(-1, -1, _ae);
    }
    break;
   case pccMenuUriHistory:
    cm_setup_uri_history(); break;
#else
   case pccMenuWindowlist: case pccMenuContextual: case pccMenuUriHistory:
     menus_were_disabled(); break;
#endif
#endif /* #if TGC_IS_GRAPHICS */
   case pccDocumentTop:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      if ( (view != NULL) && (view->origin_y > 0) )
      { view->origin_y = 0; view->flags |= wvfScrollingUp;
#if TGC_IS_CURSES
        view->aecur = INVALID_AE;
#endif
        window_redraw(current_window);
      }
    }
    break;
   case pccDocumentBottom:
    if (is_normal_window(current_window))
      __window_redraw(current_window, wrtToEnd);
    break;
   case pccGoUri:
    text = _("Go to URL: "); temp = NULL;
    go_to_uri:
    line_input_estart(eligGoToUri, text, temp, NULL, _("Enter URL"), _("Go"),
      _(strCancel));
    break;
   case pccGoUriPreset:
    { tWindowView* view = current_window->current_view;
      if (view != NULL)
      { text = _("Edit URL: "); temp = view->title; goto go_to_uri; }
    }
    break;
   case pccGoHome:
    if (config.home_uri != NULL)
    { require_normal_current_window();
      prepare_resource_request(current_window, config.home_uri, prrfRedrawOne,
        NULL, NULL);
    }
    else show_message(_("No home URL configured"), truE);
    break;
   case pccGoSearch:
    if (config.search_engine != NULL)
    { require_normal_current_window();
      prepare_resource_request(current_window, config.search_engine,
        prrfRedrawOne, NULL, NULL);
    }
    else show_message(_("No search engine URL configured"), truE);
    break;
   case pccGoBookmarks:
    if (config.bookmarks != NULL)
    { require_normal_current_window();
      prepare_resource_request(current_window, config.bookmarks, prrfRedrawOne,
        NULL, NULL);
    }
    else show_message(_("No bookmarks document URL configured"), truE);
    break;
#if CONFIG_JUMPS
   case pccJump:
    temp = NULL;
    jump:
    if (config.jumps != NULL)
    { line_input_estart(eligJump, _("Jump to: "), temp, NULL, _("Jump to URL"),
        _("Jump"), _(strCancel));
    }
    else show_message(_("No jump shortcuts configured"), truE);
    break;
   case pccJumpPreset:
    temp = previous_jump; goto jump; /*@notreached@*/ break;
#else
   case pccJump: case pccJumpPreset: fwdact(_(strFjumps)); break;
#endif
   case pccWindowNew:
    if (current_window->current_view != NULL)
    { current_window = window_create();
      window_redraw(current_window);
    }
    break;
   case pccWindowNewFromDocument:
    { tWindowView* view = current_window->current_view;
      if ( (view != NULL) && (view->title != NULL) )
      { tResourceRequest* request = view->request;
        tResource* resource;
        if (! (( (request != NULL) && ( (resource = request->resource) != NULL)
                && (resource->flags & rfCannotReload) )) )
        { char* uri = my_strdup(view->title);
          current_window = window_create();
          prepare_resource_request(current_window, uri, prrfRedrawOne, NULL,
            NULL);
          memory_deallocate(uri);
        }
      }
    }
    break;
   case pccWindowClose:
    if (config.flags & cfDontConfirmClose) window_hide(current_window, 2);
    else ask_yesno(_("Really close this window?"), cgClose, NULL);
    break;
   case pccDocumentEnforceSource:
    { tWindowView* view = current_window->current_view;
      if ( (view == NULL) || (view->flags & wvfSource) ) return;
      prepare_resource_request(current_window, view->title,
        prrfRedrawOne | prrfSource, NULL, NULL);
    }
    break;
   case pccDocumentReload: case pccDocumentReloadEnforced:
    { tWindowView* view;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (request = view->request) != NULL ) &&
           ( (code == pccDocumentReloadEnforced) ||
             ( ( (resource = request->resource) != NULL ) &&
               (!(resource->flags & rfCannotReload)) &&
               (resource->uri != NULL) ) ) )
      { tBoolean is_post = cond2boolean(view->flags & wvfPost);
        lid.prrf = ( (code == pccDocumentReloadEnforced)
          ? prrfEnforcedReload : prrfReload );
        if (is_post) lid.prrf |= prrfPost;
        if ( (!is_post) || (config.flags & cfDontConfirmRepost) )
          reload_document();
        else ask_yesno(_("Really repost form data?"), cgRepost, NULL);
      }
    }
    break;
   case pccDocumentSave:
    { tWindowView* view;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) /* &&
           (!(resource->flags & rfCannotReload)) -- CHECKME! */ )
      { const char* filename = request2filename(request);
        line_input_estart(eligSaveAs, _("Save document as: "), filename, NULL,
          _("Save Document"), _(strSave), _(strCancel));
        __dealloc(filename);
      }
    }
    break;
   case pccDump:
    { tWindowView* view;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) )
      { line_input_estart(eligDumpFilename, _("Dump to: "), NULL, NULL,
          _("Dump to File"), _("Dump"), _(strCancel));
      }
    }
    break;
   case pccDownload:
     dealloc(khmli_download_uri);
     line_input_estart(eligDownloadUri, _("Download from: "), NULL, NULL,
       _(strDownload), _("Proceed"), _(strCancel));
     break;
   case pccDownloadFromElement:
    { const tWindowView* view;
      const tResourceRequest* request;
      const tResource* resource;
      const tActiveElementBase* aebase;
      tActiveElementNumber _ae;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
#if TGC_IS_GRAPHICS
           ( (_ae = INVALID_AE) != INVALID_AE ) && /* FIXME! */
#else
           ( (_ae = view->aecur) != INVALID_AE ) &&
#endif
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) &&
           ( (aebase = resource->aebase) != NULL ) )
      { const tActiveElementKind aek = aebase[_ae].kind;
        if (aek == aekLink)
        { temp = aebase[_ae].data;
          if ( (temp != NULL) && (temp[0] != '\0') )
          { my_strdedup(khmli_download_uri, temp); ask_download_to(); }
        }
        else show_message(_("That's not a link"), truE);
      }
    }
    break;
   case pccDocumentEnforceHtml:
    { tWindowView* view;
      tResourceRequest* request;
      tResource* resource;
      if ( (is_normal_window(current_window)) &&
           ( (view = current_window->current_view) != NULL ) &&
           ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) )
      { if (config.flags & cfDontConfirmHtml) interpret_as_html();
        else ask_yesno(_("Really interpret as HTML?"), cgHtml, NULL);
      }
    }
    break;
   case pccFormReset:
    handle_reset:
    /* CHECKME: only ask if actually inside a form, otherwise fail */
    if (config.flags & cfDontConfirmReset) reset_form();
    else ask_yesno(_("Really reset this form?"), cgReset, NULL);
    break;
   case pccFormSubmit:
    handle_submit:
    /* CHECKME: only ask if actually inside a form, otherwise fail */
    if (config.flags & cfDontConfirmSubmit) submit_form();
    else ask_yesno(_("Really submit this form?"), cgSubmit, NULL);
    break;
   case pccStop:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      tResourceRequest* request;
      if ( (view != NULL) && ( (request = view->request) != NULL) )
        resource_callback(request, mckStopActivity);
    }
    break;
   case pccDocumentSearch:
    elig = eligDocumentSearch; msg = _("Search string: ");
    check_search:
    if (is_normal_window(current_window))
    { tWindowView* view = current_window->current_view;
      tResourceRequest* request;
      tResource* resource;
      if ( (view != NULL) && ( (request = view->request) != NULL ) &&
           ( (resource = request->resource) != NULL ) )
      { line_input_estart(elig, msg, search_string, NULL, _("Search String"),
          _(strSearch), _(strCancel));
      }
    }
    break;
   case pccDocumentSearchBackward:
    elig = eligDocumentSearchBackward; msg = _("Search string backwards: ");
    goto check_search; /*@notreached@*/ break;
   case pccDocumentSearchNext:
    if ( (search_string != NULL) && (is_normal_window(current_window)) )
      __window_redraw(current_window, wrtSearch);
    break;
   case pccDocumentSearchPrevious:
    if ( (search_string != NULL) && (is_normal_window(current_window)) )
      __window_redraw(current_window, wrtSearchBackward);
    break;
#if OPTION_EXECEXT & 1
   case pccExecextShell:
    if ( (config.flags & cfExecextShellCustom) &&
         (config.execext_shell == NULL) )
    { show_message(_(strResourceError[reConfigForbids]), truE); }
    else
    { line_input_estart(eligExecextShell, _("Execute shell command: "), NULL,
        NULL, _("Execute Shell Command"), _("Execute"), _(strCancel));
    }
    break;
   case pccExecextShellFlip:
    config.flags ^= cfExecextShellCustom;
    sprint_safe(strbuf, _("execext-shell configuration: %s"), (config.flags &
      cfExecextShellCustom) ? _("customized") : _("standard"));
    show_message(strbuf, falsE);
    break;
#else
   case pccExecextShell: case pccExecextShellFlip:
     fwdact(_(strFexecext)); break;
#endif
#if CONFIG_SESSIONS
   case pccSessionSave:
     { line_input_estart(eligSessionSave, _("Save session as: "),
         config.session_default, NULL, _("Save Session"), _(strSave),
         _(strCancel));
     }
     break;
   case pccSessionResume:
     { line_input_estart(eligSessionResume, _("Resume session from: "),
         config.session_default, NULL, _("Resume Session"), _("Resume"),
         _(strCancel));
     }
     break;
#else
   case pccSessionSave: case pccSessionResume: fwdact(_(strFsessions)); break;
#endif
   case pccMouseFlip: case pccMouseOff: case pccMouseOn:
     mouse_flip(code - pccMouseFlip); break;
   case pccQuit:
    if (config.flags & cfDontConfirmQuit) do_quit();
    else ask_yesno(_("Really quit retawq?"), cgQuit, NULL);
    break;
  }
}

#if TGC_IS_CURSES

#if CONFIG_DO_TEXTMODEMOUSE
static one_caller void handle_key_mouse(void)
{ MEVENT event;
  short x, _y;
  tLinenumber y, yt;
  tVisibleWindowIndex i;
  tWindow* window;
  tWindowView* view;
  const tActiveElement* aes;
  tActiveElementNumber _ae;
  mmask_t mask;
  if (getmouse(&event) != OK) return;
  mask = event.bstate;
  if (!(mask & TEXTMODEMOUSE_MASK)) return; /* "can't happen" */
  x = event.x; _y = event.y; y = (tLinenumber) _y;
  if (!is_bottom_occupied)
  { sprint_safe(strbuf, _("mouse click: %d, %d"), x, _y);
    show_message(strbuf, falsE);
  }
  if (key_handling_mode == khmLineInput)
  { if ( (lid.curr == 1) && (y == lid_area(row)) && (x >= lid_area(colmin)) &&
         (x <= lid_area(colmax)) )
    { short offset = x - lid_area(colcurr);
      if (offset < 0)
      { li_off:
        lid_area(colcurr) += offset; lid_area(pos) += offset;
        (void) move(lid_area(row), lid_area(colcurr));
      }
      else if (offset > 0)
      { short pos = lid_area(pos) + offset;
        if (pos > lid_area(len)) pos = lid_area(len);
        if (pos > lid_area(pos)) { offset = pos - lid_area(pos); goto li_off; }
      }
      /* "else": offset == 0, nothing changes */
    }
    return;
  }
#if OPTION_MENUS
  else if (key_handling_mode == khmContextualMenu)
  { if (mask & BUTTON1_CLICKED) cm_handle_mouseclick(x, _y);
    return;
  }
#endif
  if (y > LINES - 2)
  { check_other_buttons:
    _ae = INVALID_AE;
    do_check_other_buttons:
    if (mask & (BUTTON2_CLICKED | BUTTON3_CLICKED))
    {
#if OPTION_MENUS
      cm_setup(x, _y, _ae);
#else
      menus_were_disabled();
#endif
    }
    return;
  }
  if (visible_window[1] == NULL) i = 0;
  else
  { if (y == VMIDDLE) goto check_other_buttons;
    else if (y < VMIDDLE) i = 0;
    else { i = 1; y -= (VMIDDLE + 1); }
  }
  window = visible_window[i];
  view = window->current_view;
  if ( (view == NULL) || (view->aenum < 1) ) goto check_other_buttons;
  yt = y + view->origin_y; aes = view->active_element;
  for (_ae = 0; _ae < view->aenum; _ae++)
  { const tActiveElementCoordinates* aec = aes[_ae].aec;
    while (aec != NULL)
    { tLinenumber yc = aec->y;
      short x1 = aec->x1, x2 = aec->x2;
      if ( (yt == yc) && (x >= x1) && (x <= x2) )
      { if (_ae != view->aecur) activate_element(view, _ae);
        if (!(mask & BUTTON1_CLICKED)) goto do_check_other_buttons;
        if (view->request->resource->aebase[_ae].kind == aekLink)
        { javascript_handle_event(jekClick, _ae);
          prepare_resource_request(window,
            view->request->resource->aebase[_ae].data, prrfRedrawOne, view,
            NULL);
        }
        return; /* done */
      }
      aec = aec->next;
    }
  }
  goto check_other_buttons;
}
#endif /* #if CONFIG_DO_TEXTMODEMOUSE */

static one_caller tMbsIndex lookup_command_key(tKey key)
{ if (keymap_command_keys_num <= 0) return(INVALID_INDEX); /* "can't happen" */
  my_binary_search(0, keymap_command_keys_num - 1,
    my_numcmp(key, keymap_command_keys[idx].key), return(idx))
}

static one_caller void handle_key(tKey key)
{
#if CONFIG_DO_TEXTMODEMOUSE
  if (key == KEY_MOUSE) { handle_key_mouse(); return; }
#endif
#if (TGC_IS_CURSES) && (defined(KEY_RESIZE))
  if (key == KEY_RESIZE)
  { /* might happen if the curses library installed its own resize signal
       handler and our MIGHT_RESIZE_TERMINAL was 0 */
    if (key_handling_mode == khmLineInput) line_input_resize();
#if OPTION_MENUS
    else if (key_handling_mode == khmContextualMenu) cm_cancel();
#endif
    window_redraw_all();
    return;
  }
#endif

  remap_key(&key);
  if (is_khm_command)
  { if (key != KEY_CANCEL)
    { tMbsIndex idx = lookup_command_key(key);
      if (idx < 0) show_message(_("Unknown keyboard command"), truE);
      else handle_command_code(keymap_command_keys[idx].pcc);
    }
  }
#if TGC_IS_CURSES
  else if (key_handling_mode == khmLineInput) line_input_handle_key(key);
#if OPTION_MENUS
  else if (key_handling_mode == khmContextualMenu)
  { switch (key)
    { case KEY_ENTER: cm_apply(); break;
      case KEY_CANCEL: cm_cancel(); break;
      case KEY_UP: cm_line_up(); break;
      case KEY_DOWN: cm_line_down(); break;
      case KEY_PPAGE: cm_page_up(); break;
      case KEY_NPAGE: cm_page_down(); break;
    }
  }
#endif
#endif /* #if TGC_IS_CURSES */
}

#endif /* #if TGC_IS_CURSES */

static my_inline void copy_error(tResourceRequest* request,
  const tResource* resource)
/* CHECKME! */
{ if ( (resource->state == rsError) && (request->state != rrsError) )
  { request->state = rrsError; request->error = resource->error; }
}

static void handle_request_change(tResourceRequest* request)
{ tResource* resource = request->resource;
  tWindowView* view = request->window_view;
  tWindow* window;

  if ( (program_mode == pmDump) || (program_mode == pmDownload) )
  { if (request == pm_request)
    { if (request->state == rrsError)
        fatal_error(0, _(strResourceError[request->error]));
      if (resource != NULL)
      { if (resource->state == rsError)
        { tResourceError re = resource->error;
          const char* text;
#if OPTION_TLS
          if ( (re == reTls) && (tls_errtext(resource, strbuf)) ) text=strbuf;
          else
#endif
          { text = _(strResourceError[re]); }
          fatal_error(0, text);
        }
        else if (resource->flags & rfFinal)
        { if (program_mode == pmDump) content_dump_to_fd(resource, 1, 0);
          do_quit();
        }
      }
    }
    return;
  }

  if (resource != NULL) copy_error(request, resource);

  if (request->flags & rrfDownload)
  { if ( (request->state == rrsError) ||
         ( (resource != NULL) && (resource->flags & rfFinal) ) )
      request_deallocate(request);
    goto out;
  }

  if ( (view == NULL) || ( (window = view->window) == NULL ) ) goto out;

  if ( (is_environed) && (window_is_visible(window)) )
  { if (request->flags & rrfResourceChanged)
      request->flags &= ~rrfResourceChanged;
    else if (view->flags & wvfScreenFull) view->flags |= wvfDontDrawContents;
    window_redraw(window);
    cursor_reset_position();
  }

  if (resource != NULL) test_redirection(view, resource);
  out: {}
}

typedef unsigned int tRemainingWorkNum;
my_enum1 enum
{ rwkRequestChanged = 0, rwkResourceChanged = 1, rwkPrr = 2
#if CONFIG_USER_QUERY
  , rwkUserQuery = 3
#endif
} my_enum2(unsigned char) tRemainingWorkKind;
typedef struct
{ void* data;
  tObjectIdentifier id;
  tRemainingWorkKind kind;
} tRemainingWork;

typedef struct
{ const char* uri;
  tPrrFlags prrf;
  tWindowView* view;
} tRwPrr;

static tRemainingWorkNum remaining_work_num = 0, remaining_work_maxnum = 0,
  remaining_work_unhandled = 0;
static tRemainingWork* remaining_work = NULL;

static void __remaining_work_store(void* data, tObjectIdentifier id,
  tRemainingWorkKind kind)
{ tRemainingWork* rw;
  if (remaining_work_num >= remaining_work_maxnum)
  { remaining_work_maxnum += 8;
    remaining_work = (tRemainingWork*) memory_reallocate(remaining_work,
      remaining_work_maxnum * sizeof(tRemainingWork));
  }
  rw = &(remaining_work[remaining_work_num++]);
  rw->data = data; rw->id = id; rw->kind = kind;
  if (id != oiUnknown) object_watch(id);
}

static void remaining_work_store(void* data, tObjectIdentifier id,
  tRemainingWorkKind kind)
{ __remaining_work_store(data, id, kind);
  remaining_work_unhandled++;
}

static one_caller void remaining_work_do(void)
{ tRemainingWorkNum num = remaining_work_num, count;
  const tRemainingWork* rws;
  if (num <= 0) return; /* nothing to do (quite likely) */
  rws = remaining_work;
  remaining_work = NULL;
  remaining_work_num = remaining_work_maxnum = remaining_work_unhandled = 0;
  for (count = 0; count < num; count++)
  { const tRemainingWork* rw = &(rws[count]);
    tObjectIdentifier id = rw->id;
    void* data = rw->data;
    tResource* resource;
    tResourceRequest* request;
    tResourceRequest** reqlist;
    tReqlistNum rn;
    tRwPrr* rwprr;
    if (id != oiUnknown)
    { tBoolean did_vanish = object_did_vanish(id);
      object_unwatch(id);
      if (did_vanish) continue; /* object gone, work gone :-) */
    }
    switch (rw->kind)
    { case rwkRequestChanged:
        request = (tResourceRequest*) data;
        handle_request_change(request);
        break;
      case rwkResourceChanged:
        resource = (tResource*) data;
        loop:
        reqlist = resource->reqlist;
        for (rn = 0; rn < resource->reqlist_num; rn++)
        { handle_request_change(reqlist[rn]);
          if (reqlist != resource->reqlist)
          { /* e.g. queued a new request due to automatic redirection; FIXME!*/
            goto loop;
          }
        }
        break;
      case rwkPrr:
        rwprr = (tRwPrr*) data;
        prepare_resource_request(rwprr->view->window, rwprr->uri, rwprr->prrf,
          rwprr->view, NULL);
        memory_deallocate(rwprr->uri); memory_deallocate(rwprr);
        break;
#if CONFIG_USER_QUERY
      case rwkUserQuery:
        if (!is_khm_command) /* IMPROVEME! (or rather "FIXME!"?) */
          __remaining_work_store(rw->data, rw->id, rw->kind);
        else start_user_query(rw->data);
        break;
#endif
    }
  }
  memory_deallocate(rws);
}

static void test_redirection(tWindowView* view, tResource* resource)
{ if ( (resource->redirection != NULL) &&
       (!(view->flags & wvfHandledRedirection)) )
  { view->flags |= wvfHandledRedirection;
    if (view->flags & wvfSource) { /* nothing */ }
    else if (is_httplike(resource->protocol))
    { const tServerStatusCode ssc = resource->server_status_code;
      if ( (ssc == 301) || (ssc == 302) || (ssc == 303) || (ssc == 307) )
      { tRwPrr* x;
        setup_redirect: debugmsg("redirecting\n");
        x = memory_allocate(sizeof(tRwPrr));
        x->view = view; x->uri = my_strdup(resource->redirection);
        x->prrf = prrfRedrawOne | prrfIsHttpRedirection;
        remaining_work_store(x, oiUnknown, rwkPrr);
      }
    }
    else if (resource->flags & rfHtmlRedirection) goto setup_redirect;
  }
}

void main_callback(void* data, tMainCallbackKind kind)
/* resource.c wants to tell us something; we temporarily store the information
   in the "remaining work" queue first, e.g. in order to avoid infinite loops
   between main.c and resource.c */
{ tResourceRequest* request;
  tResource* resource;
#if CONFIG_USER_QUERY
  tUserQuery* query;
#endif
#if CONFIG_DEBUG
  sprint_safe(strbuf, "main_callback(%p, %d)\n", data, kind);
  debugmsg(strbuf);
#endif
  switch (kind)
  { case mckRequest:
      request = (tResourceRequest*) data;
      remaining_work_store(request, request->id, rwkRequestChanged);
      break;
    case mckResource:
      resource = (tResource*) data;
      remaining_work_store(resource, resource->id, rwkResourceChanged);
      break;
#if CONFIG_USER_QUERY
    case mckUserQuery:
      query = (tUserQuery*) data;
      remaining_work_store(query, query->id, rwkUserQuery);
      break;
#endif
  }
}

#if CONFIG_CONSOLE

typedef signed int tConsoleTaskId;
#define CONSOLE_TASK_INVALID (-1)

typedef struct
{ tResourceRequest* request;
} tConsoleTask;

static tConsoleTaskId console_task_num = 0, console_task_max = 0,
  console_task_curr = CONSOLE_TASK_INVALID;
static char* console_input = NULL;
static size_t console_input_len = 0, console_input_maxlen = 0;

static void console_print_prompt(void)
{ my_write_str(1, "\n> ");
}

static void console_input_append(char ch)
{ if (console_input_len >= console_input_maxlen)
  { console_input_maxlen += 50;
    console_input = memory_reallocate(console_input, console_input_maxlen);
  }
  console_input[console_input_len++] = ch;
}

static one_caller void console_input_interpret(void)
/* IMPLEMENTME! */
{ if (console_input_len <= 0) return; /* nothing entered, nothing to do */
  console_input_append('\0'); /* (for simplicity) */
  console_input_len = 0; /* preparing for the next command */
  my_write_str(1, strNewline);
  if (!strcmp(console_input, strQuit)) do_quit();
  else if (!strcmp(console_input, strHelp))
  { /* preliminary, very poor help text; IMPLEMENTME: more and _()! */
    my_write_str(1,
      "Enter \"quit\" to quit, \"help\" to get this help text.\n");
  }
  else my_write_str(1, "Unknown command. Try \"help\".\n");
  console_print_prompt();
}

static one_caller void console_handle_stdin(void)
{ char ch;
  int err = my_read(0, &ch, sizeof(ch));
  if (err <= 0) { /* nothing */ }
  else if (ch == '\n') console_input_interpret();
  else if (ch == config.console_backspace)
  { if (console_input_len > 0)
    { console_input_len--; my_write_str(1, "\b \b"); }
  }
  else
  { const unsigned char c = (unsigned char) ch;
    if (!is_bad_uchar(c)) { my_write(1, &ch, 1); console_input_append(ch); }
  }
}

#endif /* #if CONFIG_CONSOLE */

static one_caller void __init main_initialize(void)
{ size_t count;
  sfconv[0] = '%'; sfconv[3] = '\0';
#if OPTION_MENUS
  my_memclr_arr(uri_history);
#endif
#if TGC_IS_GRAPHICS
  istrYesUc = my_strdup_ucfirst(_(strYes));
  istrNoUc  = my_strdup_ucfirst(_(strNo));
  default_font =
    gdk_font_load("-adobe-helvetica-*-r-normal--*-*-*-*-*-*-iso8859-1");
#endif

  for (count = 0; count < ARRAY_ELEMNUM(keymap_command_defaultkeys); count++)
  { const tKeymapCommandEntry* x = &(keymap_command_defaultkeys[count]);
    (void) keymap_command_key_do_register(x->key, x->pcc);
  }
  qsort(keymap_command_keys, keymap_command_keys_num,
    sizeof(tKeymapCommandEntry), keymap_command_sorter);
  /* (The qsort() might be slow because the defaultkeys[] was already "quite
      sorted"; don't care much - it's only done once, and only at program
      initialization time, and the array isn't _that_ big:-) */
  for (count = 0; count < ARRAY_ELEMNUM(keymap_lineinput_defaultkeys); count++)
  { const tKeymapLineinputEntry* x = &(keymap_lineinput_defaultkeys[count]);
    (void) keymap_lineinput_key_do_register(x->key, x->liac);
  }
  qsort(keymap_lineinput_keys, keymap_lineinput_keys_num,
    sizeof(tKeymapLineinputEntry), keymap_lineinput_sorter);
}

int main(int argc, const char** argv)
{ /* getting start(l)ed... */
  initialize(argc, argv);
  main_initialize();

  if (is_environed)
  { unsigned short count;
#if CONFIG_SESSIONS
    { const char* filename = config.session_resume;
      if (filename != NULL) session_resume(filename);
    }
    if (windowlisthead == NULL) /* didn't actually resume anything */
#endif
    { if (launch_uri_count == 0)
      { /* The user didn't provide any URI on the command-line, so we use a
           default URI. We don't use the "home" URI because that might cause
           undesired network activity (which might cost money although the user
           maybe launched retawq only to view some local files); the user can
           go there by pressing "h", and I think that's acceptable. :-) */
        launch_uri[launch_uri_count++] = "about:retawq";
      }
    }
    for (count = 0; count < launch_uri_count; count++)
    { tWindow* w = window_create();
#if TGC_IS_CURSES
      if (count < 2) visible_window[count] = w;
#endif
      prepare_resource_request(w, launch_uri[count], prrfNone, NULL, NULL);
    }
    window_redraw_all();
  }
  else if (program_mode == pmDump)
  { const char* uri = config.pm_uri;
    tResourceError re;
    prepare_resource_request(NULL, uri, 0, NULL, &pm_request);
    test_pm_error:
    if (pm_request == NULL)
    { re = reRedirection;
      handle_pm_error: fatal_error(0, _(strResourceError[re]));
    }
    else if (pm_request->state == rrsError)
    { re = pm_request->error; goto handle_pm_error; }
  }
  else if (program_mode == pmDownload)
  { content_download_to_fd(config.pm_uri, 1); goto test_pm_error; }
#if CONFIG_CONSOLE
  else if (program_mode == pmConsole)
  { my_write_str(1, strVersion);
    my_write_str(1, _("\nEntering console mode. Use the command \"help\" for a list of commands.\n"));
    console_print_prompt();
  }
#endif

#if CONFIG_TG == TG_GTK

  if (is_environed) { need_tglib_cleanup = truE; gtk_main(); }

#elif TGC_IS_CURSES

  while (1)
  { int selectcount;
    unsigned short loopcount;
    i18n_cleanup
    while (remaining_work_unhandled > 0) remaining_work_do();
    if (is_environed) (void) refresh();
    loopcount = 0;
    select_loop:
    fds_r = fds_e = fds_read; fds_w = fds_write;
    selectcount = select(highest_select_fd + 1, &fds_r, &fds_w, &fds_e,
      NULL);
    if (selectcount <= 0)
    { if ( (selectcount == -1) && (errno == EINTR) && (++loopcount < 10000) )
        goto select_loop;
      else fatal_error(errno, _("select(main) failed"));
    }
    resource_callback(NULL, mckCalledSelect);

    if (FD_ISSET(fd_keyboard_input, &fds_r))
    { /* the user pressed a key (or similar) */
#if CONFIG_CONSOLE
      if (program_mode == pmConsole) console_handle_stdin();
      else
#endif
      { tKey key = getch();
        if (key != ERR) handle_key(key);
      }
      if (--selectcount <= 0) continue;
    }

    if (FD_ISSET(fd_any2main_read, &fds_r)) /* special signal stuff */
    { static char buf[4], count = 0;
      ssize_t err = my_read(fd_any2main_read, buf + count, 4 - count);
      if (err > 0)
      { count += err;
        if (count == 4) /* got a complete four-byte sequence */
        { count = 0;
          if (buf[0] == 'w') /* resize the terminal */
          { const unsigned char* const temp = (const unsigned char*) buf,
              cols = temp[1], rows = temp[2];
            if ( (COLS != cols) || (LINES != rows) ) /* need resize */
            {
#if OPTION_MENUS
              if (key_handling_mode == khmContextualMenu) cm_cancel();
#endif
              (void) resizeterm(rows, cols); window_redraw_all();
              if (key_handling_mode == khmLineInput) line_input_resize();
            }
          }
#if CONFIG_CONSOLE
          else if (!strncmp(buf, strConsoleDiscard, 4)) /* discard input */
          { console_input_len = 0; console_print_prompt(); }
#endif
          else if (!strncmp(buf, strQuit, 4)) /* terminated by a signal */
          { do_quit_sig(); }
        }
      }
      if (--selectcount <= 0) continue;
    }

#if CONFIG_TG == TG_XCURSES
    xcurses_check_connections(&selectcount);
    if (selectcount <= 0) continue;
#endif

    /* Here we have selectcount > 0, so it seems there's something left to do
       for resource.c: */
    { tMainCallbackFds mcf;
      mcf.selectcount = selectcount;
      resource_callback(&mcf, mckFds);
    }
  }
  /*@notreached@*/ do_quit(); /*@notreached@*/
#endif /* tg */
  return(0); /* avoid compiler warning... */
}
